﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# pragma warning(disable:4512)
# include <Leap.h>
# pragma warning(default:4512)
# pragma comment (lib, "Leap")

namespace s3d
{
	namespace LeapMotion
	{
		struct FingerType
		{
			enum Type
			{
				Thumb,
				Index,
				Middle,
				Ring,
				Pinky,
			};
		};

		struct Finger
		{
			double length;

			double width;

			Vec3 pos;

			Vec3 stabilizedPos;

			Vec3 direction;

			Vec3 joints[5];

			FingerType::Type type;

			int id;
		};

		struct Hand
		{
			/**
			*	手のひらの中央の位置 [mm]
			*/
			Vec3 pos;

			/**
			*	安定するよう調整された手のひらの中央の位置 [mm]
			*/
			Vec3 stabilizedPos;

			/**
			*	手のひらの速度 [mm/s]
			*/
			Vec3 velocity;

			/**
			*	手のひらの方向の単位ベクトル
			*/
			Vec3 normal;

			double palmWidth;

			/**
			*	手のひらから指先への方向の単位ベクトル
			*/
			Vec3 direction;

			double armWidth;

			Vec3 armDirection;

			Vec3 elbowPos;

			Vec3 wristPos;

			/**
			*	手の湾曲にフィットする仮想的な球の中心位置 [mm]
			*/
			Vec3 sphereCenter;

			/**
			*	手の湾曲にフィットする仮想的な球の半径 [mm]
			*/
			double sphereRadius;

			double pinchStrength;

			double grabStrength;

			/**
			*	Leap Motion に認識されている時間 [秒]
			*/
			double timeVisibleSec;

			double confidence;

			Array<Finger> fingers;

			/**
			*	Hand の ID
			*/
			int id;

			bool isLeft;
		};

		struct Tool
		{
			/**
			*	指またはツールが指す方向の単位ベクトル
			*/
			Vec3 direction;

			/**
			*	先端の位置 [mm]
			*/
			Vec3 pos;

			/**
			*	安定するよう調整された先端の位置 [mm]
			*/
			Vec3 stabilizedPos;

			/**
			*	先端の速度 [mm/s]
			*/
			Vec3 velocity;

			/**
			*	Leap Motion に認識されている時間 [秒]
			*/
			double timeVisibleSec;

			/**
			*	推測される長さ [mm]
			*/
			double length;

			/**
			*	推測される幅 [mm]
			*/
			double width;

			/**
			*	Pointable の ID
			*/
			int id;
		};

		/**
		*	@brief Leap Motion が認識するジェスチャーの種類
		*/
		enum class GestureType
		{
			/**
			*	未定義
			*/
			Undefined,

			/**
			*	円を描く動作
			*/
			Circle,

			/**
			*	線を描く動作
			*/
			Swipe,

			/**
			*	ピアノの鍵盤をたたくような動作
			*/
			KeyTap,

			/**
			*	正面にタッチするような動作
			*/
			ScreenTap
		};

		/**
		*	@brief Leap Motion が認識したジェスチャーの状態
		*/
		enum class GestureState
		{
			/**
			*	不正値
			*/
			Invalid,

			/**
			*	ジェスチャーが始まった
			*/
			Start,

			/**
			*	ジェスチャーが継続中である
			*/
			Update,

			/**
			*	ジェスチャーが終了した
			*/
			Stop,
		};

		/**
		*	@brief Leap Motion が認識した Circle ジェスチャーの回転方向
		*/
		enum class CircleClockwiseness
		{
			/**
			*	時計回り
			*/
			Clockwise,

			/**
			*	反時計回り
			*/
			Counterclockwise,
		};

		/**
		*	@brief Leap Motion に認識されたジェスチャー
		*/
		struct Gesture
		{
			/**
			*	ジェスチャーの種類
			*/
			GestureType type = GestureType::Undefined;

			/**
			*	ジェスチャーの名前
			*/
			Property_Get(String, gestureName) const
			{
				const static String gestureNames[]{ L"Undefined", L"Circle", L"Swipe", L"KeyTap", L"ScreenTap" };

				return gestureNames[static_cast<unsigned>(type)];
			}

			/**
			*	ジェスチャーの位置、Circle の場合は円の中心位置 [mm]
			*/
			Vec3 pos;

			/**
			*	ジェスチャーの方向の単位ベクトル
			*/
			Vec3 direction;

			/**
			*	ジェスチャーの速さ [mm/s] （Swipe のみ）
			*/
			double speed = 0.0;

			/**
			*	ジェスチャーの開始位置 [mm]（Swipe のみ）
			*/
			Vec3 startPos = { 0.0, 0.0, 0.0 };

			/**
			*	ジェスチャーの周回数（Circle のみ）
			*/
			double progress = 0.0;

			/**
			*	ジェスチャーの半径（Circle のみ）
			*/
			double radius = 0.0;

			/**
			*	ジェスチャーの回転方向（Circle のみ）
			*/
			CircleClockwiseness clockwiseness = CircleClockwiseness::Clockwise;

			/**
			*	ジェスチャーの状態
			*/
			GestureState state = GestureState::Invalid;

			/**
			*	ジェスチャーの ID
			*/
			int id;

			/**
			*	ジェスチャーに関連付けられた Hand の ID
			*/
			int handID;

			int pointableID;
		};
	}

	Vec3 ToVec3(const Leap::Vector& v)
	{
		return{ v.x, v.y, -v.z };
	}

	class LeapMotionDevice : public ISiv3DAddon
	{
	private:

		Leap::Controller m_controller;

		Array<LeapMotion::Hand> m_hands;

		Array<LeapMotion::Tool> m_tools;

		Array<LeapMotion::Gesture> m_gestures;

		bool m_connected = false;

		LeapMotion::GestureState LeapMotionDevice::convertState(Leap::Gesture::State state)
		{
			switch (state)
			{
			case Leap::Gesture::STATE_START:
				return LeapMotion::GestureState::Start;
			case Leap::Gesture::STATE_UPDATE:
				return LeapMotion::GestureState::Update;
			case Leap::Gesture::STATE_STOP:
				return LeapMotion::GestureState::Stop;
			default:
				return LeapMotion::GestureState::Invalid;
			}
		}

	public:

		static const String& name()
		{
			static const String name = L"LeapMotionSiv3D";
			return name;
		}

		String getName() const override
		{
			return name();
		}

		bool init() override
		{
			return true;
		}

		bool update() override
		{
			m_hands.clear();

			m_tools.clear();

			m_gestures.clear();

			m_connected = m_controller.isConnected();

			const Leap::Frame frame = m_controller.frame();

			for (const auto& hand : frame.hands())
			{
				const Leap::Arm arm = hand.arm();
				LeapMotion::Hand h;
				h.pos = ToVec3(hand.palmPosition());
				h.stabilizedPos = ToVec3(hand.stabilizedPalmPosition());
				h.velocity = ToVec3(hand.palmVelocity());
				h.normal = ToVec3(hand.palmNormal());
				h.palmWidth = hand.palmWidth();
				h.direction = ToVec3(hand.direction());
				h.armWidth = arm.width();
				h.armDirection = ToVec3(arm.direction());
				h.elbowPos = ToVec3(arm.elbowPosition());
				h.wristPos = ToVec3(arm.wristPosition());
				h.sphereCenter = ToVec3(hand.sphereCenter());
				h.sphereRadius = hand.sphereRadius();
				h.pinchStrength = hand.pinchStrength();
				h.grabStrength = hand.grabStrength();
				h.timeVisibleSec = hand.timeVisible();
				h.confidence = hand.confidence();
				h.isLeft = hand.isLeft();
				h.id = hand.id();

				h.fingers.reserve(hand.fingers().count());

				for (const auto& finger : hand.fingers())
				{
					LeapMotion::Finger f;
					f.length = finger.length();
					f.width = finger.width();
					f.pos = ToVec3(finger.tipPosition());
					f.stabilizedPos = ToVec3(finger.stabilizedTipPosition());
					f.direction = ToVec3(finger.direction());
					f.type = static_cast<LeapMotion::FingerType::Type>(finger.type());
					f.id = finger.id();

					f.joints[0] = ToVec3(finger.bone(Leap::Bone::TYPE_METACARPAL).prevJoint());
					f.joints[1] = ToVec3(finger.bone(Leap::Bone::TYPE_PROXIMAL).prevJoint());
					f.joints[2] = ToVec3(finger.bone(Leap::Bone::TYPE_INTERMEDIATE).prevJoint());
					f.joints[3] = ToVec3(finger.bone(Leap::Bone::TYPE_DISTAL).prevJoint());
					f.joints[4] = ToVec3(finger.bone(Leap::Bone::TYPE_DISTAL).nextJoint());

					h.fingers.push_back(f);
				}

				m_hands.push_back(h);
			}

			for (const auto& tool : frame.tools())
			{
				LeapMotion::Tool t;
				t.direction = ToVec3(tool.direction());
				t.pos = ToVec3(tool.tipPosition());
				t.stabilizedPos = ToVec3(tool.stabilizedTipPosition());
				t.velocity = ToVec3(tool.tipVelocity());
				t.timeVisibleSec = tool.timeVisible();
				t.length = tool.length();
				t.width = tool.width();
				t.id = tool.id();

				m_tools.push_back(t);
			}

			for (const auto& gesture : frame.gestures())
			{
				LeapMotion::Gesture g;
				g.id = gesture.id();
				g.handID = gesture.hands().isEmpty() ? -1 : gesture.hands().frontmost().id();
				g.pointableID = gesture.pointables().isEmpty() ? -1 : gesture.pointables().frontmost().id();
				g.state = convertState(gesture.state());

				switch (gesture.type())
				{
				case Leap::Gesture::TYPE_CIRCLE:
					{
						const Leap::CircleGesture& circleGesture = gesture;
						g.type = LeapMotion::GestureType::Circle;
						g.pos = ToVec3(circleGesture.center());
						g.direction = ToVec3(circleGesture.normal());
						g.progress = circleGesture.progress();
						g.radius = circleGesture.radius();
						g.clockwiseness =
							circleGesture.pointable().direction().angleTo(circleGesture.normal()) <= Leap::PI / 4
							? LeapMotion::CircleClockwiseness::Clockwise : LeapMotion::CircleClockwiseness::Counterclockwise;
						break;
					}
				case Leap::Gesture::TYPE_SWIPE:
					{
						const Leap::SwipeGesture& swipeGesture = gesture;
						g.type = LeapMotion::GestureType::Swipe;
						g.pos = ToVec3(swipeGesture.position());
						g.direction = ToVec3(swipeGesture.direction());
						g.speed = swipeGesture.speed();
						g.startPos = ToVec3(swipeGesture.startPosition());
						break;
					}
				case Leap::Gesture::TYPE_KEY_TAP:
					{
						const Leap::KeyTapGesture& keytapGesture = gesture;
						g.type = LeapMotion::GestureType::KeyTap;
						g.pos = ToVec3(keytapGesture.position());
						g.direction = ToVec3(keytapGesture.position());
						break;
					}
				case Leap::Gesture::TYPE_SCREEN_TAP:
					{
						const Leap::ScreenTapGesture& screentapGesture = gesture;
						g.type = LeapMotion::GestureType::ScreenTap;
						g.pos = ToVec3(screentapGesture.position());
						g.direction = ToVec3(screentapGesture.direction());
						break;
					}
				default:
					{
						g.type = LeapMotion::GestureType::Undefined;
						break;
					}
				}

				m_gestures.push_back(g);
			}

			return true;
		}

		bool isConnected() const
		{
			return m_connected;
		}

		void enableGesture(LeapMotion::GestureType type, bool enable)
		{
			switch (type)
			{
			case LeapMotion::GestureType::Circle:
				m_controller.enableGesture(Leap::Gesture::TYPE_CIRCLE, enable); break;
			case LeapMotion::GestureType::Swipe:
				m_controller.enableGesture(Leap::Gesture::TYPE_SWIPE, enable); break;
			case LeapMotion::GestureType::KeyTap:
				m_controller.enableGesture(Leap::Gesture::TYPE_KEY_TAP, enable); break;
			case LeapMotion::GestureType::ScreenTap:
				m_controller.enableGesture(Leap::Gesture::TYPE_SCREEN_TAP, enable); break;
			default:
				break;
			}
		}

		const Array<LeapMotion::Hand>& hands() const
		{
			return m_hands;
		}

		const Array<LeapMotion::Tool>& tools() const
		{
			return m_tools;
		}

		const Array<LeapMotion::Gesture>& gestures() const
		{
			return m_gestures;
		}
	};

	namespace LeapMotion
	{
		inline bool RegisterAddon()
		{
			return Addon::Register<LeapMotionDevice>();
		}

		inline bool IsConnected()
		{
			if (auto p = Addon::GetAddon<LeapMotionDevice>())
			{
				return p->isConnected();
			}
			else
			{
				return false;
			}
		}

		inline Array<Hand> Hands()
		{
			if (auto p = Addon::GetAddon<LeapMotionDevice>())
			{
				return p->hands();
			}
			else
			{
				return{};
			}
		}

		inline Array<Tool> Tools()
		{
			if (auto p = Addon::GetAddon<LeapMotionDevice>())
			{
				return p->tools();
			}
			else
			{
				return{};
			}
		}

		inline Array<Gesture> Gestures()
		{
			if (auto p = Addon::GetAddon<LeapMotionDevice>())
			{
				return p->gestures();
			}
			else
			{
				return{};
			}
		}

		inline void EnableGesture(GestureType type, bool enable = true)
		{
			if (auto p = Addon::GetAddon<LeapMotionDevice>())
			{
				return p->enableGesture(type, enable);
			}
		}
	}
}
