﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <memory>
# include "Siv3DFwd.hpp"
# include "SivIReader.hpp"

namespace s3d
{
	/// <summary>
	/// 読み込み用バイナリファイル
	/// </summary>
	class BinaryReader : public IReader
	{
	private:

		class CBinaryReader;

		std::shared_ptr<CBinaryReader> pReader;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		BinaryReader();

		/// <summary>
		/// バイナリファイルを開きます。
		/// </summary>
		/// <param name="path">
		/// ファイルパス
		/// </param>
		explicit BinaryReader(const FilePath& path);

		/// <summary>
		/// デストラクタ
		/// </summary>
		~BinaryReader() = default;

		/// <summary>
		/// バイナリファイルを開きます。
		/// </summary>
		/// <param name="path">
		/// ファイルパス
		/// </param>
		/// <returns>
		/// ファイルのオープンに成功した場合 true, それ以外の場合は false
		/// </returns>
		bool open(const FilePath& path);

		/// <summary>
		/// バイナリファイルをクローズします。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void close();

		/// <summary>
		/// バイナリファイルがオープンされているかを返します。
		/// </summary>
		/// <returns>
		/// ファイルがオープンされている場合 true, それ以外の場合は false
		/// </returns>
		bool isOpened() const override;

		/// <summary>
		/// バイナリファイルがオープンされているかを返します。
		/// </summary>
		/// <returns>
		/// ファイルがオープンされている場合 true, それ以外の場合は false
		/// </returns>
		explicit operator bool() const { return isOpened(); }

		/// <summary>
		/// バイナリファイルのサイズを返します。
		/// </summary>
		/// <returns>
		/// バイナリファイルのサイズ（バイト）
		/// </returns>
		long long size() const override;

		/// <summary>
		/// 現在の読み込み位置を返します。
		/// </summary>
		/// <returns>
		/// 現在の読み込み位置（バイト）
		/// </returns>
		long long getPos() const override;

		/// <summary>
		/// 読み込み位置を変更します。
		/// </summary>
		/// <param name="pos">
		/// 新しい読み込み位置（バイト）
		/// </param>
		/// <returns>
		/// 読み込み位置の変更に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool setPos(long long pos) override;

		/// <summary>
		/// ファイルを読み飛ばし、読み込み位置を変更します。
		/// </summary>
		/// <param name="offset">
		/// 読み飛ばすサイズ（バイト）
		/// </param>
		/// <returns>
		/// 新しい読み込み位置
		/// </returns>
		long long skip(long long offset) override;

		/// <summary>
		/// ファイルからデータを読み込みます。
		/// </summary>
		/// <param name="buffer">
		/// 読み込み先
		/// </param>
		/// <param name="size">
		/// 読み込むサイズ（バイト）
		/// </param>
		/// <returns>
		/// 実際に読み込んだサイズ（バイト）
		/// </returns>
		long long read(void* buffer, long long size) override;

		/// <summary>
		/// 読み込み位置を変更しないデータ読み込みをサポートしているかを示します。
		/// </summary>
		/// <returns>
		/// つねに true
		/// </returns>
		bool supportsLookahead() const override { return true; }

		/// <summary>
		/// 読み込み位置を変更しないでファイルからデータを読み込みます。
		/// </summary>
		/// <param name="buffer">
		/// 読み込み先
		/// </param>
		/// <param name="size">
		/// 読み込むサイズ（バイト）
		/// </param>
		/// <returns>
		/// 現在の読み込み位置（バイト）
		/// </returns>
		long long lookahead(void* buffer, long long size) const override;

		/// <summary>
		/// 読み込み位置を変更しないでファイルからデータを読み込みます。
		/// </summary>
		/// <param name="buffer">
		/// 読み込み先
		/// </param>
		/// <param name="pos">
		/// 先頭から数えた読み込み開始位置（バイト）
		/// </param>
		/// <param name="size">
		/// 読み込むサイズ（バイト）
		/// </param>
		/// <returns>
		/// 現在の読み込み位置（バイト）
		/// </returns>
		long long lookahead(void* buffer, long long pos, long long size) const override;

		/// <summary>
		/// 現在の読み込み位置以降のすべてのデータを読み込み、MemoryReader として返します。
		/// </summary>
		/// <returns>
		/// 現在の読み込み位置以降のすべてのデータ
		/// </returns>
		MemoryReader readWhole();

		/// <summary>
		/// 現在の読み込み位置から指定したサイズのデータを読み込み、MemoryReader として返します。
		/// </summary>
		/// <param name="size">
		/// 読み込むサイズ（バイト）
		/// </param>
		/// <returns>
		/// 読み込んだデータ
		/// </returns>
		MemoryReader readSubset(long long size);

		/// <summary>
		/// 読み込み範囲を指定してデータを読み込み、MemoryReader として返します。
		/// </summary>
		/// <param name="pos">
		/// 先頭から数えた読み込み開始位置（バイト）
		/// </param>
		/// <param name="size">
		/// 読み込むサイズ（バイト）
		/// </param>
		/// <returns>
		/// 読み込んだデータ
		/// </returns>
		MemoryReader readSubset(long long pos, long long size);

		/// <summary>
		/// オープンしているファイルのパスを示します。
		/// </summary>
		/// <remarks>
		/// クローズしている場合は空の文字列です。
		/// </remarks>
		Property_Get(FilePath, path) const;
	};
}
