﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivOptional.hpp"
# include "SivPointVector.hpp"
# include "SivQuaternion.hpp"
# include "SivMat4x4.hpp"

namespace s3d
{
	/// <summary>
	/// 直方体
	/// </summary>
	struct Box
	{
		Quaternion rotation = Quaternion::Identity();

#pragma warning(disable:4201)
		union
		{
			struct { Vec3 center; };
			struct { double x, y, z; };
		};

		union
		{
			struct { Vec3 size; };
			struct { double w, h, d; };
		};
#pragma warning(default:4201)

		Box() {}

		explicit Box(
			double _size,
			const Quaternion& _rotation = Quaternion::Identity()
			)
			: Box({ 0, 0, 0 }, { _size, _size, _size }, _rotation) {}

		Box(
			double _x,
			double _y,
			double _z,
			double _size,
			const Quaternion& _rotation = Quaternion::Identity()
			)
			: Box({ _x, _y, _z }, { _size, _size, _size }, _rotation) {}


		Box(
			const Vec3& pos,
			double _size,
			const Quaternion& _rotation = Quaternion::Identity()
			)
			: Box(pos, { _size, _size, _size }, _rotation) {}

		Box(
			double _x,
			double _y,
			double _z,
			double _w,
			double _h,
			double _d,
			const Quaternion& _rotation = Quaternion::Identity()
			)
			: Box({ _x, _y, _z }, { _w, _h, _d }, _rotation) {}

		Box(
			const Vec3& pos,
			double _w,
			double _h,
			double _d,
			const Quaternion& _rotation = Quaternion::Identity()
			)
			: Box(pos, { _w, _h, _d }, _rotation) {}

		Box(
			double _x,
			double _y,
			double _z,
			const Vec3& _size,
			const Quaternion& _rotation = Quaternion::Identity()
			)
			: Box({ _x, _y, _z }, _size, _rotation) {}

		Box(
			const Vec3& pos,
			const Vec3& _size,
			const Quaternion& _rotation = Quaternion::Identity()
			)
			: center(pos)
			, size(_size)
			, rotation(_rotation) {}


		Box& setPos(double _x, double _y, double _z) { return setPos({ _x, _y, _z }); }

		Box& setPos(const Vec3& _pos) { center.set(_pos); return *this; }

		Box& setSize(double _size) { return setSize({ _size, _size, _size }); }

		Box& setSize(double _w, double _h, double _d) { return setSize({ _w, _h, _d }); }

		Box& setSize(const Vec3& _size) { size.set(_size); return *this; }

		Box& moveBy(double _x, double _y, double _z) { return moveBy({ _x, _y, _z }); }

		Box& moveBy(const Vec3& v) { center.moveBy(v); return *this; }


		Box movedBy(double _x, double _y, double _z) const { return movedBy({ _x, _y, _z }); }

		Box movedBy(const Vec3& v) const { return{ center.movedBy(v), size, rotation }; }

		Box stretched(double s) const { return stretched(s, s, s); }

		Box stretched(double _x, double _y, double _z) const { return{ center, size.movedBy(_x, _y, _z), rotation }; }

		Box scaled(double s) const { return scaled({ s, s, s }); }

		Box scaled(double sx, double sy, double sz) const { return scaled({ sx, sy, sz }); }

		Box scaled(const Vec3& s) const { return{ center, size * s, rotation }; }

		Box rollPitchYaw(double roll, double pitch, double yaw) const { return{ center, size, rotation.rollPitchYaw(roll, pitch, yaw) }; }

		Box rollPitchYaw(const Quaternion& _rotation) const { return{ center, size, rotation * _rotation }; }

		Mat4x4 __vectorcall getMatrix() const { return Mat4x4::AffineTransform(size, rotation, center); }

		/**
		*
		*	@param primitive
		*	@return
		*/
		template<typename Primitive>
		bool intersects(const Primitive& primitive) const
		{
			return Geometry3D::Intersect(*this, primitive);
		}

		template<typename Primitive>
		bool contains(const Primitive& primitive) const
		{
			return Geometry3D::Contains(*this, primitive);
		}

		void draw(const ColorF& color = ColorF(1.0)) const;

		void draw(const Texture& texture, const ColorF& color = ColorF(1.0)) const;

		void drawForward(const ColorF& color = ColorF(1.0)) const;

		void drawForward(const Texture& texture, const ColorF& color = ColorF(1.0)) const;
	};
}
