﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <memory>
# include "Siv3DFwd.hpp"
# include "SivParse.hpp"
# include "SivOptional.hpp"
# include "SivString.hpp"
# include "SivPropertyMacro.hpp"

namespace s3d
{
	/// <summary>
	/// CSV ファイルの読み込み
	/// </summary>
	class CSVReader
	{
	private:

		class CCSVReader;

		std::shared_ptr<CCSVReader> pFile;

		Optional<String> getItem(unsigned row, unsigned column) const;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		CSVReader();
		
		/// <summary>
		/// CSV ファイルを開きます。
		/// </summary>
		/// <param name="path">
		/// ファイルパス
		/// </param>
		explicit CSVReader(const FilePath& path);
		
		/// <summary>
		/// デストラクタ
		/// </summary>
		~CSVReader();

		/// <summary>
		/// CSV ファイルを開きます。
		/// </summary>
		/// <param name="path">
		/// ファイルパス
		/// </param>
		/// <returns>
		/// ファイルのオープンに成功した場合 true, それ以外の場合は false
		/// </returns>
		bool open(const FilePath& path);

		/// <summary>
		/// CSV ファイルをクローズします。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void close();

		/// <summary>
		/// CSV ファイルがオープンされているかを返します。
		/// </summary>
		/// <returns>
		/// ファイルがオープンされている場合 true, それ以外の場合は false
		/// </returns>
		bool isOpened() const;

		/// <summary>
		/// CSV ファイルの内容が変更されたかを返します。
		/// </summary>
		/// <remarks>
		/// ファイルの内容が変更された場合、reload() を呼ぶまで true を返します。
		/// </remarks>
		/// <returns>
		/// CSV ファイルの内容が変更された場合 true, それ以外の場合は false
		/// </returns>
		bool hasChanged() const;

		/// <summary>
		/// CSV ファイルが変更された場合、データの内容を更新します。
		/// </summary>
		/// <returns>
		/// 更新に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool reload();

		/// <summary>
		/// CSV ファイルがオープンされているかを返します。
		/// </summary>
		/// <returns>
		/// ファイルがオープンされている場合 true, それ以外の場合は false
		/// </returns>
		explicit operator bool() const { return isOpened(); }

		/// <summary>
		/// 指定した行の列数を返します。
		/// </summary>
		/// <param name="row">
		/// 行
		/// </param>
		/// <returns>
		/// 指定した行の列数
		/// </returns>
		unsigned columns(unsigned row) const;

		/// <summary>
		/// 指定した位置のデータを取得します。
		/// </summary>
		/// <param name="row">
		/// 行
		/// </param>
		/// <param name="column">
		/// 列
		/// </param>
		/// <returns>
		/// 指定した位置のデータ。変換に失敗した場合はデフォルトの値
		/// </returns>
		template<class Type> Type get(unsigned row, unsigned column) const
		{
			if (auto opt = getOpt<Type>(row, column))
			{
				return opt.value();
			}

			return Type();
		}

		/// <summary>
		/// 指定した位置のデータを取得します。
		/// </summary>
		/// <param name="row">
		/// 行
		/// </param>
		/// <param name="column">
		/// 列
		/// </param>
		/// <param name="defaultValue">
		/// 変換に失敗したときに返す値
		/// </param>
		/// <returns>
		/// 指定した位置のデータ。変換に失敗した場合は defaultValue
		/// </returns>
		template<class Type> Type getOr(unsigned row, unsigned column, Type&& defaultValue) const
		{
			return getOpt<Type>(row, column).value_or(std::move(defaultValue));
		}

		/// <summary>
		/// 指定した位置のデータを取得します。
		/// </summary>
		/// <param name="row">
		/// 行
		/// </param>
		/// <param name="column">
		/// 列
		/// </param>
		/// <returns>
		/// 指定した位置のデータの Optional 型。変換に失敗した場合は none
		/// </returns>
		template<typename Type> Optional<Type> getOpt(unsigned row, unsigned column) const
		{
			if (const auto item = getItem(row, column))
			{
				return ParseOpt<Type>(item.value());
			}

			return none;
		}

		/// <summary>
		/// CSV データが空かどうかを示します。
		/// </summary>
		Property_Get(bool, isEmpty) const;

		/// <summary>
		/// CSV データの行数を示します。
		/// </summary>
		Property_Get(unsigned, rows) const;

		/// <summary>
		/// CSV ファイルのパスを示します。
		/// </summary>
		Property_Get(FilePath, path) const;
	};
}
