﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "SivHalfFloat.hpp"
# include "SivColor.hpp"
# include "SivTextureFormat.hpp"

namespace s3d
{
	struct R32F
	{
		const static TextureFormat Format = TextureFormat::R32_Float;

		float r;

		R32F()
		{

		}

		R32F(float _r)
			: r(_r)
		{

		}

		Color toColor() const
		{
			unsigned char cr = r >= 1.0f ? 255 : r <= 0.0f ? 0 : static_cast<unsigned char>(r * 255.0f + 0.5f);

			return{ cr, 255 };
		}

		ColorF toColorF() const
		{
			return{ r, 1.0 };
		}

		void fromColor(const Color& color)
		{
			r = color.r / 255.0f;
		}

		void fromColorF(const ColorF& color)
		{
			r = static_cast<float>(color.r);
		}
	};

	union RGBA16F
	{
		const static TextureFormat Format = TextureFormat::R16G16B16A16_Float;

# pragma warning ( disable: 4201 )	
		struct
		{
			HalfFloat r;
			HalfFloat g;
			HalfFloat b;
			HalfFloat a;
		};
# pragma warning ( default: 4201 )	

		uint16_t elem[4];

		uint64_t v;

		RGBA16F()
		{

		}

		RGBA16F(float _r, float _g, float _b, float _a)
		{
			const float values[4] = { _r, _g, _b, _a };

			DirectX::PackedVector::XMConvertFloatToHalfStream(elem, sizeof(HalfFloat), values, sizeof(float), 4);
		}

		Color toColor() const
		{
			float values[4];

			DirectX::PackedVector::XMConvertHalfToFloatStream(values, sizeof(float), elem, sizeof(HalfFloat), 4);

			unsigned char cr = values[0] >= 1.0f ? 255 : values[0] <= 0.0f ? 0 : static_cast<unsigned char>(values[0] * 255.0f + 0.5f);
			unsigned char cg = values[1] >= 1.0f ? 255 : values[1] <= 0.0f ? 0 : static_cast<unsigned char>(values[1] * 255.0f + 0.5f);
			unsigned char cb = values[2] >= 1.0f ? 255 : values[2] <= 0.0f ? 0 : static_cast<unsigned char>(values[2] * 255.0f + 0.5f);
			unsigned char ca = values[3] >= 1.0f ? 255 : values[3] <= 0.0f ? 0 : static_cast<unsigned char>(values[3] * 255.0f + 0.5f);

			return{ cr, cg, cb, ca };
		}

		ColorF toColorF() const
		{
			float values[4];

			DirectX::PackedVector::XMConvertHalfToFloatStream(values, sizeof(float), elem, sizeof(HalfFloat), 4);

			return{ values[0], values[1], values[2], values[3] };
		}

		void fromColor(const Color& color)
		{
			static const __m128 inv255 = ::_mm_set_ps1(1.0f / 255.0f);

			const __m128 v = ::_mm_mul_ps(::_mm_cvtepi32_ps(::_mm_set_epi32(color.a, color.b, color.g, color.r)), inv255);

			DirectX::PackedVector::XMConvertFloatToHalfStream(elem, sizeof(HalfFloat), v.m128_f32, sizeof(float), 4);
		}

		void fromColorF(const ColorF& color)
		{
			r.set(static_cast<float>(color.r));
			g.set(static_cast<float>(color.g));
			b.set(static_cast<float>(color.b));
			a.set(static_cast<float>(color.a));
		}
	};

	union RGBA32F
	{
		const static TextureFormat Format = TextureFormat::R32G32B32A32_Float;

		__m128 v;

# pragma warning ( disable: 4201 )	
		struct
		{
			float r, g, b, a;
		};
# pragma warning ( default: 4201 )	

		RGBA32F()
		{

		}

		RGBA32F(__m128 _v)
			: v(_v)
		{

		}

		RGBA32F(float _r, float _g, float _b, float _a)
			: v(::_mm_setr_ps(_r, _g, _b, _a))
		{

		}

		Color toColor() const
		{
			static const __m128i toColorShuffleMask = ::_mm_setr_epi8(0, 4, 8, 12, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1);
			const __m128 sv = ::_mm_min_ps(::_mm_max_ps(v, ::_mm_setzero_ps()), ::_mm_set_ps1(1.0f));
			const __m128 result = ::_mm_mul_ps(sv, ::_mm_set_ps1(255.0f));
			const __m128i r = _mm_cvtps_epi32(result);

			Color color;
			color.c = ::_mm_cvtsi128_si32(::_mm_shuffle_epi8(r, toColorShuffleMask));
			return color;
		}

		ColorF toColorF() const
		{
			return{ r, g, b, a };
		}

		void fromColor(const Color& color)
		{
			static const __m128 inv255 = ::_mm_set_ps1(1.0f / 255.0f);
			v = ::_mm_mul_ps(::_mm_cvtepi32_ps(::_mm_set_epi32(color.a, color.b, color.g, color.r)), inv255);
		}

		void fromColorF(const ColorF& color)
		{
			r = static_cast<float>(color.r);
			g = static_cast<float>(color.g);
			b = static_cast<float>(color.b);
			a = static_cast<float>(color.a);
		}
	};
}
