﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "SivPropertyMacro.hpp"
# include "SivString.hpp"
# include "SivDuration.hpp"

namespace s3d
{
	/// <summary>
	/// 日付
	/// </summary>
	struct Date
	{
		/// <summary>
		/// 西暦
		/// </summary>
		int year;

		/// <summary>
		/// 月 [1-12]
		/// </summary>
		int month;

		/// <summary>
		/// 日 [1-31]
		/// </summary>
		int day;

		/// <summary>
		/// 曜日
		/// </summary>
		enum DayOfWeek : int
		{
			/// <summary>
			/// 日曜日
			/// </summary>
			Sunday,

			/// <summary>
			/// 月曜日
			/// </summary>
			Monday,

			/// <summary>
			/// 火曜日
			/// </summary>
			Tuesday,

			/// <summary>
			/// 水曜日
			/// </summary>
			Wednesday,

			/// <summary>
			/// 木曜日
			/// </summary>
			Thursday,

			/// <summary>
			/// 金曜日
			/// </summary>
			Friday,

			/// <summary>
			/// 土曜日
			/// </summary>
			Saturday
		};

		/// <summary>
		/// 曜日を示します。
		/// </summary>
		Property_Get(int, dayOfWeek) const;

		/// <summary>
		/// 曜日を日本語で示します。
		/// </summary>
		Property_Get(String, dayOfWeekJP) const;

		/// <summary>
		/// 曜日を英語で示します。
		/// </summary>
		Property_Get(String, dayOfWeekEN) const;

		/// <summary>
		/// 現在のローカルの年月日と一致するかを示します。
		/// </summary>
		Property_Get(bool, isToday) const;

		/// <summary>
		/// うるう年であるかを示します。
		/// </summary>
		Property_Get(bool, isLeapYear) const;

		/// <summary>
		/// 月の日数を示します。
		/// </summary>
		Property_Get(int, daysInMonth) const;

		/// <summary>
		/// 1 年の日数を示します。
		/// </summary>
		Property_Get(int, daysInYear) const	{ return isLeapYear ? 366 : 365; }

		/// <summary>
		/// 日付の妥当性を示します。
		/// </summary>
		/// <remarks>
		/// 日付が正しい範囲の値であれば true, それ以外の場合は false
		/// </remarks>
		Property_Get(bool, isValid) const;

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		Date() {}

		/// <summary>
		/// 日付を作成します。
		/// </summary>
		/// <param name="_year">
		/// 西暦
		/// </param>
		/// <param name="_month">
		/// 月
		/// </param>
		/// <param name="_day">
		/// 日
		/// </param>
		explicit Date(int _year, int _month = 1, int _day = 1)
			: year(_year)
			, month(_month)
			, day(_day) {}

		/// <summary>
		/// 日付を指定したフォーマットの文字列で返します。
		/// フォーマット指定のパターンは以下の通りです。
		///
		/// yyyy	4 桁の年 (0001-)
		/// yy		2 桁の年 (00-99)
		/// y		年 (1-)
		/// MMMM	英語の月 (January-December)
		/// MMM		英語の月の略称 (Jan-Dec)
		/// MM		2 桁の月 (01-12)
		/// M		1-2 桁の月 (1-12)
		/// dd		2 桁の日 (01-31)
		/// d		1-2 桁の日 (1-31)
		/// EEEE	英語の曜日 (Sunday-Saturday)
		/// EEE		英語の曜日の略称 (Sun-Sat)
		/// E		日本語の曜日 (日-土)
		///
		/// 引用符で囲まれていないアルファベットはパターン文字として解釈されます。
		/// '' は単一引用符を表します。
		/// </summary>
		/// <param name="format">
		/// フォーマット指定
		/// </param>
		/// <returns>
		/// フォーマットされた日付
		/// </returns>
		String format(const String& format = L"yyyy/M/d") const;

		/// <summary>
		/// 昨日のローカルの日付を返します。
		/// </summary>
		/// <returns>
		/// 昨日のローカルの日付
		/// </returns>
		static Date Yesterday();

		/// <summary>
		/// 現在のローカルの日付を返します。
		/// </summary>
		/// <returns>
		/// 現在のローカルの日付
		/// </returns>
		static Date Today();

		/// <summary>
		/// 明日のローカルの日付を返します。
		/// </summary>
		/// <returns>
		/// 明日のローカルの日付
		/// </returns>
		static Date Tomorrow();

		/// <summary>
		/// 日付を進めます。
		/// </summary>
		/// <param name="days">
		/// 進める日数
		/// </param>
		/// <returns>
		/// *this
		/// </returns>
		Date& operator += (const Days& days);

		/// <summary>
		/// 日付を戻します。
		/// </summary>
		/// <param name="days">
		/// 戻す日数
		/// </param>
		/// <returns>
		/// *this
		/// </returns>
		Date& operator -= (const Days& days)
		{
			return *this += (-days);
		}
	};

	/// <summary>
	/// 出力ストリームに日付を渡します。
	/// </summary>
	/// <param name="os">
	/// 出力ストリーム
	/// </param>
	/// <param name="date">
	/// 日付
	/// </param>
	/// <remarks>
	/// L"yyyy/M/d" 形式でフォーマットされます。
	/// </remarks>
	/// <returns>
	/// 渡した後の出力ストリーム
	/// </returns>
	inline std::wostream & operator << (std::wostream& os, const Date& date)
	{
		return os << date.format();
	}

	/// <summary>
	/// 日付を進めます。
	/// </summary>
	/// <param name="date">
	/// 日付
	/// </param>
	/// <param name="days">
	/// 進める日数
	/// </param>
	/// <returns>
	/// 進めた結果の日付
	/// </returns>
	Date operator + (const Date& date, const Days& days);

	/// <summary>
	/// 日付を戻します。
	/// </summary>
	/// <param name="date">
	/// 日付
	/// </param>
	/// <param name="days">
	/// 戻す日数
	/// </param>
	/// <returns>
	/// 戻した結果の日付
	/// </returns>
	inline Date operator - (const Date& date, const Days& days)
	{
		return date + (-days);
	}

	/// <summary>
	/// 2 つの日付の間の日数を計算します。
	/// </summary>
	/// <param name="to">
	/// 終わりの日付
	/// </param>
	/// <param name="from">
	/// はじめの日付
	/// </param>
	/// <returns>
	/// 2 つの日付の間の日数
	/// </returns>
	Days operator - (const Date& to, const Date& from);

	/// <summary>
	/// 2 つの日付が等しいかを調べます。
	/// </summary>
	/// <param name="a">
	/// 比較する日付
	/// </param>
	/// <param name="b">
	/// 比較する日付
	/// </param>
	/// <returns>
	/// 2 つの日付が等しい場合 true, それ以外の場合は false
	/// </returns>
	bool operator == (const Date& a, const Date& b);

	/// <summary>
	/// 2 つの日付が異なるかを調べます。
	/// </summary>
	/// <param name="a">
	/// 比較する日付
	/// </param>
	/// <param name="b">
	/// 比較する日付
	/// </param>
	/// <returns>
	/// 2 つの日付が異なる場合 true, それ以外の場合は false
	/// </returns>
	inline bool operator != (const Date& a, const Date& b)
	{
		return !(a == b);
	}

	/// <summary>
	/// 日付の &lt; 比較を行います。
	/// </summary>
	/// <param name="a">
	/// 比較する日付
	/// </param>
	/// <param name="b">
	/// 比較する日付
	/// </param>
	/// <returns>
	/// 比較結果
	/// </returns>
	bool operator < (const Date& a, const Date& b);

	/// <summary>
	/// 日付の &gt; 比較を行います。
	/// </summary>
	/// <param name="a">
	/// 比較する日付
	/// </param>
	/// <param name="b">
	/// 比較する日付
	/// </param>
	/// <returns>
	/// 比較結果
	/// </returns>
	bool operator > (const Date& a, const Date& b);

	/// <summary>
	/// 日付の &lt;= 比較を行います。
	/// </summary>
	/// <param name="a">
	/// 比較する日付
	/// </param>
	/// <param name="b">
	/// 比較する日付
	/// </param>
	/// <returns>
	/// 比較結果
	/// </returns>
	inline bool operator <= (const Date& a, const Date& b)
	{
		return !(a > b);
	}

	/// <summary>
	/// 日付の &gt;= 比較を行います。
	/// </summary>
	/// <param name="a">
	/// 比較する日付
	/// </param>
	/// <param name="b">
	/// 比較する日付
	/// </param>
	/// <returns>
	/// 比較結果
	/// </returns>
	inline bool operator >= (const Date& a, const Date& b)
	{
		return !(a < b);
	}
}
