﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "SivPropertyMacro.hpp"
# include "SivString.hpp"
# include "SivDate.hpp"

namespace s3d
{
	/// <summary>
	/// 日付と時刻
	/// </summary>
	struct DateTime : Date
	{
		/// <summary>
		/// 時 [0-23]
		/// </summary>
		int hour;

		/// <summary>
		/// 分 [0-59]
		/// </summary>
		int minute;

		/// <summary>
		/// 秒 [0-59]
		/// </summary>
		int second;

		/// <summary>
		/// ミリ秒 [0-999]
		/// </summary>
		int milliseconds;

		/// <summary>
		/// 日付と時刻の妥当性を示します。
		/// </summary>
		/// <remarks>
		/// 日付と時刻が正しい範囲の値であれば true, それ以外の場合は false
		/// </remarks>
		Property_Get(bool, isValid) const;

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		DateTime() {}

		/// <summary>
		/// 日付と時刻を作成します。
		/// </summary>
		/// <param name="_year">
		/// 年
		/// </param>
		/// <param name="_month">
		/// 月 [1-12]
		/// </param>
		/// <param name="_day">
		/// 日 [1-31]
		/// </param>
		/// <param name="_hour">
		/// 時 [0-23]、デフォルトでは 0
		/// </param>
		/// <param name="_minute">
		/// 分 [0-59]、デフォルトでは 0
		/// </param>
		/// <param name="_second">
		/// 秒 [0-59]、デフォルトでは 0
		/// </param>
		/// <param name="_milliseconds">
		/// ミリ秒 [0-999]、デフォルトでは 0
		/// </param>
		/// <remarks>
		/// 指定された時刻や日付が存在するかはチェックされません。
		/// </remarks>
		DateTime(int _year, int _month, int _day, int _hour = 0, int _minute = 0, int _second = 0, int _milliseconds = 0)
			: Date(_year, _month, _day)
			, hour(_hour)
			, minute(_minute)
			, second(_second)
			, milliseconds(_milliseconds) {}

		/// <summary>
		/// 日付と時刻を作成します。
		/// </summary>
		/// <param name="date">
		/// 年月日
		/// </param>
		/// <param name="_hour">
		/// 時 [0-23]、デフォルトでは 0
		/// </param>
		/// <param name="_minute">
		/// 分 [0-59]、デフォルトでは 0
		/// </param>
		/// <param name="_second">
		/// 秒 [0-59]、デフォルトでは 0
		/// </param>
		/// <param name="_milliseconds">
		/// ミリ秒 [0-999]、デフォルトでは 0
		/// </param>
		/// <remarks>
		/// 指定された時刻や日付が存在するかはチェックされません。
		/// </remarks>
		DateTime(const Date& date, int _hour = 0, int _minute = 0, int _second = 0, int _milliseconds = 0)
			: Date(date)
			, hour(_hour)
			, minute(_minute)
			, second(_second)
			, milliseconds(_milliseconds) {}

		/// <summary>
		/// 現在のローカルの日付と時刻を返します。
		/// </summary>
		/// <returns>
		/// 現在のローカルの日付と時刻
		/// </returns>
		static DateTime Now();

		/// <summary>
		/// 現在の協定世界時刻 (UTC) の日付と時刻を返します。
		/// </summary>
		/// <returns>
		/// 現在の協定世界時刻 (UTC) の日付と時刻
		/// </returns>
		static DateTime NowUTC();

		/// <summary>
		/// 昨日 0:00 のローカルの日付と時刻を返します。
		/// </summary>
		/// <returns>
		/// 昨日 0:00 のローカルの日付と時刻
		/// </returns>
		static DateTime Yesterday()
		{
			return Date::Yesterday();
		}

		/// <summary>
		/// 今日 0:00 のローカルの日付と時刻を返します。
		/// </summary>
		/// <returns>
		/// 今日 0:00 のローカルの日付と時刻
		/// </returns>
		static DateTime Today()
		{
			return Date::Today();
		};

		/// <summary>
		/// 明日 0:00 のローカルの日付と時刻を返します。
		/// </summary>
		/// <returns>
		/// 明日 0:00 のローカルの日付と時刻
		/// </returns>
		static DateTime Tomorrow()
		{
			return Date::Tomorrow();
		}

		/// <summary>
		/// 日付と時刻を指定したフォーマットの文字列で返します。
		/// フォーマット指定のパターンは以下の通りです。
		///
		/// yyyy	4 桁の年 (0001-)
		/// yy		2 桁の年 (00-99)
		/// y		年 (1-)
		/// MMMM	英語の月 (January-December)
		/// MMM		英語の月の略称 (Jan-Dec)
		/// MM		2 桁の月 (01-12)
		/// M		1-2 桁の月 (1-12)
		/// dd		2 桁の日 (01-31)
		/// d		1-2 桁の日 (1-31)
		/// EEEE	英語の曜日 (Sunday-Saturday)
		/// EEE		英語の曜日の略称 (Sun-Sat)
		/// E		日本語の曜日 (日-土)
		/// a		午前/午後 (AM/PM)
		/// HH		24 時間表記の 2 桁の時 (00-23)
		/// H		24 時間表記の時 (0-23)
		/// hh		12 時間表記の 2 桁の時 (00-11)
		/// h		12 時間表記の時 (0-11)
		/// mm		2 桁の分 (00-59)
		/// ss		2 桁の秒 (00-59)
		/// S		小数点以下 1 桁の秒 (0-9)
		/// SS		小数点以下 2 桁の秒 (00-99)
		/// SSS		小数点以下 3 桁の秒 (000-999)
		///
		/// 引用符で囲まれていないアルファベットはパターン文字として解釈されます。
		/// '' は単一引用符を表します。
		/// </summary>
		/// <param name="format">
		/// フォーマット指定
		/// </param>
		/// <returns>
		/// フォーマットされた日付と時刻
		/// </returns>
		String format(const String& pattern = L"yyyy/M/d HH:mm:ss") const;

		/// <summary>
		/// 日付と時刻を進めます。
		/// </summary>
		/// <param name="days">
		/// 進める日数
		/// </param>
		/// <returns>
		/// *this
		/// </returns>
		DateTime& operator += (const Days& days)
		{
			Date::operator+=(days);
			return *this;
		}

		/// <summary>
		/// 日付と時刻を戻します。
		/// </summary>
		/// <param name="days">
		/// 戻す日数
		/// </param>
		/// <returns>
		/// *this
		/// </returns>
		DateTime& operator -= (const Days& days)
		{
			Date::operator-=(days);
			return *this;
		}

		/// <summary>
		/// 日付と時刻を進めます。
		/// </summary>
		/// <param name="days">
		/// 進める時間（ミリ秒）
		/// </param>
		/// <returns>
		/// *this
		/// </returns>
		DateTime& operator += (const Milliseconds& _milliseconds);

		/// <summary>
		/// 日付と時刻を戻します
		/// </summary>
		/// <param name="days">
		/// 戻す時間（ミリ秒）
		/// </param>
		/// <returns>
		/// *this
		/// </returns>
		DateTime& operator -= (const Milliseconds& _milliseconds)
		{
			return *this += (-_milliseconds);
		}
	};

	/// <summary>
	/// 出力ストリームに日付と時刻を渡します。
	/// </summary>
	/// <param name="os">
	/// 出力ストリーム
	/// </param>
	/// <param name="dateTime">
	/// 日付と時刻
	/// </param>
	/// <remarks>
	/// L"yyyy/M/d HH:mm:ss" 形式でフォーマットされます。
	/// </remarks>
	/// <returns>
	/// 渡した後の出力ストリーム
	/// </returns>
	inline std::wostream & operator << (std::wostream& os, const DateTime& dateTime)
	{
		return os << dateTime.format();
	}

	/// <summary>
	/// 日付と時刻を進めます。
	/// </summary>
	/// <param name="dateTime">
	/// 日付と時刻
	/// </param>
	/// <param name="days">
	/// 進める日数
	/// </param>
	/// <returns>
	/// 進めた結果の日付と時刻
	/// </returns>
	inline DateTime operator + (const DateTime& dateTime, const Days& days)
	{
		return DateTime{ dateTime } += days;
	}

	/// <summary>
	/// 日付と時刻を戻します。
	/// </summary>
	/// <param name="dateTime">
	/// 日付と時刻
	/// </param>
	/// <param name="days">
	/// 戻す日数
	/// </param>
	/// <returns>
	/// 戻した結果の日付と時刻
	/// </returns>
	inline DateTime operator - (const DateTime& dateTime, const Days& days)
	{
		return DateTime{ dateTime } -= days;
	}

	/// <summary>
	/// 日付と時刻を進めます。
	/// </summary>
	/// <param name="dateTime">
	/// 日付と時刻
	/// </param>
	/// <param name="milliseconds">
	/// 進める時間（ミリ秒）
	/// </param>
	/// <returns>
	/// 進めた結果の時刻と日付
	/// </returns>
	inline DateTime operator + (const DateTime& dateTime, const Milliseconds& milliseconds)
	{
		return DateTime{ dateTime } += milliseconds;
	}

	/// <summary>
	/// 日付と時刻を戻します
	/// </summary>
	/// <param name="dateTime">
	/// 日付と時刻
	/// </param>
	/// <param name="milliseconds">
	/// 戻す時間（ミリ秒）
	/// </param>
	/// <returns>
	/// 戻した結果の時刻と日付
	/// </returns>
	inline DateTime operator - (const DateTime& dateTime, const Milliseconds& milliseconds)
	{
		return DateTime{ dateTime } -= milliseconds;
	}

	/// <summary>
	/// 2 つの日付と時刻の間の時間を計算します。
	/// </summary>
	/// <param name="to">
	/// 終わりの日付と時刻
	/// </param>
	/// <param name="from">
	/// はじめの日付と時刻
	/// </param>
	/// <returns>
	/// 2 つの日付の間の時間（ミリ秒）
	/// </returns>
	Milliseconds operator - (const DateTime& a, const DateTime& b);

	/// <summary>
	/// 2 つの日付と時刻が等しいかを調べます。
	/// </summary>
	/// <param name="a">
	/// 比較する日付と時刻
	/// </param>
	/// <param name="b">
	/// 比較する日付と時刻
	/// </param>
	/// <returns>
	/// 2 つの日付と時刻が等しい場合 true, それ以外の場合は false
	/// </returns>
	bool operator == (const DateTime& a, const DateTime& b);

	/// <summary>
	/// 2 つの日付と時刻が異なるかを調べます。
	/// </summary>
	/// <param name="a">
	/// 比較する日付と時刻
	/// </param>
	/// <param name="b">
	/// 比較する日付と時刻
	/// </param>
	/// <returns>
	/// 2 つの日付と時刻が異なる場合 true, それ以外の場合は false
	/// </returns>
	inline bool operator != (const DateTime& a, const DateTime& b)
	{
		return !(a == b);
	}

	/// <summary>
	/// 日付と時刻の &lt; 比較を行います。
	/// </summary>
	/// <param name="a">
	/// 比較する日付と時刻
	/// </param>
	/// <param name="b">
	/// 比較する日付と時刻
	/// </param>
	/// <returns>
	/// 比較結果
	/// </returns>
	bool operator < (const DateTime& a, const DateTime& b);

	/// <summary>
	/// 日付と時刻の &gt; 比較を行います。
	/// </summary>
	/// <param name="a">
	/// 比較する日付と時刻
	/// </param>
	/// <param name="b">
	/// 比較する日付と時刻
	/// </param>
	/// <returns>
	/// 比較結果
	/// </returns>
	bool operator > (const DateTime& a, const DateTime& b);

	/// <summary>
	/// 日付と時刻の &lt;= 比較を行います。
	/// </summary>
	/// <param name="a">
	/// 比較する日付と時刻
	/// </param>
	/// <param name="b">
	/// 比較する日付と時刻
	/// </param>
	/// <returns>
	/// 比較結果
	/// </returns>
	inline bool operator <= (const DateTime& a, const DateTime& b)
	{
		return !(a > b);
	}

	/// <summary>
	/// 日付と時刻の &gt;= 比較を行います。
	/// </summary>
	/// <param name="a">
	/// 比較する日付と時刻
	/// </param>
	/// <param name="b">
	/// 比較する日付と時刻
	/// </param>
	/// <returns>
	/// 比較結果
	/// </returns>
	inline bool operator >= (const DateTime& a, const DateTime& b)
	{
		return !(a < b);
	}
}
