﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <utility>

namespace s3d
{
	enum class DepthFunc : unsigned short
	{
		Never			= 1,
		Less			= 2,
		Equal			= 3,
		LessEqual		= 4,
		Greater			= 5,
		NotEqual		= 6,
		GreaterEqual	= 7,
		Always			= 8
	};

	struct DepthState
	{
		DepthState(
			bool _enable = true,
			bool _writeEnable = true,
			DepthFunc _func = DepthFunc::GreaterEqual
			)
			: enable(_enable)
			, writeEnable(_writeEnable)
			, func(_func) {}

		using DataType = unsigned;

#pragma warning(disable:4201)
		union
		{
			struct
			{
				bool enable;

				bool writeEnable;

				DepthFunc func;
			};

			DataType _data;
		};
#pragma warning(default:4201)

		bool operator == (const DepthState& d) const
		{
			return _data == d._data;
		}

		bool operator != (const DepthState& d) const
		{
			return _data != d._data;
		}

		enum class Predefined
		{
			None,
			TestOnly,
			WriteOnly,
			TestWrite,
			Default2D = None,
			Default3D = TestWrite,
		};

		DepthState(Predefined predefined);

		const static Predefined None = Predefined::None;

		const static Predefined TestOnly = Predefined::TestOnly;

		const static Predefined WriteOnly = Predefined::WriteOnly;

		const static Predefined TestWrite = Predefined::TestWrite;

		const static Predefined Default2D = Predefined::Default2D;

		const static Predefined Default3D = Predefined::Default3D;
	};

	enum class StencilOp : unsigned char
	{
		Keep		= 1,
		Zero		= 2,
		Replace		= 3,
		IncrSat		= 4,
		DecrSat		= 5,
		Invert		= 6,
		Incr		= 7,
		Decr		= 8
	};

	enum class StencilFunc : unsigned char
	{
        Never			= 1,
        Less			= 2,
        Equal			= 3,
        LessEqual		= 4,
        Greater			= 5,
        NotEqual		= 6,
        GreaterEqual	= 7,
        Always			= 8
	};

	struct StencilState
	{
		StencilState(
			bool _enable = false,
			bool _stencilWriteOnly = false,
			StencilFunc _func = StencilFunc::Always,
			StencilOp _passOp = StencilOp::Keep,
			StencilOp _failOp = StencilOp::Keep,
			StencilFunc _backFunc = StencilFunc::Always,
			StencilOp _backPassOp = StencilOp::Keep,
			StencilOp _backFailOp = StencilOp::Keep
			)
			: enable(_enable)
			, stencilWriteOnly(_stencilWriteOnly)
			, func(_func)
			, passOp(_passOp)
			, failOp(_failOp)
			, backFunc(_backFunc)
			, backPassOp(_backPassOp)
			, backFailOp(_backFailOp) {}

		StencilState(
			StencilFunc _func,
			StencilOp _passOp,
			StencilOp _failOp = StencilOp::Keep
			)
			: enable(true)
			, stencilWriteOnly(false)
			, func(_func)
			, passOp(_passOp)
			, failOp(_failOp)
			, backFunc(_func)
			, backPassOp(_passOp)
			, backFailOp(_failOp) {}

		using DataType = unsigned;

#pragma warning(disable:4201)
		union
		{
			struct
			{
				bool enable : 4;

				bool stencilWriteOnly : 4;

				StencilFunc func : 4;

				StencilOp passOp : 4;

				StencilOp failOp : 4;

				StencilFunc backFunc : 4;

				StencilOp backPassOp : 4;

				StencilOp backFailOp : 4;
			};

			DataType _data;
		};
#pragma warning(default:4201)

		bool operator == (const StencilState& s) const
		{
			return _data == s._data;
		}

		bool operator != (const StencilState& s) const
		{
			return _data != s._data;
		}

		enum class Predefined
		{
			None,
			Increment,
			Decrement,
			Replace,
			Default = None,
		};

		StencilState(Predefined predefined);

		const static Predefined None = Predefined::None;

		const static Predefined Increment = Predefined::Increment;

		const static Predefined Decrement = Predefined::Decrement;

		const static Predefined Replace = Predefined::Replace;

		const static Predefined Default = Predefined::Default;

		static StencilState Test(StencilFunc _func)
		{
			return{ _func, StencilOp::Keep };
		}
	};

	struct DepthStencilState
	{
		DepthStencilState(
			const DepthState& _depth = DepthState::Default2D,
			const StencilState& _stencil = StencilState::Default
			)
			: depth(_depth)
			, stencil(_stencil) {}

		using DataType = unsigned __int64;

#pragma warning(disable:4201)
		union
		{
			struct
			{
				DepthState depth;

				StencilState stencil;
			};

			DataType _data;
		};
#pragma warning(default:4201)

		bool operator == (const DepthStencilState& d) const
		{
			return _data == d._data;
		}

		bool operator != (const DepthStencilState& d) const
		{
			return _data != d._data;
		}

		enum class Predefined
		{
			Default2D,
			Default3D,
		};

		DepthStencilState(Predefined predefined);

		const static Predefined Default2D = Predefined::Default2D;

		const static Predefined Default3D = Predefined::Default3D;
	};

	static_assert(sizeof(DepthState) == sizeof(DepthState::DataType), "");
	static_assert(sizeof(StencilState) == sizeof(StencilState::DataType), "");
	static_assert(sizeof(DepthStencilState) == sizeof(DepthStencilState::DataType), "");
}

template<>
struct std::hash<s3d::DepthState>
{
	size_t operator()(const s3d::DepthState& keyVal) const
	{
		return std::hash<s3d::DepthState::DataType>()(keyVal._data);
	}
};

template<>
struct std::hash<s3d::StencilState>
{
	size_t operator()(const s3d::StencilState& keyVal) const
	{
		return std::hash<s3d::StencilState::DataType>()(keyVal._data);
	}
};

template<>
struct std::hash<s3d::DepthStencilState>
{
	size_t operator()(const s3d::DepthStencilState& keyVal) const
	{
		return std::hash<s3d::DepthStencilState::DataType>()(keyVal._data);
	}
};
