﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivPropertyMacro.hpp"
# include "SivString.hpp"

namespace s3d
{
	/// <summary>
	/// @brief 浮動小数点数
	/// </summary>
	/// <remarks>
	/// 浮動小数点数のユーティリティ
	/// </remarks>
	template <typename Type, typename BitsType>
	struct FloatBase
	{
		static_assert(
			sizeof(Type) == sizeof(BitsType),
			"error: sizeof(Type)!=sizeof(BitsType)"
			);

		/// <summary>
		/// 浮動小数点数
		/// </summary>
		union { Type value; BitsType bits; };

		/// <summary>
		/// コンストラクタ
		/// </summary>
		FloatBase() {}

		/// <summary>
		/// 数値を設定します。
		/// </summary>
		/// <param name="_value">
		/// コピーする数値
		/// </param>
		FloatBase(Type _value)
			: value(_value) {}

		/// <summary>
		/// 文字列から数値に変換します。
		/// </summary>
		/// <param name="str">
		/// 変換される文字列（例: L"123.45")
		/// </param>
		FloatBase(StringView str);

		/// <summary>
		/// 組み込みの浮動小数点数型に変換します。
		/// </summary>
		/// <returns>
		/// 組み込みの浮動小数点数型の値
		/// </returns>
		operator Type() const { return value; }

		/// <summary>
		/// 新しい数値を代入します。
		/// </summary>
		/// <param name="_value">
		/// 新しい数値
		/// </param>
		/// <returns>
		/// 新しい数値
		/// </returns>
		FloatBase operator = (Type _value)
		{
			value = _value;
			return *this;
		}

		/// <summary>
		/// 文字列で示された新しい数値を代入します。
		/// </summary>
		/// <param name="str">
		/// 新しい数値
		/// </param>
		/// <returns>
		/// 新しい数値
		/// </returns>
		FloatBase operator = (StringView str)
		{
			return *this = FloatBase<Type, BitsType>(str);
		}

		/// <summary>
		/// 数値を文字列で表します。
		/// </summary>
		/// <param name="decimalPlace">
		/// 小数点以下の最大桁数
		/// </param>
		/// <returns>
		/// 変換した文字列
		/// </returns>
		String toString(int decimalPlace) const;

		/// <summary>
		/// 数値を文字列で表します。
		/// </summary>
		/// <remarks>
		/// 小数点以下最大 3 桁まで表します。
		/// </remarks>
		Property_Get(String, string) const { return toString(3); }

		/// <summary>
		/// 浮動小数点数のビット表現を 2 進数の文字列で表します。
		/// </summary>
		Property_Get(String, binary) const;

		/// <summary>
		/// 浮動小数点数のビット表現を 8 進数の文字列で表します。
		/// </summary>
		Property_Get(String, octal) const;

		/// <summary>
		/// 浮動小数点数のビット表現を 16 進数の文字列で表します。
		/// </summary>
		Property_Get(String, hex) const;

		/// <summary>
		/// 数値が非数 (NaN) であるかを示します。
		/// </summary>
		Property_Get(bool, isNaN) const { return !!::_isnan(value); }

		/// <summary>
		/// 数値が有限値であるかを示します。
		/// </summary>
		Property_Get(bool, isFinite) const { return !!::_finite(value); }

		/// <summary>
		/// 数値が無限であるかを示します。
		/// </summary>
		Property_Get(bool, isInfinity) const { return !!(::_fpclass(value) & (_FPCLASS_NINF | _FPCLASS_PINF)); }

		/// <summary>
		/// 表現できる最大の数値を返します。
		/// </summary>
		/// <returns>
		/// 表現できる最大の数値
		/// </returns>
		static Type Max()
		{
			return std::numeric_limits<Type>::max();
		}

		/// <summary>
		/// 表現できる最小の数値を返します。
		/// </summary>
		/// <returns>
		/// 表現できる最小の数値
		/// </returns>
		static Type Min()
		{
			return std::numeric_limits<Type>::min();
		}

		/// <summary>
		/// 正の無限大を返します。
		/// </summary>
		/// <returns>
		/// 正の無限大
		/// </returns>
		static Type Infinity()
		{
			return std::numeric_limits<Type>::infinity();
		}
	};

	/// <summary>
	/// 単精度浮動小数点数型
	/// </summary>
	using Float = FloatBase<float, unsigned __int32>;

	/// <summary>
	/// 倍精度浮動小数点数型
	/// </summary>
	using Double = FloatBase<double, unsigned __int64>;
}
