﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <utility>
# include <array>
# include <vector>
# include <sstream>
# include "Siv3DFwd.hpp"
# include "SivUtility.hpp"
# include "SivOptional.hpp"
# include "SivArray.hpp"
# include "SivGrid.hpp"
# include "SivString.hpp"

namespace s3d
{
	/// <summary>
	/// 値を指定した文字数の文字列に変換します。
	/// </summary>
	/// <param name="value">
	/// 変換する値
	/// </param>
	/// <param name="padding">
	/// 文字数と詰め文字のペア
	/// </param>
	/// <returns>
	/// 変換した文字列
	/// </returns>
	template<typename Type> inline String Pad(const Type& value, const std::pair<int, wchar_t>& padding)
	{
		return Format(value).padLeft(padding.first, padding.second);
	}

	/// <summary>
	/// Format の内部で使用するデータ
	/// </summary>
	struct FormatData
	{
		String string;

		struct DecimalPlace
		{
			int value = 5;
		} decimalPlace;
	};

	/// <summary>
	/// Format の内部で使用する関数
	/// </summary>
	inline void Format(const FormatData&)
	{
		return;
	}

	/// <summary>
	/// Format の内部で使用する関数
	/// </summary>
	template <typename Type, typename ... Args>
	inline
		void Format(FormatData& formatData, const Type& value, const Args& ... args)
	{
		Formatter(formatData, value);

		Format(formatData, args...);
	}

	/// <summary>
	/// 一連の引数を文字列に変換します。
	/// </summary>
	/// <param name="args">
	/// 変換する値
	/// </param>
	/// <returns>
	/// 引数を文字列に変換して連結した文字列
	/// </returns>
	template <typename ... Args> inline String Format(const Args& ... args)
	{
		FormatData formatData;

		Format(formatData, args...);

		return formatData.string;
	}

	inline String Format() = delete;

	/// <summary>
	/// 引数を文字列に変換します。
	/// </summary>
	/// <param name="ch">
	/// 変換する値
	/// </param>
	/// <returns>
	/// 引数を文字列に変換した文字列
	/// </returns>
	inline String Format(wchar_t ch)
	{
		return String(1, ch);
	}

	/// <summary>
	/// 引数を文字列に変換します。
	/// </summary>
	/// <param name="str">
	/// 変換する文字列
	/// </param>
	/// <returns>
	/// 引数を文字列に変換した文字列
	/// </returns>
	inline String Format(const wchar_t* const str)
	{
		return String(str);
	}

	/// <summary>
	/// 引数を文字列に変換します。
	/// </summary>
	/// <param name="str">
	/// 変換する文字列
	/// </param>
	/// <returns>
	/// 引数を文字列に変換した文字列
	/// </returns>
	inline String Format(const String& str)
	{
		return str;
	}

	/// <summary>
	/// 引数を文字列に変換します。
	/// </summary>
	/// <param name="str">
	/// 変換する文字列
	/// </param>
	/// <returns>
	/// 引数を文字列に変換した文字列
	/// </returns>
	inline String Format(String&& str)
	{
		return String(std::move(str));
	}

	/// <summary>
	/// 浮動小数点数を変換する際の小数点以下の最大桁数を指定するマニピュレータを返します。
	/// </summary>
	/// <param name="width">
	/// 小数点以下の最大桁数
	/// </param>
	/// <remarks>
	/// この関数が返すマニピュレータを Format の引数にすると、
	/// 該当 Format() 内のそれ以降の浮動小数点数の変換に適用されます。
	/// </remarks>
	/// <returns>
	/// Format に渡すマニピュレータ
	/// </returns>
	inline FormatData::DecimalPlace DecimalPlace(int width)
	{
		FormatData::DecimalPlace decimalPlace;
		decimalPlace.value = width;
		return decimalPlace;
	}

	void Formatter(FormatData& formatData, long value);
	void Formatter(FormatData& formatData, unsigned long value);
	void Formatter(FormatData& formatData, long long value);
	void Formatter(FormatData& formatData, unsigned long long value);
	void Formatter(FormatData& formatData, double value);
	void Formatter(FormatData& formatData, const void* value);
	void Formatter(FormatData& formatData, const char* const str) = delete;
	void Formatter(FormatData& formatData, bool value);

	inline void Formatter(FormatData& formatData, signed char value)
	{
		Formatter(formatData, static_cast<long>(value));
	}

	inline void Formatter(FormatData& formatData, unsigned char value)
	{
		Formatter(formatData, static_cast<unsigned long>(value));
	}

	inline void Formatter(FormatData& formatData, char value)
	{
		Formatter(formatData, static_cast<long>(value));
	}

	inline void Formatter(FormatData& formatData, short value)
	{
		Formatter(formatData, static_cast<long>(value));
	}

	inline void Formatter(FormatData& formatData, unsigned short value)
	{
		Formatter(formatData, static_cast<unsigned long>(value));
	}

	inline void Formatter(FormatData& formatData, wchar_t value)
	{
		formatData.string.push_back(value);
	}

	inline void Formatter(FormatData& formatData, int value)
	{
		Formatter(formatData, static_cast<long>(value));
	}

	inline void Formatter(FormatData& formatData, unsigned value)
	{
		Formatter(formatData, static_cast<unsigned long>(value));
	}

	inline void Formatter(FormatData& formatData, float value)
	{
		Formatter(formatData, static_cast<double>(value));
	}

	inline void Formatter(FormatData& formatData, const FormatData::DecimalPlace decimalPlace)
	{
		formatData.decimalPlace = decimalPlace;
	}

	inline void Formatter(FormatData& formatData, const wchar_t* const str)
	{
		formatData.string.append(str);
	}

	inline void Formatter(FormatData& formatData, const String& str)
	{
		formatData.string.append(str);
	}

	template <typename Type>
	inline void Formatter(FormatData& formatData, const Optional<Type>& opt)
	{
		if (opt)
		{
			formatData.string.append(L"Optional ");

			Formatter(formatData, static_cast<Type>(opt.value()));
		}
		else
		{
			formatData.string.append(L"none");
		}
	}

	inline void Formatter(FormatData& formatData, nullopt_t)
	{
		formatData.string.append(L"none");
	}

	template <typename Iterator>
	inline
		void
		Formatter(
		FormatData& formatData,
		Iterator begin,
		Iterator end
		)
	{
		formatData.string.push_back(L'{');

		for (bool b = false; begin != end; ++begin)
		{
			if (Exchange(b, true))
			{
				formatData.string.push_back(L',');
			}

			Formatter(formatData, *begin);
		}

		formatData.string.push_back(L'}');
	}

	template <typename Type, size_t N>
	inline
		void
		Formatter(
		FormatData& formatData,
		const Type(&values)[N]
		)
	{
		Formatter(formatData, std::begin(values), std::end(values));
	}

	template <typename Type, size_t N>
	inline
		void
		Formatter(
		FormatData& formatData,
		const std::array<Type, N>& v
		)
	{
		Formatter(formatData, v.begin(), v.end());
	}

	template <typename Type>
	inline
		void
		Formatter(
		FormatData& formatData,
		const std::vector<Type>& v
		)
	{
		Formatter(formatData, v.begin(), v.end());
	}

	template <typename Type>
	inline
		void
		Formatter(
		FormatData& formatData,
		const aligned_vector<Type>& v
		)
	{
		Formatter(formatData, v.begin(), v.end());
	}

	template <typename Type>
	inline
		void
		Formatter(
		FormatData& formatData,
		const Grid<Type>& g
		)
	{
		formatData.string.push_back(L'{');

		bool b = false;

		for (int y = 0; y < g.height; ++y)
		{
			if (Exchange(b, true))
			{
				formatData.string.append(L',');
			}

			Formatter(formatData, g[y], g[y] + g.width);
		}

		formatData.string.push_back(L'}');
	}

	template <typename Type>
	inline
		void
		Formatter(
		FormatData& formatData,
		const Type& value
		)
	{
		std::wostringstream wos;

		wos << value;

		formatData.string.append(wos.str());
	}

	void Formatter(FormatData& formatData, const Color& value);
	void Formatter(FormatData& formatData, const ColorF& value);
	void Formatter(FormatData& formatData, const HSV& value);
	void Formatter(FormatData& formatData, const Point& value);
	void Formatter(FormatData& formatData, const Float2& value);
	void Formatter(FormatData& formatData, const Vec2& value);
	void Formatter(FormatData& formatData, const Float3& value);
	void Formatter(FormatData& formatData, const Vec3& value);
	void Formatter(FormatData& formatData, const Float4& value);
	void Formatter(FormatData& formatData, const Vec4& value);
	void Formatter(FormatData& formatData, const __m128& value);
	void Formatter(FormatData& formatData, const Circular& value);
	void Formatter(FormatData& formatData, const LineInt& value);
	void Formatter(FormatData& formatData, const Line& value);
	void Formatter(FormatData& formatData, const Rect& value);
	void Formatter(FormatData& formatData, const RectF& value);
	void Formatter(FormatData& formatData, const Circle& value);
	void Formatter(FormatData& formatData, const Ellipse& value);
	void Formatter(FormatData& formatData, const Triangle& value);
	void Formatter(FormatData& formatData, const Quad& value);
	void Formatter(FormatData& formatData, const RoundRect& value);
	void Formatter(FormatData& formatData, const Quaternion& value);
	void Formatter(FormatData& formatData, const Mat4x4& value);
	void Formatter(FormatData& formatData, const Ray& value);
}
