﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivBlendState.hpp"
# include "SivRasterizerState.hpp"
# include "SivDepthStencilState.hpp"
# include "SivSamplerState.hpp"
# include "SivRenderState.hpp"

namespace s3d
{
	/// <summary>
	/// 2D グラフィックス
	/// </summary>
	/// <remarks>
	/// 2D　グラフィックスに関連する機能を提供します。
	/// </remarks>
	namespace Graphics2D
	{
		/// <summary>
		/// 2D 描画のスケールとオフセットを設定します。
		/// </summary>
		/// <param name="">
		/// スケーリング
		/// </param>
		/// <param name="">
		/// オフセット（ピクセル）
		/// </param>
		/// <remarks>
		/// GUI と Println 等による画面表示には適用されません。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void SetTransform(const Vec2& scale = { 1.0, 1.0 }, const Vec2& offset = { 0.0, 0.0 });

		/// <summary>
		/// 現在の 2D 描画のスケールとオフセットを取得します。
		/// </summary>
		/// <returns>
		/// 現在の 2D 描画のスケールとオフセット
		/// </returns>
		std::pair<Vec2, Vec2> GetTransform();

		/// <summary>
		/// 2D 描画のブレンドステートを設定します。
		/// </summary>
		/// <param name="blendState">
		/// ブレンドステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetBlendState(const BlendState& blendState);

		/// <summary>
		/// 現在の 2D 描画のブレンドステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の 2D 描画のブレンドステート
		/// </returns>
		BlendState GetBlandState();

		/// <summary>
		/// 2D 描画のラスタライザーステートを設定します。
		/// </summary>
		/// <param name="rasterizerState">
		/// ラスタライザーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetRasterizerState(const RasterizerState& rasterizerState);

		/// <summary>
		/// 現在の 2D 描画のラスタライザーステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の 2D 描画のラスタライザーステート
		/// </returns>
		RasterizerState GetRasterizerState();

		/// <summary>
		/// 2D 描画のステンシルステートを設定します。
		/// </summary>
		/// <param name="stencilState">
		/// ステンシルステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetStencilState(const StencilState& stencilState);

		/// <summary>
		/// 現在の 2D 描画のステンシルステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の 2D 描画のステンシルステート
		/// </returns>
		StencilState GetStencilState();

		/// <summary>
		/// 2D 描画のステンシル値を設定します。
		/// </summary>
		/// <param name="value">
		/// ステンシル値
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetStencilValue(unsigned char value);

		/// <summary>
		/// 現在の 2D 描画のステンシル値を取得します。
		/// </summary>
		/// <returns>
		/// 現在の 2D 描画のステンシル値
		/// </returns>
		unsigned char GetStencilValue();

		/// <summary>
		/// 2D 描画のサンプラーステートを設定します。
		/// </summary>
		/// <param name="samplerState">
		/// サンプラーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetSamplerState(const SamplerState& samplerState);

		/// <summary>
		/// 2D 描画のサンプラーステートをスロットを指定して設定します。
		/// </summary>
		/// <param name="slot">
		/// スロット [0 または 1]
		/// </param>
		/// <param name="samplerState">
		/// サンプラーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetSamplerState(unsigned slot, const SamplerState& samplerState);

		/// <summary>
		/// 現在の 2D 描画のサンプラーステートを取得します。
		/// </summary>
		/// <param name="slot">
		/// スロット [0 または 1]
		/// </param>
		/// <returns>
		/// 現在の 2D 描画のサンプラーステート
		/// </returns>
		SamplerState GetSamplerState(unsigned slot = 0);

		/// <summary>
		/// 現在の 2D 描画のレンダーステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の 2D 描画のレンダーステート
		/// </returns>
		RenderState GetRenderState();

		/// <summary>
		/// 2D 描画のレンダーステートを設定します。
		/// </summary>
		/// <param name="state">
		/// レンダーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetRenderState(const RenderState& state);
	}
}