﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivBlendState.hpp"
# include "SivRasterizerState.hpp"
# include "SivDepthStencilState.hpp"
# include "SivSamplerState.hpp"
# include "SivRenderState.hpp"
# include "SivCamera.hpp"
# include "SivFog.hpp"
# include "SivLight.hpp"

namespace s3d
{
	/// <summary>
	/// レンダラー
	/// </summary>
	enum class Renderer
	{
		/// <summary>
		/// デフォルトの Deferred レンダラー
		/// </summary>
		Deferred,

		/// <summary>
		/// Forward レンダラー
		/// </summary>
		Forward,
	};

	/// <summary>
	/// アンチエイリアシングのレベル
	/// </summary>
	enum class AntiAliasing
	{
		/// <summary>
		/// なし
		/// </summary>
		None,

		/// <summary>
		/// 低
		/// </summary>
		Low,

		/// <summary>
		/// 中
		/// </summary>
		Medium,

		/// <summary>
		/// 高
		/// </summary>
		High,

		/// <summary>
		/// デフォルト [低]
		/// </summary>
		Default = Low
	};

	/// <summary>
	/// 3D グラフィックス
	/// </summary>
	/// <remarks>
	/// 3D　グラフィックスに関連する機能を提供します。
	/// </remarks>
	namespace Graphics3D
	{
		////////////////////////////////////////////////////////////////
		//
		//		カメラ
		//
		////////////////////////////////////////////////////////////////

		/// <summary>
		/// カメラを設定します。
		/// </summary>
		/// <param name="state">
		/// カメラの設定
		/// </param>
		/// <remarks>
		/// 毎フレーム、最後に適用されたカメラ設定がシーンの描画に使用されます。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void SetCamera(const Camera& state);

		/// <summary>
		/// 現在のカメラの設定を取得します。
		/// </summary>
		/// <returns>
		/// 現在のカメラの設定
		/// </returns>
		const Camera& GetCamera();

		/// <summary>
		/// デバッグ用のカメラを適用します。
		/// </summary>
		/// <param name="speed">
		/// カメラの移動速度
		/// </param>
		/// <remarks>
		/// A, W, S, D, E, X, 矢印キーを使ってカメラを操作できます。
		/// Shift, Ctrl キーを同時に押すと移動速度が上がります。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void FreeCamera(double speed = 0.1);

		/// <summary>
		/// 現在のカメラの設定を使って、ワールド座標をスクリーン座標に変換します。
		/// </summary>
		/// <param name="worldPos">
		/// ワールド座標
		/// </param>
		/// <returns>
		/// スクリーン座標。z は深度
		/// </returns>
		Vec3 ToScreenPos(const Vec3& worldPos);

		/// <summary>
		/// 現在のカメラの設定を使って、スクリーン座標をワールド座標に変換します。
		/// </summary>
		/// <param name="screenPos">
		/// スクリーン座標
		/// </param>
		/// <param name="depth">
		/// 深度
		/// </param>
		/// <returns>
		/// ワールド座標
		/// </returns>
		Vec3 ToWorldPos(const Vec2& screenPos, double depth);

		/// <summary>
		/// 現在のカメラの設定を使って、スクリーン座標をレイに変換します。
		/// </summary>
		/// <param name="screenPos">
		/// スクリーン座標
		/// </param>
		/// <returns>
		/// レイ
		/// </returns>
		Ray ToRay(const Vec2& screenPos);

		////////////////////////////////////////////////////////////////
		//
		//		レンダーステート
		//
		////////////////////////////////////////////////////////////////

		/// <summary>
		/// Forward レンダラ―のブレンドステートを設定します。
		/// </summary>
		/// <param name="blendState">
		/// ブレンドステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetBlendStateForward(const BlendState& blendState);

		/// <summary>
		/// 現在の Forward レンダラ―のブレンドステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の Forward レンダラ―のブレンドステート
		/// </returns>
		BlendState GetBlandStateForward();

		/// <summary>
		/// Deferred レンダラ―のラスタライザーステートを設定します。
		/// </summary>
		/// <param name="rasterizerState">
		/// ラスタライザーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetRasterizerState(const RasterizerState& rasterizerState);

		/// <summary>
		/// Forward レンダラ―のラスタライザーステートを設定します。
		/// </summary>
		/// <param name="rasterizerState">
		/// ラスタライザーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetRasterizerStateForward(const RasterizerState& rasterizerState);

		/// <summary>
		/// 現在の Deferred レンダラ―のラスタライザーステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の Deferred レンダラ―のラスタライザーステート
		/// </returns>
		RasterizerState GetRasterizerState();

		/// <summary>
		/// 現在の Forward レンダラ―のラスタライザーステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の Forward レンダラ―のラスタライザーステート
		/// </returns>
		RasterizerState GetRasterizerStateForward();

		/// <summary>
		/// Deferred レンダラ―のステンシルステートを設定します。
		/// </summary>
		/// <param name="stencilState">
		/// ステンシルステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetStencilState(const StencilState& stencilState);

		/// <summary>
		/// Forward レンダラ―のステンシルステートを設定します。
		/// </summary>
		/// <param name="stencilState">
		/// ステンシルステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetStencilStateForward(const StencilState& stencilState);

		/// <summary>
		/// 現在の Deferred レンダラ―のステンシルステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の Deferred レンダラ―のステンシルステート
		/// </returns>
		StencilState GetStencilState();

		/// <summary>
		/// 現在の Forward レンダラ―のステンシルステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の Forward レンダラ―のステンシルステート
		/// </returns>
		StencilState GetStencilStateForward();

		/// <summary>
		/// Deferred レンダラ―のデプスステートを設定します。
		/// </summary>
		/// <param name="depthState">
		/// デプスステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetDepthState(const DepthState& depthState);

		/// <summary>
		/// Forward レンダラ―のデプスステートを設定します。
		/// </summary>
		/// <param name="depthState">
		/// デプスステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetDepthStateForward(const DepthState& depthState);

		/// <summary>
		/// 現在の Deferred レンダラ―のデプスステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の Deferred レンダラ―のデプスステート
		/// </returns>
		DepthState GetDepthState();

		/// <summary>
		/// 現在の Forward レンダラ―のデプスステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の Forward レンダラ―のデプスステート
		/// </returns>
		DepthState GetDepthStateForward();

		/// <summary>
		/// Deferred レンダラ―のステンシル値を設定します。
		/// </summary>
		/// <param name="value">
		/// ステンシル値
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetStencilValue(unsigned char value);

		/// <summary>
		/// Forward レンダラ―のステンシル値を設定します。
		/// </summary>
		/// <param name="value">
		/// ステンシル値
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetStencilValueForward(unsigned char value);

		/// <summary>
		/// 現在の Deferred レンダラ―のステンシル値を取得します。
		/// </summary>
		/// <returns>
		/// 現在の Deferred レンダラ―のステンシル値
		/// </returns>
		unsigned char GetStencilValue();

		/// <summary>
		/// 現在の Forward レンダラ―のステンシル値を取得します。
		/// </summary>
		/// <returns>
		/// 現在の Forward レンダラ―のステンシル値
		/// </returns>
		unsigned char GetStencilValueForward();

		/// <summary>
		/// Deferred レンダラ―のサンプラーステートを設定します。
		/// </summary>
		/// <param name="samplerState">
		/// サンプラーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetSamplerState(const SamplerState& samplerState);

		/// <summary>
		/// Forward レンダラ―のサンプラーステートを設定します。
		/// </summary>
		/// <param name="samplerState">
		/// サンプラーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetSamplerStateForward(const SamplerState& samplerState);

		/// <summary>
		/// Deferred レンダラ―のサンプラーステートを設定します。
		/// </summary>
		/// <param name="slot">
		/// スロット番号 [0 または 1]
		/// </param>
		/// <param name="samplerState">
		/// サンプラーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetSamplerState(unsigned slot, const SamplerState& samplerState);

		/// <summary>
		/// Forward レンダラ―のサンプラーステートを設定します。
		/// </summary>
		/// <param name="slot">
		/// スロット番号 [0 または 1]
		/// </param>
		/// <param name="samplerState">
		/// サンプラーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetSamplerStateForward(unsigned slot, const SamplerState& samplerState);

		/// <summary>
		/// 現在の Deferred レンダラ―のサンプラーステートを取得します。
		/// </summary>
		/// <param name="slot">
		/// スロット番号 [0 または 1]
		/// </param>
		/// <returns>
		/// 現在の Deferred レンダラ―のサンプラーステート
		/// </returns>
		SamplerState GetSamplerState(unsigned slot = 0);

		/// <summary>
		/// 現在の Forward レンダラ―のサンプラーステートを取得します。
		/// </summary>
		/// <param name="slot">
		/// スロット番号 [0 または 1]
		/// </param>
		/// <returns>
		/// 現在の Forward レンダラ―のサンプラーステート
		/// </returns>
		SamplerState GetSamplerStateForward(unsigned slot = 0);

		/// <summary>
		/// 現在の Deferred レンダラ―のレンダーステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の Deferred レンダラ―のレンダーステート
		/// </returns>
		RenderState GetRenderState();

		/// <summary>
		/// 現在の Forward レンダラ―のレンダーステートを取得します。
		/// </summary>
		/// <returns>
		/// 現在の Forward レンダラ―のレンダーステート
		/// </returns>
		RenderState GetRenderStateForward();

		/// <summary>
		/// Deferred レンダラ―のレンダーステートを設定します。
		/// </summary>
		/// <param name="state">
		/// レンダーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetRenderState(const RenderState& state);

		/// <summary>
		/// Forward レンダラ―のレンダーステートを設定します。
		/// </summary>
		/// <param name="state">
		/// レンダーステート
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetRenderStateForward(const RenderState& state);


		////////////////////////////////////////////////////////////////
		//
		//		ライトとフォグ
		//
		////////////////////////////////////////////////////////////////

		/// <summary>
		/// Deferred レンダラ―にフォグを設定します。
		/// </summary>
		/// <param name="fog">
		/// フォグ
		/// </param>
		/// <remarks>
		/// 毎フレーム、最後に適用された設定がシーンの描画に使用されます。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void SetFog(const Fog& fog);

		/// <summary>
		/// 現在の Deferred レンダラ―のフォグ設定を取得します。
		/// </summary>
		/// <returns>
		/// 現在の Deferred レンダラ―のフォグ設定
		/// </returns>
		const Fog& GetFog();

		/// <summary>
		/// Forward レンダラ―にフォグを設定します。
		/// </summary>
		/// <param name="fog">
		/// フォグ
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetFogForward(const Fog& fog);

		/// <summary>
		/// 現在の Forward レンダラ―のフォグ設定を取得します。
		/// </summary>
		/// <returns>
		/// 現在の Forward レンダラ―のフォグ設定
		/// </returns>
		const Fog& GetFogForward();

		/// <summary>
		/// Deferred レンダラ―に環境光を設定します。
		/// </summary>
		/// <param name="color">
		/// 環境光の強さ
		/// </param>
		/// <remarks>
		/// 毎フレーム、最後に適用された設定がシーンの描画に使用されます。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void SetAmbientLight(const ColorF& color);

		/// <summary>
		/// 現在の Deferred レンダラ―の環境光設定を取得します。
		/// </summary>
		/// <returns>
		/// 現在の Deferred レンダラ―の環境光の強さ
		/// </returns>
		const ColorF& GetAmbientLight();

		/// <summary>
		/// Forward レンダラ―に環境光を設定します。
		/// </summary>
		/// <param name="color">
		/// 環境光の強さ
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetAmbientLightForward(const ColorF& color);

		/// <summary>
		/// 現在の Forward レンダラ―の環境光設定を取得します。
		/// </summary>
		/// <returns>
		/// 現在の Forward レンダラ―の環境光の強さ
		/// </returns>
		const ColorF& GetAmbientLightForward();

		enum
		{
			/// <summary>
			/// Deferred レンダラ―最大の光源数
			/// </summary>
			MaxLight = 128,

			/// <summary>
			/// Deferred レンダラ―最大の光源インデックス
			/// </summary>
			MaxLightIndex = MaxLight - 1,
		};

		/// <summary>
		/// Deferred レンダラ―にライトを設定します。
		/// </summary>
		/// <param name="index">
		/// ライトインデックス [0-127]
		/// </param>
		/// <param name="light">
		/// ライトの設定
		/// </param>
		/// <remarks>
		/// 毎フレーム、ライトインデックスごとに最後に適用された設定がシーンの描画に使用されます。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void SetLight(unsigned index, const Light& light);

		/// <summary>
		/// 現在の Deferred レンダラ―のライトの設定を取得します。
		/// </summary>
		/// <param name="index">
		/// ライトインデックス [0-127]
		/// </param>
		/// <returns>
		/// 現在の Deferred レンダラ―のライトの設定
		/// </returns>
		const Light& GetLight(unsigned index);

		/// <summary>
		/// Forward レンダラ―にライトを設定します。
		/// </summary>
		/// <param name="light">
		/// ライトの設定
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void SetLightForward(const Light& light);

		/// <summary>
		/// 現在の Forward レンダラ―のライトの設定を取得します。
		/// </summary>
		/// <returns>
		/// 現在の Forward レンダラ―のライトの設定
		/// </returns>
		const Light& GetLightForward();


		////////////////////////////////////////////////////////////////
		//
		//		画質
		//
		////////////////////////////////////////////////////////////////

		/// <summary>
		/// アンチエイリアシングの品質を設定します。
		/// </summary>
		/// <param name="antiAliasing">
		/// アンチエイリアシングの設定
		/// </param>
		/// <remarks>
		/// 毎フレーム、最後に適用された設定がシーンの描画に使用されます。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void SetAntiAliasing(AntiAliasing antiAliasing);

		/// <summary>
		/// 現在のアンチエイリアシングの設定を取得します。
		/// </summary>
		/// <returns>
		/// 現在のアンチエイリアシングの設定
		/// </returns>
		AntiAliasing GetAntiAliasing();

		/// <summary>
		/// ライトバッファの品質を変更します。
		/// </summary>
		/// <param name="quality">
		/// ライトバッファの品質。 [1-100] の範囲で指定します
		/// </param>
		/// <remarks>
		/// デフォルトでは 100 で、品質を小さくすると複雑なシーンでの画質が低下しますが、
		/// ライティングにかかるコストは数値に比例して減少します。
		/// quality は 1 から 100 の範囲にクランプされます
		/// </remarks>
		/// <remarks>
		/// 毎フレーム、最後に適用された設定がシーンの描画に使用されます。
		/// </remarks>
		/// <returns>
		/// 正常に変更できた場合 true, メモリ不足などのエラーが発生した場合は false
		/// </returns>
		bool SetLightBufferQuality(unsigned quality = 100);

		/// <summary>
		/// 現在のライトバッファの品質を取得します。
		/// </summary>
		/// <returns>
		/// ライトバッファの品質 [1-100]
		/// </returns>
		unsigned GetLightBufferQuality();

		/// <summary>
		/// カメラ、ライト、フォグ、画質の設定を初期設定に戻します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		inline void ResetScene()
		{
			SetAmbientLight(ColorF(0.1));

			SetAmbientLightForward(ColorF(0.1));

			for (int i = 0; i <= MaxLightIndex; ++i)
			{
				SetLight(0, Light::None());
			}

			SetLight(0, Light::Default());

			SetLightForward(Light::Default());

			SetFog(Fog::None());

			SetFogForward(Fog::None());

			SetCamera(Camera());

			SetAntiAliasing(AntiAliasing::Default);

			SetLightBufferQuality(100);
		}
	}
}
