﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <iostream>
# include "SivColor.hpp"
# include "SivPointVector.hpp"

namespace s3d
{
	/// <summary>
	/// HSV カラー
	/// </summary>
	/// <remarks>
	/// 色を 色相(H), 彩度(S), 明度(V) で表します。
	/// </remarks>
	struct HSV
	{
		/// <summary>
		/// 色相 [0.0, 360.0)
		/// </summary>
		double h;

		/// <summary>
		/// 彩度 [0.0, 1.0]
		/// </summary>
		double s;
			
		/// <summary>
		/// 明度 [0.0, 1.0]
		/// </summary>
		double v;

		/**
		*
		*/
		HSV() {}

		/**
		*
		*	@param _h
		*/
		explicit HSV(double _h)
			: h(_h)
			, s(1.0)
			, v(1.0) {}

		/**
		*
		*	@param _h
		*	@param _s
		*	@param _v
		*/
		HSV(double _h, double _s, double _v)
			: h(_h)
			, s(_s)
			, v(_v) {}

		/**
		*
		*	@param col
		*/
		HSV(const Color& col);

		/**
		*
		*	@param col
		*/
		HSV(const ColorF& col);

		/**
		*
		*	@param hsv
		*	@return
		*/
		HSV operator + (const HSV& hsv) const;

		/**
		*
		*	@param hsv
		*	@return
		*/
		HSV operator - (const HSV& hsv) const;

		/**
		*
		*	@return
		*/
		operator Color() const
		{
			return toColor();
		}

		/// <summary>
		/// Color 型に変換します。
		/// </summary>
		/// <param name="alpha">
		/// アルファ値 [0, 255]
		/// </param>
		/// <returns>
		/// Color 型の色
		/// </returns>
		Color toColor(unsigned alpha = 255) const;

		/// <summary>
		/// ColorF 型に変換します。
		/// </summary>
		/// <param name="alpha">
		/// アルファ値 [0.0, 1.0]
		/// </param>
		/// <returns>
		/// ColorF 型の色
		/// </returns>
		ColorF toColorF(double alpha = 1.0) const;

		/**
		*
		*	@param r
		*	@param g
		*	@param b
		*/
		void convertFrom(double r, double g, double b);

		/**
		*
		*/
		Property_Get(Vec3, hsv) const
		{
			return{ h, s, v };
		}
	};

	/**
	*
	*	@param os
	*	@param hsv
	*	@return
	*/
	template<typename CharType>
	inline
	std::basic_ostream<CharType>&
		operator << (
		std::basic_ostream<CharType>& os,
		const HSV& hsv
		)
	{
		return os << CharType('(')
			<< hsv.h << CharType(',')
			<< hsv.s << CharType(',')
			<< hsv.v << CharType(')');
	}

	/**
	*
	*	@param is
	*	@param hsv
	*	@return
	*/
	template<typename CharType>
	inline
	std::basic_istream<CharType>&
		operator >> (
		std::basic_istream<CharType>& is,
		HSV& hsv
		)
	{
		CharType unused;

		return	is >> unused
			>> hsv.h >> unused
			>> hsv.s >> unused
			>> hsv.v >> unused;
	}

	/**
	*
	*	@param color
	*	@param hsv
	*	@return
	*/
	inline HSV operator + (const Color& color, const HSV& hsv)
	{
		return HSV(color) + hsv;
	}

	/**
	*
	*	@param color
	*	@param hsv
	*	@return
	*/
	inline HSV operator - (const Color& color, const HSV& hsv)
	{
		return HSV(color) - hsv;
	}

	/// <summary>
	/// 色相から Color を作成します。
	/// </summary>
	/// <param name="hue">
	/// 色相 [0.0, 360.0)
	/// </param>
	/// <returns>
	/// 作成したカラー
	/// </returns>
	Color HueToColor(double hue);

	/// <summary>
	/// 色相から ColorF を作成します。
	/// </summary>
	/// <param name="hue">
	/// 色相 [0.0, 360.0)
	/// </param>
	/// <returns>
	/// 作成したカラー
	/// </returns>
	ColorF HueToColorF(double hue);
}
