﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <unordered_map>
# include "SivString.hpp"
# include "SivTextWriter.hpp"

namespace s3d
{
	/// <summary>
	/// INI データの書き出し
	/// </summary>
	class INIWriter
	{
	private:

		std::unordered_map<String, String> m_globalSection;

		std::unordered_map<String, std::unordered_map<String, String>> m_sections;

		TextWriter m_writer;

		void flush()
		{
			for (const auto& name : m_globalSection)
			{
				m_writer.write(name.first);
				m_writer.write(L" = ");
				m_writer.write(name.second);
				m_writer.write(L"\r\n");
			}

			m_writer.write(L"\r\n");

			for (const auto& section : m_sections)
			{
				m_writer.write(L'[');
				m_writer.write(section.first);
				m_writer.write(L"]\r\n");

				for (const auto& name : section.second)
				{
					m_writer.write(name.first);
					m_writer.write(L" = ");
					m_writer.write(name.second);
					m_writer.write(L"\r\n");
				}

				m_writer.write(L"\r\n");
			}
		}

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		INIWriter() {}

		/// <summary>
		/// 書き出し用のファイルを開きます。
		/// </summary>
		/// <param name="path">
		/// ファイルパス
		/// </param>
		INIWriter(const FilePath& path)
			: m_writer(path) {}

		/// <summary>
		/// デストラクタ
		/// </summary>
		~INIWriter()
		{
			close();
		}

		/// <summary>
		/// 書き出し用のファイルを開きます。
		/// </summary>
		/// <param name="path">
		/// ファイルパス
		/// </param>
		/// <returns>
		/// ファイルのオープンに成功した場合 true, それ以外の場合は false
		/// </returns>
		bool open(const FilePath& path)
		{
			return m_writer.open(path);
		}

		/// <summary>
		/// INI データを書き出し、ファイルをクローズします。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void close()
		{
			if (m_writer)
			{
				flush();
			}

			m_writer.close();
		}

		/// <summary>
		/// 書き出し用ファイルがオープンされているかを返します。
		/// </summary>
		/// <returns>
		/// ファイルがオープンされている場合 true, それ以外の場合は false
		/// </returns>
		bool isOpened() const
		{
			return m_writer.isOpened();
		}

		/// <summary>
		/// 書き出し用ファイルがオープンされているかを返します。
		/// </summary>
		/// <returns>
		/// ファイルがオープンされている場合 true, それ以外の場合は false
		/// </returns>
		explicit operator bool() const { return isOpened(); }

		/// <summary>
		/// 書き出したデータをすべて削除します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void clear()
		{
			m_globalSection.clear();

			m_sections.clear();
		}

		/// <summary>
		/// プロパティを書き込みます。
		/// </summary>
		/// <param name="name">
		/// 値の名前
		/// </param>
		/// <param name="value">
		/// 値
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		template<class Type> void write(const String& name, const Type& value)
		{
			m_globalSection[name] = Format(value);
		}

		/// <summary>
		/// プロパティを書き込みます。
		/// </summary>
		/// <param name="section">
		/// セクション
		/// </param>
		/// <param name="name">
		/// 値の名前
		/// </param>
		/// <param name="value">
		/// 値
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		template<class Type> void write(const String& section, const String& name, const Type& value)
		{
			m_sections[section][name] = Format(value);
		}

		/// <summary>
		/// オープンしているファイルのパスを示します。
		/// </summary>
		/// <remarks>
		/// クローズしている場合は空の文字列です。
		/// </remarks>
		Property_Get(FilePath, path) const
		{
			return m_writer.path;
		}
	};
}
