﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivPropertyMacro.hpp"
# include "SivString.hpp"

namespace s3d
{
	/// <summary>
	/// 整数
	/// </summary>
	/// <remarks>
	///	整数のユーティリティ
	///	</remarks>
	template <typename Type>
	struct IntegerBase
	{
		/// <summary>
		///	整数
		///	</summary>
		Type value;

		/// <summary>
		/// コンストラクタ
		/// </summary>
		IntegerBase() {}

		/// <summary>
		/// 数値を設定します。
		/// </summary>
		/// <param name="_value">
		///	コピーする整数
		///	</param>
		IntegerBase(Type _value)
			: value(_value) {}

		/// <summary>
		/// 文字列から数値に変換します。
		/// </summary>
		/// <param name="str">
		///	変換される文字列（例: L"12345", L"0xFF")
		///	</param>
		IntegerBase(StringView str)
			: IntegerBase(str, 0) {}

		/// <summary>
		/// 基数を指定して文字列から数値に変換します。
		/// </summary>
		/// <param name="str">
		///	変換される文字列（例: L"12345", L"0xFF")
		///	</param>
		/// <param name="radix">
		///	使用する基数
		///	</param>
		IntegerBase(StringView str, int radix);

		/// <summary>
		/// 組み込みの整数型に変換します。
		/// </summary>
		/// <returns>
		///	組み込みの整数型の値
		///	</returns>
		operator Type() const { return value; }

		/// <summary>
		/// 新しい整数を代入します。
		/// </summary>
		/// <param name="_value">
		///	新しい整数
		///	</param>
		/// <returns>
		///	新しい整数
		///	</returns>
		IntegerBase operator = (Type _value)
		{
			value = _value;
			return *this;
		}

		/// <summary>
		/// 文字列で示された新しい整数を代入します。
		/// </summary>
		/// <param name="str">
		///	新しい整数
		///	</param>
		/// <returns>
		///	新しい整数
		///	</returns>
		IntegerBase operator = (StringView str)
		{
			return *this = IntegerBase<Type>(str);
		}

		/// <summary>
		/// 整数を指定した基数の文字列で表します。
		/// </summary>
		/// <param name="radix">
		///	基数
		///	</param>
		/// <remarks>
		///	基数は [2-36] の範囲で指定します。
		///	</remarks>
		/// <returns>
		///	変換した文字列
		///	</returns>
		String toString(int radix) const;

		/// <summary>
		/// 整数を 10 進数の文字列で表します。
		/// </summary>
		Property_Get(String, string) const { return toString(10); }

		/// <summary>
		/// 整数を 2 進数の文字列で表します。
		/// </summary>
		Property_Get(String, binary) const { return toString(2); }

		/// <summary>
		/// 整数を 8 進数の文字列で表します。
		/// </summary>
		Property_Get(String, octal) const { return toString(8); }

		/// <summary>
		/// 整数を 16 進数の文字列で表します。
		/// </summary>
		Property_Get(String, hex) const { return toString(16); }

		/// <summary>
		/// 表現できる最大の整数を返します
		/// </summary>
		/// <returns>
		///	表現できる最大の整数
		///	</returns>
		static Type Max()
		{
			return std::numeric_limits<Type>::max();
		}

		/// <summary>
		/// 表現できる最小の整数を返します。
		/// </summary>
		/// <returns>
		///	表現できる最小の整数
		///	</returns>
		static Type Min()
		{
			return std::numeric_limits<Type>::min();
		}
	};

	/// <summary>
	/// 符号付き 8 ビット整数型
	/// </summary>
	using Integer8 = IntegerBase<char>;

	/// <summary>
	/// 符号無し 8 ビット整数型
	/// </summary>
	using IntegerU8 = IntegerBase<unsigned __int8>;

	/// <summary>
	/// 符号付き 16 ビット整数型
	/// </summary>
	using Integer16 = IntegerBase<__int16>;

	/// <summary>
	/// 符号無し 16 ビット整数型
	/// </summary>
	using IntegerU16 = IntegerBase<unsigned __int16>;

	/// <summary>
	/// 符号付き 32 ビット整数型
	/// </summary>
	using Integer32 = IntegerBase<__int32>;

	/// <summary>
	/// 符号付き 32 ビット整数型
	/// </summary>
	using Integer = Integer32;

	/// <summary>
	/// 符号無し 32 ビット整数型
	/// </summary>
	using IntegerU32 = IntegerBase<unsigned __int32>;

	/// <summary>
	/// 符号付き 64 ビット整数型
	/// </summary>
	using Integer64 = IntegerBase<__int64>;

	/// <summary>
	/// 符号無し 64 ビット整数型
	/// </summary>
	using IntegerU64 = IntegerBase<unsigned __int64>;
}
