﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivOptional.hpp"
# include "SivString.hpp"
# include "SivArray.hpp"

namespace s3d
{
	/// <summary>
	/// ダウンロードの進行状況
	/// </summary>
	enum class DownloadStatus
	{
		/// <summary>
		/// 無効な状態
		/// </summary>
		Invalid,

		/// <summary>
		/// ダウンロード中
		/// </summary>
		Running,
		
		/// <summary>
		/// 一時停止中
		/// </summary>
		Paused,
		
		/// <summary>
		/// ダウンロード完了
		/// </summary>
		Succeeded,
		
		/// <summary>
		/// ダウンロード失敗
		/// </summary>
		Failed,
		
		/// <summary>
		/// ダウンロードがキャンセルされた
		/// </summary>
		Canceled,
	};

	/// <summary>
	/// ファイルのダウンロード状況
	/// </summary>
	struct FileDownloadState
	{
		/// <summary>
		/// ダウンロード済みのサイズ（バイト）
		/// </summary>
		long long downloadedSize = 0;

		/// <summary>
		/// ファイルのサイズの Optional. 不明の場合 none
		/// </summary>
		Optional<long long> fileSize;

		/// <summary>
		/// ダウンロードの進行状況
		/// </summary>
		DownloadStatus status = DownloadStatus::Invalid;

		/// <summary>
		/// ファイルの URL
		/// </summary>
		FilePath uri;

		/// <summary>
		/// ファイルの保存パス
		/// </summary>
		FilePath savePath;
	};

	/// <summary>
	/// インターネット
	/// </summary>
	/// <remarks>
	/// インターネット上のデータにアクセスする機能を提供します。
	/// </remarks>
	namespace Internet
	{
		/// <summary>
		/// コンピューターがインターネットに接続されているかを調べます。
		/// </summary>
		/// <returns>
		/// コンピューターがインターネットに接続されていれば true, されていなければ false
		/// </returns>
		bool IsConnected();

		/// <summary>
		/// 指定した URL のファイルをダウンロードします。
		/// </summary>
		/// <param name="url">
		/// ダウンロードするファイルの URL
		/// </param>
		/// <param name="savePath">
		/// ダウンロードしたファイルの保存パス
		/// </param>
		/// <remarks>
		/// ダウンロードが完了するか失敗するまでこの関数は制御を返しません。
		/// </remarks>
		/// <returns>
		/// ダウンロードが完了したら true, それ以外の場合は false
		/// </returns>
		bool Download(const FilePath& url, const FilePath& savePath);

		/// <summary>
		/// 指定した URL のファイルの非同期ダウンロードを開始します。
		/// </summary>
		/// <param name="url">
		/// ダウンロードするファイルの URL
		/// </param>
		/// <param name="savePath">
		/// ダウンロードしたファイルの保存パス
		/// </param>
		/// <returns>
		/// ダウンロード開始に成功した場合 true, 失敗した場合 false
		/// </returns>
		bool DownloadAsync(const FilePath& url, const FilePath& savePath);

		/// <summary>
		/// ダウンロードの進行状況を取得します。
		/// </summary>
		/// <returns>
		/// ダウンロードの進行状況の Optional, ダウンロードを実行中でなければ none
		/// </returns>
		Optional<FileDownloadState> GetFileDownloadState();

		/// <summary>
		/// ダウンロードを中止します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void CancelDownload();

		/// <summary>
		/// 指定した URL を Internet Explorer で開きます。
		/// </summary>
		/// <param name="url">
		/// http:// もしくは https:// で始まる URL
		/// </param>
		/// <returns>
		/// 成功した場合 true, 失敗した場合 false
		/// </returns>
		bool LaunchIE(const FilePath& url);

		/// <summary>
		/// 指定した URL をユーザの既定のブラウザで開きます。
		/// </summary>
		/// <param name="url">
		/// http:// もしくは https:// で始まる URL
		/// </param>
		/// <returns>
		/// 成功した場合 true, 失敗した場合 false
		/// </returns>
		bool LaunchWebBrowser(const FilePath& url);
	}
}

