﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivPropertyMacro.hpp"
# include "SivArray.hpp"

namespace s3d
{
	/// <summary>
	/// キー
	/// </summary>
	/// <remarks>
	/// キーボードの特定のキーを示します。
	/// </remarks>
	class Key
	{
	private:

		unsigned char m_virtualKeyCode = 0;

		friend struct Input;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		Key() {}

		/// <summary>
		/// アルファベットか数字でキーを指定します。
		/// </summary>
		/// <param name="alnum">
		/// アルファベットか数字
		/// </param>
		explicit Key(wchar_t alnum);

		/// <summary>
		/// キーがクリックされたことを示します。
		/// </summary>
		Property_Get(bool, clicked) const;

		/// <summary>
		/// キーが押されていることを示します。
		/// </summary>
		Property_Get(bool, pressed) const;

		/// <summary>
		/// キーが離されたことを示します。
		/// </summary>
		Property_Get(bool, released) const;

		/// <summary>
		/// 仮想キーコードを示します。
		/// </summary>
		Property_Get(unsigned char, code) const
		{
			return m_virtualKeyCode;
		}
	};

	/// <summary>
	/// 2 つのキーが同じキーを示しているかを返します。
	/// </summary>
	/// <param name="key1">
	/// 比較するキー
	/// </param>
	/// <param name="key2">
	/// 比較するキー
	/// </param>
	/// <returns>
	/// 2 つのキーが同じキーを示している場合 true, それ以外の場合は false
	/// </returns>
	inline bool operator == (const Key& key1, const Key& key2)
	{
		return key1.code == key2.code;
	}

	/// <summary>
	/// 2 つのキーが異なるキーを示しているかを返します。
	/// </summary>
	/// <param name="key1">
	/// 比較するキー
	/// </param>
	/// <param name="key2">
	/// 比較するキー
	/// </param>
	/// <returns>
	/// 2 つのキーが異なるキーを示している場合 true, それ以外の場合は false
	/// </returns>
	inline bool operator != (const Key& key1, const Key& key2)
	{
		return key1.code != key2.code;
	}

	/// <summary>
	/// キーの組み合わせ（プラス）
	/// </summary>
	/// <remarks>
	/// 1 つ目のキーが押されている状態のとき 2 つ目のキー入力を調べます。
	/// </remarks>
	class PlusCombinedKeys
	{
	private:

		Key m_key1, m_key2;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		PlusCombinedKeys() {}

		/// <summary>
		/// キーの組み合わせを設定します。
		/// </summary>
		/// <param name="key1">
		/// 1 つ目のキー
		/// </param>
		/// <param name="key2">
		/// 2 つ目のキー
		/// </param>
		PlusCombinedKeys(Key key1, Key key2)
			: m_key1(key1)
			, m_key2(key2) {}

		/// <summary>
		/// 1 つ目のキーが押されているときに、2 つ目のキーがクリックされたことを示します。
		/// </summary>
		Property_Get(bool, clicked) const
		{
			return m_key1.pressed && m_key2.clicked;
		}

		/// <summary>
		/// 1 つ目のキーが押されているときに、2 つ目のキーが押されていることを示します。
		/// </summary>
		Property_Get(bool, pressed) const
		{
			return m_key1.pressed && m_key2.pressed;
		}

		/// <summary>
		/// 1 つ目のキーが押されているときに、2 つ目のキーが離されたことを示します。
		/// </summary>
		Property_Get(bool, released) const
		{
			return m_key1.pressed && m_key2.released;
		}
	};

	/// <summary>
	/// キーの組み合わせ（OR）
	/// </summary>
	/// <remarks>
	/// 1 つ目のキーか 2 つ目のキーの入力を調べます。
	/// </remarks>
	class OrCombinedKeys
	{
	private:

		Key m_key1, m_key2;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		OrCombinedKeys() {}

		/// <summary>
		/// キーの組み合わせを設定します。
		/// </summary>
		/// <param name="key1">
		/// 1 つ目のキー
		/// </param>
		/// <param name="key2">
		/// 2 つ目のキー
		/// </param>
		OrCombinedKeys(Key key1, Key key2)
			: m_key1(key1)
			, m_key2(key2) {}

		/// <summary>
		/// どちらかのキーがクリックされたことを示します。
		/// </summary>
		Property_Get(bool, clicked) const
		{
			return m_key1.clicked || m_key2.clicked;
		}

		/// <summary>
		/// どちらかのキーがクリック押されていることを示します。
		/// </summary>
		Property_Get(bool, pressed) const
		{
			return m_key1.pressed || m_key2.pressed;
		}

		/// <summary>
		/// どちらかのキーが離されたことを示します。
		/// </summary>
		Property_Get(bool, released) const
		{
			return m_key1.released || m_key2.released;
		}
	};

	/// <summary>
	/// キーの組み合わせ（プラス）を設定します。
	/// </summary>
	/// <param name="key1">
	/// 1 つ目のキー
	/// </param>
	/// <param name="key2">
	/// 2 つ目のキー
	/// </param>
	/// <returns>
	/// キーの組み合わせ（プラス）
	/// </returns>
	inline PlusCombinedKeys operator + (Key key1, Key key2)
	{
		return{ key1, key2 };
	}

	/// <summary>
	/// キーの組み合わせ (OR) を設定します。
	/// </summary>
	/// <param name="key1">
	/// 1 つ目のキー
	/// </param>
	/// <param name="key2">
	/// 2 つ目のキー
	/// </param>
	/// <returns>
	/// キーの組み合わせ (OR)
	/// </returns>
	inline OrCombinedKeys operator | (Key key1, Key key2)
	{
		return{ key1, key2 };
	}
}
