﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <array>
# include "Siv3DFwd.hpp"
# include "SivColor.hpp"
# include "SivHSV.hpp"
# include "SivKinectV1.hpp"

namespace s3d
{
	/// <summary>
	/// Kinect v2 の取得データフラグ
	/// </summary>
	struct KinectV2DataType
	{
		enum Type
		{
			Color		= 0x01,

			Depth		= 0x02,

			BodyIndex	= 0x04,

			Body		= 0x08,

			Infrared	= 0x10,

			AudioBeam	= 0x20,

			Default		= Color | Depth | BodyIndex | Body,
		};
	};

	/// <summary>
	/// 手の状態
	/// </summary>
	enum class HandState
	{
		Unknown,

		NotTracked,

		Open,

		Closed,

		Lasso,
	};

	struct V2JointType
	{
		enum Type
		{
			SpineBase,
			SpineMid,
			Neck,
			Head,
			ShoulderLeft,
			ElbowLeft,
			WristLeft,
			HandLeft,
			ShoulderRight,
			ElbowRight,
			WristRight,
			HandRight,
			HipLeft,
			KneeLeft,
			AnkleLeft,
			FootLeft,
			HipRight,
			KneeRight,
			AnkleRight,
			FootRight,
			SpineShoulder,
			HandTipLeft,
			ThumbLeft,
			HandTipRight,
			ThumbRight,
		};
	};

	struct KinectV2Joint
	{
		Vec3 cameraSpacePos = Vec3::Zero;

		Vec2 colorSpacePos = Vec2::Zero;

		Vec2 depthSpacePos = Vec2::Zero;

		TrackingState trackingState = TrackingState::NotTracked;
	};

	struct KinectV2Body
	{
		enum { Num_Joints = 25 };

		KinectV2Joint joints[Num_Joints];

		Vec2 lean = Vec2::Zero;

		HandState leftHand = HandState::Unknown;

		HandState rightHand = HandState::Unknown;

		bool leftHandHighConfidence = false;

		bool rightHandHighConfidence = false;
	};

	struct KinectV2AudioBeam
	{
		double angle = 0.0;

		double confidence = 0.0;
	};

	/// <summary>
	/// Kinect v2
	/// </summary>
	/// <remarks>
	/// Kinect v2 の機能を提供します。
	/// </remarks>
	namespace KinectV2
	{
		inline Color DefaultDepthColoring(unsigned short depth)
		{
			return depth != 0 ? HueToColor(depth / 4.0) : Palette::Black;
		}

		inline Color DefaultBodyIndexColoring(unsigned char index)
		{
			return index != 255 ? HueToColor(index * 60) : Palette::Black;
		}

		inline Color DefaultInfraredColoring(unsigned short infrared)
		{
			return Color(infrared / 256);
		}

		bool IsAvailable();

		bool Start(int dataType = KinectV2DataType::Default);


		bool StartColorFrame();

		void StopColorFrame();

		bool HasNewColorFrame();

		bool GetColorFrame(Image& image);

		bool GetColorFrame(DynamicTexture& texture);


		bool StartDepthFrame();

		void StopDepthFrame();

		bool HasNewDepthFrame();

		const Grid<unsigned short>& GetRawDepthData();

		bool GetDepthFrame(Image& image, std::function<Color(unsigned short)> coloring = DefaultDepthColoring);
		
		bool GetDepthFrame(DynamicTexture& texture, std::function<Color(unsigned short)> coloring = DefaultDepthColoring);

	
		bool StartBodyIndexFrame();

		void StopBodyIndexFrame();

		bool HasNewBodyIndexFrame();

		const Grid<unsigned char>& GetRawBodyIndexData();

		bool GetBodyIndexFrame(Image& image, std::function<Color(unsigned char)> coloring = DefaultBodyIndexColoring);

		bool GetBodyIndexFrame(DynamicTexture& texture, std::function<Color(unsigned char)> coloring = DefaultBodyIndexColoring);


	
		bool StarBodyFrame();

		void StopBodyFrame();

		bool HasNewBodyFrame();

		bool GetBodyFrame(std::array<Optional<KinectV2Body>, 6>& bodies);


		bool StartInfraredFrame();

		void StopInfraredFrame();

		bool HasNewInfraredFrame();

		const Grid<unsigned short>& GetRawInfraredData();

		bool GetInfraredFrame(Image& image, std::function<Color(unsigned short)> coloring = DefaultInfraredColoring);

		bool GetInfraredFrame(DynamicTexture& texture, std::function<Color(unsigned short)> coloring = DefaultInfraredColoring);


		bool StarAudioBeamFrame();

		void StopAudioBeamFrame();

		bool HasNewAudioBeamFrame();

		bool GetAudioBeamFrame(KinectV2AudioBeam& audioBeam);
	}
}
