﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "SivArray.hpp"
# include "SivUtility.hpp"
# include "SivPointVector.hpp"
# include "SivLine.hpp"

namespace s3d
{
	/**
	*	@brief 線の配列
	*/
	class LineString
	{
	private:

		Array<Vec2> m_data;

	public:

		/**
		*
		*/
		LineString() {}

		/**
		*
		*	@param pts
		*/
		explicit LineString(const String& pts)
		{
			std::wistringstream(pts.str()) >> *this;
		}

		/**
		*
		*	@param pts
		*/
		explicit LineString(const Array<Point>& pts)
		{
			m_data.reserve(pts.size());

			for (const auto& p : pts)
			{
				m_data.push_back(p);
			}
		}

		/**
		*
		*	@param pts
		*/
		explicit LineString(const Array<Vec2>& pts)
			: m_data(pts) {}

		/**
		*
		*	@param pts
		*/
		explicit LineString(Array<Vec2>&& pts)
			: m_data(std::move(pts)) {}

		/**
		*
		*	@param pts
		*	@param size
		*/
		LineString(const Vec2* pts, size_t size)
			: m_data(pts, pts+size) {}

		/**
		*
		*	@param pts
		*/
		template<typename Iterator>
		LineString(Iterator first, Iterator last)
			: m_data(first, last) {}

		/**
		*
		*	@param il
		*/
		explicit LineString(std::initializer_list<Vec2> il)
			: LineString(il.begin(), il.size()) {}

		LineString movedBy(double x, double y) const { return movedBy({ x, y }); }

		LineString movedBy(const Vec2& v) const;

		LineString& moveBy(double x, double y) { return moveBy({ x, y }); }

		LineString& moveBy(const Vec2& v);

		/**
		*
		*	@return
		*/
		Property_Get(bool, isEmpty) const { return m_data.empty(); }

		/**
		*
		*/
		Property_Get(size_t, num_points) const { return m_data.size(); }

		Property_Get(size_t, num_lines) const { return isEmpty ? 0 : m_data.size() - 1; }

		/**
		*
		*	@param index
		*	@return
		*/
		size_t size() const { return m_data.size(); }

		/**
		*
		*	@param index
		*	@return
		*/
		Vec2& operator [] (size_t index) { return m_data[index]; }

		/**
		*
		*	@param index
		*	@return
		*/
		const Vec2& operator [] (size_t index) const { return m_data[index]; }

		/**
		*
		*	@return
		*/
		Array<Vec2>::iterator begin() { return m_data.begin(); }

		/**
		*
		*	@return
		*/
		Array<Vec2>::iterator end() { return m_data.end(); }

		/**
		*
		*	@return
		*/
		Array<Vec2>::const_iterator begin() const { return m_data.begin(); }

		/**
		*
		*	@return
		*/
		Array<Vec2>::const_iterator end() const { return m_data.end(); }

		/**
		*
		*	@return
		*/
		Array<Vec2>& data() { return m_data; }

		/**
		*
		*	@return
		*/
		const Array<Vec2>& data() const { return m_data; }

		/**
		*
		*	@return
		*/
		void push_back(const Vec2& point) { m_data.push_back(point); }

		/**
		*
		*	@return
		*/
		void pop_back() { m_data.pop_back(); }

		/**
		*
		*	@return
		*/
		void clear() { m_data.clear(); }

		/**
		*
		*	@return
		*/
		void reserve(size_t size) { m_data.reserve(size); }

		/**
		*
		*	@return
		*/
		void shrink_to_fit() { m_data.shrink_to_fit(); }

		Line line(unsigned index) const
		{
			return{ m_data[index], m_data[index + 1] };
		}

		/**
		*
		*	@param offset
		*	@return
		*/
		LineString
			substr(
			size_t offset
			) const
		{
			return LineString(begin() + offset, end());
		}

		/**
		*
		*	@param offset
		*	@param count
		*	@return
		*/
		LineString
			substr(
			size_t offset,
			size_t count
			) const
		{
			return LineString(begin() + offset, begin() + offset + count);
		}

		LineString
			toCatmullRomSpline(
			bool isClosedCurve = false,
			int interpolation = 24
			) const;

		/**
		*
		*	@param shape
		*	@return
		*/
		template<typename Shape>
		bool intersects(const Shape& shape) const
		{
			return Geometry2D::Intersect(*this, shape);
		}

		/**
		*
		*	@param image
		*	@param color
		*/
		void write(Image& image, const Color& color = Palette::Black, bool isClosed = false) const;

		/**
		*
		*	@param image
		*	@param thickness
		*	@param color
		*/
		void write(Image& image, int thickness, const Color& color = Palette::Black, bool isClosed = false) const;

		/**
		*
		*	@param image
		*	@param color
		*/
		void overwrite(Image& image, const Color& color = Palette::Black, bool isClosed = false) const;

		/**
		*
		*	@param image
		*	@param thickness
		*	@param color
		*/
		void overwrite(Image& image, int thickness, const Color& color = Palette::Black, bool isClosed = false) const;

		/**
		*
		*	@param image
		*	@param thickness
		*	@param color
		*	@param isClosed
		*/
		void
			writeCatmullRomSpline(
			Image& image,
			int thickness = 1,
			const Color& color = Palette::Black,
			bool isClosed= false,
			int interpolation = 24
			) const;

		void
			overwriteCatmullRomSpline(
			Image& image,
			int thickness = 1,
			const Color& color = Palette::Black,
			bool isClosed = false,
			int interpolation = 24
			) const;

		/**
		*
		*	@param color
		*	@param isClosedCurve
		*/
		void draw(const Color& color = Palette::White, bool isClosedCurve = false) const;

		/**
		*
		*	@param thickness
		*	@param color
		*	@param isClosedCurve
		*/
		void draw(double thickness, const Color& color = Palette::White, bool isClosedCurve = false) const;

		/**
		*
		*	@param thickness
		*	@param color
		*	@param isClosedCurve
		*/
		void
			drawCatmullRomSpline(
			double thickness = 1.0,
			const Color& color = Palette::White,
			bool isClosedCurve = false,
			int interpolation = 24
			) const;

		Polygon
			asPolygon(
			double thickness = 1.0,
			bool isClosed = false
			) const;
	};

	/**
	*
	*	@param os
	*	@param lineStr
	*	@return
	*/
	template<typename CharType>
	inline
	std::basic_ostream<CharType>&
		operator << (
		std::basic_ostream<CharType>& os,
		const LineString& lineStr
		)
	{
		os << CharType('(');

		bool b = false;

		for (const auto& point : lineStr)
		{
			if (Exchange(b, true))
			{
				os << CharType(',');
			}

			os << point;
		}

		return os << CharType(')');
	}

	/**
	*
	*	@param is
	*	@param lineStr
	*	@return
	*/
	template<typename CharType>
	inline
	std::basic_istream<CharType>&
		operator >> (
		std::basic_istream<CharType>& is,
		LineString& lineStr
		)
	{
		lineStr.clear();

		CharType unused;

		is >> unused;

		for (;;)
		{
			Vec2 v;

			is >> v;

			if (!is)
			{
				break;
			}

			is >> unused;

			lineStr.push_back(v);

			if (unused == L')')
			{
				break;
			}
		}

		return is;
	}
}
