﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <memory>
# include "Siv3DFwd.hpp"
# include "SivIWriter.hpp"

namespace s3d
{
	/// <summary>
	/// 書き込み用バッファ
	/// </summary>
	class MemoryWriter : public IWriter
	{
	private:

		class CMemoryWriter;

		std::shared_ptr<CMemoryWriter> pWriter;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		MemoryWriter();

		/// <summary>
		/// デストラクタ
		/// </summary>
		~MemoryWriter() = default;

		/// <summary>
		/// 書き込みバッファを解放します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void release();

		/// <summary>
		/// 書き込み用バッファが使用可能かを示します。
		/// </summary>
		/// <returns>
		/// つねに true
		/// </returns>
		bool isOpened() const override { return true; }

		/// <summary>
		/// 書き込み用バッファが使用可能かを示します。
		/// </summary>
		/// <returns>
		/// つねに true
		/// </returns>
		explicit operator bool() const { return isOpened(); }

		/// <summary>
		/// 書き込み用バッファのサイズを返します。
		/// </summary>
		/// <returns>
		/// 書き込み用バッファのサイズ（バイト）
		/// </returns>
		long long size() const override;

		/// <summary>
		/// 現在の書き込み位置を返します。
		/// </summary>
		/// <returns>
		/// 現在の書き込み位置（バイト）
		/// </returns>
		long long getPos() const override;

		/// <summary>
		/// 書き込み位置を変更します。
		/// </summary>
		/// <param name="pos">
		/// 新しい書き込み位置（バイト）
		/// </param>
		/// <returns>
		/// 書き込み位置の変更に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool setPos(long long pos) override;

		/// <summary>
		/// バッファにデータを書き込みます。
		/// </summary>
		/// <param name="buffer">
		/// 書き込むデータ
		/// </param>
		/// <param name="size">
		/// 書き込むサイズ（バイト）
		/// </param>
		/// <returns>
		/// 実際に書き込んだサイズ（バイト）
		/// </returns>
		unsigned write(const void* buffer, unsigned size) override;

		/// <summary>
		/// 書き込みバッファの先頭ポインタを返します。
		/// </summary>
		/// <remarks>
		/// 現在の書き込み位置に関係なく、バッファの先頭のポインタを返します。
		/// </remarks>
		/// <returns>
		/// 書き込みバッファの先頭ポインタ
		/// </returns>
		const unsigned char* data() const;

		/// <summary>
		/// バッファの内容をファイルに保存します。
		/// </summary>
		/// <param name="path">
		/// ファイルパス
		/// </param>
		/// <remarks>
		/// 現在の書き込み位置に関係なく、バッファの内容全てを保存します。
		/// </remarks>
		/// <returns>
		/// 保存に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool save(const FilePath& path) const;

		/// <summary>
		/// バッファの内容を MemoryReader にコピーして返します。
		/// </summary>
		/// <returns>
		/// バッファの内容をコピーした MemoryReader
		/// </returns>
		MemoryReader toMemoryReader();
	};
}
