﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivPointVector.hpp"
# include "SivBox.hpp"

namespace s3d
{
	struct MeshVertex
	{
		Float3 position;

		Float3 normal;

		Float2 texcoord;
	};

	enum class PlaneFace
	{
		XZ,

		XY,

		YZ,
	};

	struct MeshData
	{
		Array<MeshVertex> vertices;

		Array<unsigned> indices;

		MeshData() {}

		MeshData(
			unsigned vertexSize,
			unsigned indexSize
			)
			: vertices(vertexSize)
			, indices(indexSize) {}

		MeshData(
			std::vector<MeshVertex> _vertices,
			std::vector<unsigned> _indices
			)
			: vertices(std::move(_vertices))
			, indices(std::move(_indices)) {}

		/// <summary>
		/// 法線を生成します。
		/// </summary>
		/// <returns>
		/// 法線の生成に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool computeNormals();

		/// <summary>
		/// バウンディングボックスを計算します。
		/// </summary>
		/// <returns>
		/// バウンディングボックス
		/// </returns>
		Box computeBoundingBox() const;

		static MeshData
			Plane(
			double size,
			const Vec2& uvScale = { 1, 1 },
			const Vec2& uvOffset = { 0, 0 },
			PlaneFace face = PlaneFace::XZ
			);

		static MeshData
			Plane(
			const Vec2& size = { 1, 1 },
			const Vec2& uvScale = { 1, 1 },
			const Vec2& uvOffset = { 0, 0 },
			PlaneFace face = PlaneFace::XZ
			);

		static MeshData
			PlaneXZ(
			double size,
			const Vec2& uvScale = { 1, 1 },
			const Vec2& uvOffset = { 0, 0 }
			);

		static MeshData
			PlaneXZ(
			const Vec2& size = { 1, 1 },
			const Vec2& uvScale = { 1, 1 },
			const Vec2& uvOffset = { 0, 0 }
			);

		static MeshData
			PlaneXY(
			double size,
			const Vec2& uvScale = { 1, 1 },
			const Vec2& uvOffset = { 0, 0 }
		);

		static MeshData
			PlaneXY(
			const Vec2& size = { 1, 1 },
			const Vec2& uvScale = { 1, 1 },
			const Vec2& uvOffset = { 0, 0 }
			);

		static MeshData
			PlaneYZ(
			double size,
			const Vec2& uvScale = { 1, 1 },
			const Vec2& uvOffset = { 0, 0 }
		);

		static MeshData
			PlaneYZ(
			const Vec2& size = { 1, 1 },
			const Vec2& uvScale = { 1, 1 },
			const Vec2& uvOffset = { 0, 0 }
			);

		static MeshData
			Box(
			double size,
			bool flipNormals = false
			);

		static MeshData
			Box(
			const Vec3& size = { 1, 1, 1 },
			bool flipNormals = false
			);

		static MeshData
			Sphere(
			double r = 1.0,
			unsigned quality = 24,
			bool flipNormals = false
			);

		static MeshData
			Disc(
			double r = 1.0,
			PlaneFace face = PlaneFace::XZ,
			unsigned quality = 63
			);

		static MeshData
			Cylinder(
			double r = 1.0,
			double h = 1.0
			);

		static MeshData
			Cone(
			double r = 1.0,
			double h = 1.0
			);

		static MeshData
			Polygon(
			const s3d::Polygon& polygon,
			const Vec2& uvScale,
			const Vec2& uvOffset
			);

		/// <summary>
		/// Polygon に厚みを持たせて MeshData を作成します。
		/// </summary>
		/// <param name="polygon">
		/// Polygon
		/// </param>
		/// <param name="height">
		/// 厚み
		/// </param>
		/// <param name="uvScale">
		/// UV マップスケール
		/// </param>
		/// <param name="uvOffset">
		/// UV マップオフセット
		/// </param>
		/// <param name="uvY">
		/// 厚み部分の UV マップパラメータ
		/// </param>
		/// <returns>
		/// 作成された MeshData
		/// </returns>
		static MeshData Polygon(const s3d::Polygon& polygon, double height, const Vec2& uvScale, const Vec2& uvOffset, const Vec2& uvY = { 1.0, 1.0 });
	};
}
