﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"

namespace s3d
{
	/// <summary>
	/// MIDI
	/// </summary>
	/// <remarks>
	/// MIDI サウンドを再生する機能を提供します。
	/// </remarks>
	namespace Midi
	{
		/// <summary>
		/// 新しい MIDI ファイルを読み込みます。
		/// </summary>
		/// <param name="path">
		/// MIDI ファイル名
		/// </param>
		/// <remarks>
		/// 以前に読み込まれていた MIDI は停止・破棄され、音量・再生速度はリセットされます。
		/// </remarks>
		/// <returns>
		/// MIDI ファイルの読み込みに成功した場合 true, それ以外の場合は false
		/// </returns>
		bool Open(const FilePath& path);

		/// <summary>
		/// Midi::Open() で読み込んだ MIDI を再生します。
		/// </summary>
		/// <remarks>
		/// MIDI は同時に 1 つのファイルしか再生できません。
		/// ほかのアプリケーションが MIDI デバイスを使用していた場合、再生に失敗する場合があります。
		/// </remarks>
		/// <returns>
		/// MIDI の再生開始に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool Play();

		/// <summary>
		/// Midi::Open() で読み込んだ MIDI をループ再生します。
		/// </summary>
		/// <remarks>
		/// MIDI は同時に 1 つのファイルしか再生できません。
		/// ほかのアプリケーションが MIDI デバイスを使用していた場合、再生に失敗する場合があります。
		/// </remarks>
		/// <returns>
		/// MIDI の再生開始に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool PlayLoop();

		/// <summary>
		/// 再生中の MIDI を一時停止します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void Pause();

		/// <summary>
		/// 再生中の MIDI を停止し、再生位置を曲の先頭に戻します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void Stop();

		/// <summary>
		/// MIDI が再生中かどうかを示します。
		/// </summary>
		/// <returns>
		/// MIDI が再生中の場合 true, それ以外の場合は false
		/// </returns>
		bool IsPlaying();

		/// <summary>
		/// MIDI の再生音量を指定します。
		/// </summary>
		/// <param name="volume">
		/// 音量 (0.0 が最小、1.0 が最大)
		/// </param>
		/// <remarks>
		/// volume は [0.1, 1.0] の範囲にクランプされます。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void SetVolume(double volume);

		/// <summary>
		/// 現在設定されている MIDI の再生音量を返します。
		/// </summary>
		/// <returns>
		/// 現在設定されている MIDI の再生音量 [0.0-1.0]
		/// </returns>
		double GetVolume();

		/// <summary>
		/// MIDI の再生速度を変更します。
		/// </summary>
		/// <param name="speed">
		/// 再生倍率 (0.1 が最小、10.0 が最大)
		/// </param>
		/// <remarks>
		/// speed は [0.1, 10.0] の範囲にクランプされます。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void SetTempo(double speed);

		/// <summary>
		/// 現在設定されている MIDI の再生速度を返します。
		/// </summary>
		/// <returns>
		/// 現在設定されている MIDI の再生速度
		/// </returns>
		double GetTempo();

		/// <summary>
		/// 現在読み込まれている MIDI を通常速度で再生した場合の曲の長さを秒で返します。
		/// </summary>
		/// <returns>
		/// MIDI の長さ (秒)
		/// </returns>
		double GetLengthSec();

		/// <summary>
		/// 現在読み込まれている MIDI の再生位置を秒で返します。
		/// </summary>
		/// <returns>
		/// MIDI の再生位置 (秒)
		/// </returns>
		double GetPosSec();

		/// <summary>
		/// MIDI の再生位置を変更します。
		/// </summary>
		/// <param name="pos">
		/// MIDI の再生位置 (秒)
		/// </param>
		/// <remarks>
		/// MIDI が再生中の場合は、そこから再生を再開します。
		/// </remarks>
		/// <returns>
		/// なし
		/// </returns>
		void SetPosSec(double pos);
	}
}
