﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <unordered_map>
# include "SivHandle.hpp"
# include "SivString.hpp"
# include "SivArray.hpp"
# include "SivMesh.hpp"

namespace s3d
{
	/// <summary>
	/// マテリアル
	/// </summary>
	struct Material
	{
		ColorF ambient;

		ColorF diffuse;

		ColorF specular;

		ColorF transmittance;

		ColorF emission;

		float shininess;

		float ior;
		
		float dissolve;

		int illum;

		String name;

		String ambientTextureName;

		String diffuseaTextureName;

		String specularTextureName;

		String normalTextureName;
	};

	/// <summary>
	/// モデルノード
	/// </summary>
	struct ModelNode
	{
		ModelNode() {}

		ModelNode(const String& _name, const Mesh& _mesh, const Material& _material)
			: name(_name)
			, mesh(_mesh)
			, material(_material)
		{

		}

		String name;

		Mesh mesh;

		Material material;
	};

	/// <summary>
	/// モデル
	/// </summary>
	class Model
	{
	protected:

		class Handle{};

		using ModelHandle = Siv3DHandle<Handle>;

		std::shared_ptr<ModelHandle> m_handle;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		Model();

		/// <summary>
		/// 3D モデルファイルからモデルをロードします。
		/// </summary>
		/// <param name="path">
		/// 3D モデルファイルのパス
		/// </param>
		/// <remarks>
		/// Wavefront OBJ 形式のみサポートしています。
		/// </remarks>
		explicit Model(const FilePath& path);

		/// <summary>
		/// デストラクタ
		/// </summary>
		virtual ~Model();

		/// <summary>
		/// モデルをリリースします。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void release();

		explicit operator bool() const { return !isEmpty; }

		Property_Get(HandleIDType, id) const;

		Property_Get(bool, isEmpty) const;

		bool operator == (const Model& model) const;

		bool operator != (const Model& model) const;

		/// <summary>
		/// モデルノードの一覧を取得します。
		/// </summary>
		/// <returns>
		/// モデルノードの一覧
		/// </returns>
		const Array<ModelNode>& nodes() const;
	};
}
