﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <memory>
# include "Siv3DFwd.hpp"
# include "SivArray.hpp"
# include "SivPolygon.hpp"

namespace s3d
{
	class MultiPolygon
	{
	private:

		struct MultiPolygonDetail;

		std::shared_ptr<MultiPolygonDetail> m_detail;

		MultiPolygon(MultiPolygonDetail&& detail);

	public:

		using iterator			= Array<Polygon>::iterator;
		using const_iterator	= Array<Polygon>::const_iterator;
		using reverse_iterator	= Array<Polygon>::reverse_iterator;	
		using const_reverse_iterator = Array<Polygon>::const_reverse_iterator;

		MultiPolygon();

		explicit MultiPolygon(const String& pts)
		{
			std::wistringstream(pts.str()) >> *this;
		}

		explicit MultiPolygon(const Array<Polygon>& polygons);

		explicit MultiPolygon(Array<Polygon>&& polygons);

		explicit operator bool() const { return !isEmpty; }

		Property_Get(bool, isEmpty) const;

		MultiPolygon clone() const;

		MultiPolygon operator + (const Vec2& v) const { return movedBy(v); }

		MultiPolygon operator - (const Vec2& v) const { return movedBy(-v); }

		MultiPolygon movedBy(double x, double y) const { return movedBy({ x, y }); }

		MultiPolygon movedBy(const Vec2& v) const;

		MultiPolygon& moveBy(double x, double y) { return moveBy({ x, y }); }

		MultiPolygon& moveBy(const Vec2& v);

		MultiPolygon rotated(double angle) const { return rotatedAt({ 0, 0 }, angle); }

		MultiPolygon rotatedAt(const Vec2& pos, double angle) const;

		double area() const;

		double perimeter() const;

		//Vec2 centroid() const;

		Property_Get(size_t, num_polygons) const;

		const Polygon& polygon(unsigned index) const;

		Property_Get(RectF, boundingRect) const;

		//Polygon computeConvexHull() const;

		MultiPolygon simplified(double maxDistance = 2.0) const;

		const_iterator begin() const;

		const_iterator end() const;

		reverse_iterator rbegin();

		reverse_iterator rend();

		const_reverse_iterator rbegin() const;

		const_reverse_iterator rend() const;

		const_iterator cbegin() const;

		const_iterator cend() const;

		const_reverse_iterator crbegin() const;

		const_reverse_iterator crend() const;

		/**
		*
		*	@param shape
		*	@return
		*/
		template<typename Shape>
		bool intersects(const Shape& shape) const
		{
			return Geometry2D::Intersect(*this, shape);
		}

		/**
		*
		*	@param shape
		*	@return
		*/
		template<typename Shape>
		bool contains(const Shape& shape) const
		{
			return Geometry2D::Within(shape, *this);
		}

		/**
		*
		*/
		Property_Get(bool, leftClicked) const;

		/**
		*
		*/
		Property_Get(bool, leftPressed) const;

		/**
		*
		*/
		Property_Get(bool, leftReleased) const;

		/**
		*
		*/
		Property_Get(bool, rightClicked) const;

		/**
		*
		*/
		Property_Get(bool, rightPressed) const;

		/**
		*
		*/
		Property_Get(bool, rightReleased) const;

		/**
		*
		*/
		Property_Get(bool, mouseOver) const;

		/**
		*
		*	@param image
		*	@param color
		*/
		void write(Image& image, const Color& color = Palette::Black) const;

		/**
		*
		*	@param image
		*	@param x
		*	@param y
		*	@param color
		*/
		void write(Image& image, int x, int y, const Color& color = Palette::Black) const;

		/**
		*
		*	@param image
		*	@param pos
		*	@param color
		*/
		void write(Image& image, const Point& pos, const Color& color = Palette::Black) const;

		/**
		*
		*	@param image
		*	@param color
		*/
		void overwrite(Image& image, const Color& color = Palette::Black) const;

		/**
		*
		*	@param image
		*	@param x
		*	@param y
		*	@param color
		*/
		void overwrite(Image& image, int x, int y, const Color& color = Palette::Black) const;

		/**
		*
		*	@param image
		*	@param pos
		*	@param color
		*/
		void overwrite(Image& image, const Point& pos, const Color& color = Palette::Black) const;

		/**
		*
		*	@param thickness
		*	@param col
		*/
		void writeFrame(Image& image, const Color& color = Palette::Black) const;

		/**
		*
		*	@param thickness
		*	@param col
		*/
		void writeFrame(Image& image, int thickness, const Color& color = Palette::Black) const;

		/**
		*
		*	@param x
		*	@param y
		*	@param thickness
		*	@param color
		*/
		void writeFrame(Image& image, int x, int y, int thickness = 1, const Color& color = Palette::Black) const;

		/**
		*
		*	@param pos
		*	@param thickness
		*	@param color
		*/
		void writeFrame(Image& image, const Point& pos, int thickness = 1, const Color& color = Palette::Black) const;

		/**
		*
		*	@param thickness
		*	@param col
		*/
		void overwriteFrame(Image& image, const Color& color = Palette::Black) const;

		/**
		*
		*	@param thickness
		*	@param col
		*/
		void overwriteFrame(Image& image, int thickness, const Color& color = Palette::Black) const;

		/**
		*
		*	@param x
		*	@param y
		*	@param thickness
		*	@param color
		*/
		void overwriteFrame(Image& image, int x, int y, int thickness = 1, const Color& color = Palette::Black) const;

		/**
		*
		*	@param pos
		*	@param thickness
		*	@param color
		*/
		void overwriteFrame(Image& image, const Point& pos, int thickness = 1, const Color& color = Palette::Black) const;

		/**
		*
		*	@param color
		*/
		void draw(const Color& color = Palette::White) const;

		/**
		*
		*	@param x
		*	@param y
		*	@param color
		*/
		void draw(double x, double y, const Color& color = Palette::White) const;

		/**
		*
		*	@param pos
		*	@param color
		*/
		void draw(const Vec2& pos, const Color& color = Palette::White) const;

		/**
		*
		*	@param thickness
		*	@param col
		*/
		void drawFrame(double thickness = 1.0, const Color& color = Palette::White) const;

		/**
		*
		*	@param x
		*	@param y
		*	@param thickness
		*	@param color
		*/
		void drawFrame(double x, double y, double thickness = 1.0, const Color& color = Palette::White) const;

		/**
		*
		*	@param pos
		*	@param thickness
		*	@param color
		*/
		void drawFrame(const Vec2& pos, double thickness = 1.0, const Color& color = Palette::White) const;

		/**
		*
		*	@param color
		*/
		void drawWireframe(const Color& color = Palette::White) const;

		/**
		*
		*	@param x
		*	@param y
		*	@param color
		*/
		void drawWireframe(double x, double y, const Color& color = Palette::White) const;

		/**
		*
		*	@param pos
		*	@param color
		*/
		void drawWireframe(const Vec2& pos, const Color& color = Palette::White) const;

		const MultiPolygonDetail* detail() const;
	};

	/**
	*
	*	@param os
	*	@param lineStr
	*	@return
	*/
	template<typename CharType>
	inline
		std::basic_ostream<CharType>&
		operator << (
		std::basic_ostream<CharType>& os,
		const MultiPolygon& multiPolygon
		)
	{
		os << CharType('(');

		bool b = false;

		for (const auto& polygon : multiPolygon)
		{
			if (Exchange(b, true))
			{
				os << CharType(',');
			}

			os << polygon;
		}

		return os << CharType(')');
	}

	/**
	*
	*	@param is
	*	@param lineStr
	*	@return
	*/
	template<typename CharType>
	inline
		std::basic_istream<CharType>&
		operator >> (
		std::basic_istream<CharType>& is,
		MultiPolygon& multiPolygon
		)
	{
		CharType unused;

		is >> unused;

		Array<Polygon> polygons;

		do
		{
			Polygon polygon;

			is >> polygon;

			is >> unused;

			polygons.push_back(polygon);
		}
		while (is && unused != (')'));

		multiPolygon = MultiPolygon(polygons);

		return is;
	}
}
