﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <type_traits>
# include <utility>
# include <new>
# include <cassert>
# include <initializer_list>
# include <stdexcept>
# include <functional>

# ifndef SIV_CPP11_IMPLEMENTED

#	ifndef SIV_ALIGNOF
#		define SIV_ALIGNOF(x) __alignof(x)
#	endif

#	ifndef SIV_ALIGNAS
#		define SIV_ALIGNAS(x) __declspec(align(x))
#	endif

#	define SIV_CONSTEXPR
#	define SIV_NOEXCEPT

# else

#	ifndef SIV_ALIGNOF
#		define SIV_ALIGNOF(x) alignof(x)
#	endif

#	ifndef SIV_ALIGNAS
#		define SIV_ALIGNAS(x) alignas(x)
#	endif

#	define SIV_CONSTEXPR constexpr
#	define SIV_NOEXCEPT noexcept

# endif

# define SIV_REQUIRES(R) static_assert(std::R::value==true, "");

namespace s3d
{
	namespace detail
	{
		template<unsigned int Align>
		struct type_with_alignment {};

#define TYPE_WITH_ALIGNMENT_IMPL(A, Name)	\
		struct SIV_ALIGNAS(A) Name			\
		{							 		\
			char m[A]; 						\
			using type = Name;				\
		};									\
		template<>							\
		struct type_with_alignment<A> : public Name {}

#define TYPE_WITH_ALIGNMENT(A) TYPE_WITH_ALIGNMENT_IMPL(A, aligned_with_##A)

		TYPE_WITH_ALIGNMENT(1);
		TYPE_WITH_ALIGNMENT(2);
		TYPE_WITH_ALIGNMENT(4);
		TYPE_WITH_ALIGNMENT(8);
		TYPE_WITH_ALIGNMENT(16);
		TYPE_WITH_ALIGNMENT(32);
		TYPE_WITH_ALIGNMENT(64);
		TYPE_WITH_ALIGNMENT(128);

#undef TYPE_WITH_ALIGNMENT
#undef TYPE_WITH_ALIGNMENT_IMPL

		template<typename T>
		class aligned_storage
		{
		private:

			union aligned_impl
			{
				char data[sizeof(T)];
				type_with_alignment<SIV_ALIGNOF(T)> dummy;
			};

			aligned_impl m_storage;

		public:

			const void* address() const
			{
				return static_cast<const void*>(&m_storage);
			}

			void* address()
			{
				return static_cast<void*>(&m_storage);
			}
		};
	}

	//
	//	In-place construction
	//
	struct in_place_t {};
	const SIV_CONSTEXPR in_place_t in_place{};

	//
	//	No-value state indicator
	//
	struct nullopt_helper {};
	typedef int nullopt_helper::* nullopt_t;

	/// <summary>
	/// 無効値
	/// </summary>
	const SIV_CONSTEXPR nullopt_t none{};

	//
	//	Class bad_optional_access
	//
	class bad_optional_access : public std::logic_error
	{
	public:
		explicit bad_optional_access(const std::string& what_arg)
			: logic_error(what_arg) {}

		explicit bad_optional_access(const char* what_arg)
			: logic_error(what_arg) {}
	};

	/// <summary>
	/// Optional
	/// </summary>
	template<typename T>
	class Optional
	{
	private:

		detail::aligned_storage<T> m_value;

		bool m_initialized = false;

	public:

		using cv_removed_type = std::remove_cv_t<T>;
		static_assert(!std::is_same<cv_removed_type, nullopt_t>::value, "bad T");
		static_assert(!std::is_same<cv_removed_type, in_place_t>::value, "bad T");
		static_assert(!std::is_reference<T>::value, "the reference type is not supported");

		using value_type = T;

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		SIV_CONSTEXPR Optional() SIV_NOEXCEPT{}

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param>
		/// 無効値
		/// </param>
		SIV_CONSTEXPR Optional(nullopt_t) SIV_NOEXCEPT{}

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param="another">
		/// 他の Optional オブジェクト
		/// </param>
		Optional(const Optional& another)
		{
			SIV_REQUIRES(is_copy_constructible<value_type>);

			if (another)
			{
				construct(another.value());
			}
		}

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param="another">
		/// 他の Optional オブジェクト
		/// </param>
		Optional(Optional&& another) SIV_NOEXCEPT
		{
			SIV_REQUIRES(is_move_constructible<value_type>);

			if (another)
			{
				emplace(std::move(another.value()));
			}

			another = none;
		}

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param="v">
		/// コピーする値
		/// </param>
		SIV_CONSTEXPR Optional(const value_type& v)
		{
			SIV_REQUIRES(is_copy_constructible<value_type>);

			construct(v);
		}

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param="v">
		/// コピーする値
		/// </param>
		SIV_CONSTEXPR Optional(T&& v)
		{
			SIV_REQUIRES(is_move_constructible<value_type>);

			emplace(v);
		}

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param="args">
		/// 値のコンストラクタ引数
		/// </param>
		template <class... Args> SIV_CONSTEXPR explicit Optional(in_place_t, Args&&... args)
		{
			static_assert(std::is_constructible<value_type, Args&&...>::value, "");

			destroy();

			::new (m_value.address()) value_type(std::forward<Args>(args)...);

			m_initialized = true;
		}

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param="args">
		/// 値のコンストラクタ引数
		/// </param>
		template <class U, class... Args> SIV_CONSTEXPR explicit Optional(in_place_t, std::initializer_list<U> ilist, Args&&... args)
		{
			destroy();

			::new (m_value.address()) value_type(ilist, std::forward<Args>(args)...);

			m_initialized = true;
		}

		/// <summary>
		/// デストラクタ
		/// </summary>
		~Optional()
		{
			destroy();
		}

		/// <summary>
		/// 代入演算子
		/// </summary>
		/// <returns>
		/// *this
		/// </returns>
		Optional& operator=(nullopt_t) SIV_NOEXCEPT
		{
			destroy();

			return *this;
		}

		/// <summary>
		/// 代入演算子
		/// </summary>
		/// <param="another">
		/// 他の Optional オブジェクト
		/// </param>
		/// <returns>
		/// *this
		/// </returns>
		Optional& operator=(const Optional& another)
		{
			SIV_REQUIRES(is_copy_constructible<value_type>);
			SIV_REQUIRES(is_copy_assignable<value_type>);

			if (m_initialized == true && another.m_initialized == false) destroy();
			else if (m_initialized == false && another.m_initialized == true) construct(*another);
			else if (m_initialized == true && another.m_initialized == true) m_value = another.m_value;
			return *this;
		}

		/// <summary>
		/// 代入演算子
		/// </summary>
		/// <param="another">
		/// 他の Optional オブジェクト
		/// </param>
		/// <returns>
		/// *this
		/// </returns>
		Optional& operator=(Optional&& another) SIV_NOEXCEPT
		{
			SIV_REQUIRES(is_move_constructible<value_type>);
			SIV_REQUIRES(is_move_assignable<value_type>);

			if (m_initialized == true && another.m_initialized == false) destroy();
			else if (another.m_initialized == true) construct(std::move(*another));
			return *this;
		}

		/// <summary>
		/// 代入演算子
		/// </summary>
		/// <param="val">
		/// コピーする値
		/// </param>
		/// <returns>
		/// *this
		/// </returns>
		template <class U> auto operator=(U&& val) -> typename std::enable_if<std::is_same<typename std::remove_reference_t<U>, T>::value, Optional&>::type
		{
			static_assert(std::is_constructible<value_type, U>::value, "");
			static_assert(std::is_assignable<value_type&, U>::value, "");

			if (static_cast<bool>(*this))
			{
				value() = std::forward<U>(val);
			}
			else
			{
				emplace(std::forward<U>(val));
			}
			return *this;
		}

		/// <summary>
		/// Optional オブジェクトを初期化します。
		/// </summary>
		/// <param="args">
		/// 値のコンストラクタ引数
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		template<typename... Args> void emplace(Args&&... args)
		{
			static_assert(std::is_constructible<value_type, Args&&...>::value, "");

			destroy();

			::new (m_value.address()) T(std::forward<Args>(args)...);

			m_initialized = true;
		}

		/// <summary>
		/// Optional オブジェクトを初期化します。
		/// </summary>
		/// <param="ilist">
		/// 値のコンストラクタ引数
		/// </param>
		/// <param="args">
		/// 値のコンストラクタ引数
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		template <class U, class... Args> void emplace(std::initializer_list<U> ilist, Args&&... args)
		{
			static_assert(std::is_constructible<value_type, std::initializer_list<U>&, Args&&...>::value, "");

			destroy();

			::new (m_value.address()) T(ilist, std::forward<Args>(args)...);

			m_initialized = true;
		}

		/// <summary>
		/// 別の Optional オブジェクトと中身を入れ替えます。
		/// </summary>
		/// <param name="another">
		/// 別の Optional オブジェクト
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void swap(Optional& another) SIV_NOEXCEPT
		{
			SIV_REQUIRES(is_move_constructible<value_type>);

			Optional tmp(another);

			another = std::move(*this);

			*this = std::move(tmp);
		}

		/// <summary>
		/// 中身へのポインタを返します。
		/// </summary>
		/// <remarks>
		/// 中身が none の場合の動作は未定義です。
		/// </remarks>
		/// <returns>
		/// 中身へのポインタ
		/// </returns>
		SIV_CONSTEXPR const T* operator ->() const
		{
			return get_ptr();
		}

		/// <summary>
		/// 中身へのポインタを返します。
		/// </summary>
		/// <remarks>
		/// 中身が none の場合の動作は未定義です。
		/// </remarks>
		/// <returns>
		/// 中身へのポインタ
		/// </returns>
		T* operator ->()
		{
			return get_ptr();
		}

		/// <summary>
		/// 中身の参照を返します。
		/// </summary>
		/// <remarks>
		/// 中身が none の場合の動作は未定義です。
		/// </remarks>
		/// <returns>
		/// 中身の参照
		/// </returns>
		SIV_CONSTEXPR const T& operator *() const
		{
			return *get_ptr();
		}

		/// <summary>
		/// 中身の参照を返します。
		/// </summary>
		/// <remarks>
		/// 中身が none の場合の動作は未定義です。
		/// </remarks>
		/// <returns>
		/// 中身の参照
		/// </returns>
		T& operator *()
		{
			return *get_ptr();
		}

		/// <summary>
		/// 中身があるかを返します。
		/// </summary>
		/// <returns>
		/// 中身が none ではない場合 true, それ以外の場合は false
		/// </returns>
		SIV_CONSTEXPR explicit operator bool() const SIV_NOEXCEPT
		{
			return m_initialized;
		}

		/// <summary>
		/// 中身の参照を返します。
		/// </summary>
		/// <exception cref="bad_optional_access">
		/// 中身が none の場合 throw されます。
		/// </exception>
		/// <returns>
		/// 中身の参照
		/// </returns>
		SIV_CONSTEXPR const T& value() const
		{
			if (!m_initialized)
			{
				throw bad_optional_access("bad access");
			}

			return *get_ptr();
		}

		/// <summary>
		/// 中身の参照を返します。
		/// </summary>
		/// <exception cref="bad_optional_access">
		/// 中身が none の場合 throw されます。
		/// </exception>
		/// <returns>
		/// 中身の参照
		/// </returns>
		T& value()
		{
			if (!m_initialized)
			{
				throw bad_optional_access("bad access");
			}

			return *get_ptr();
		}

		/// <summary>
		/// 中身がある場合はその値を、それ以外の場合は v を返します。
		/// </summary>
		/// <param name="v">
		/// 中身が none の場合に返ってくる値
		/// </param>
		/// <returns>
		/// 中身がある場合はその値、それ以外の場合は v
		/// </returns>
		template <class U> SIV_CONSTEXPR value_type value_or(U&& v) const
#ifdef SIV_CPP11_IMPLEMENTED
			&
#endif
		{
			SIV_REQUIRES(is_copy_constructible<value_type>);
			static_assert(std::is_convertible<U&&, value_type>::value, "");

			if (static_cast<bool>(*this))
			{
				return **this;
			}

			return static_cast<value_type>(std::forward<U>(v));
		}

#ifdef SIV_CPP11_IMPLEMENTED
		template <class U> T&& value_or(U&& v) &&
		{
			SIV_REQUIRES(is_move_constructible<value_type>);
			static_assert(std::is_convertible<U&&, value_type>::value, "");

			if (static_cast<bool>(*this))
			{
				return std::move(**this);
			}

			return static_cast<value_type>(std::forward<U>(v));
		}
#endif

		/// <summary>
		/// 中身を none にします。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void reset() SIV_NOEXCEPT
		{
			destroy();
		}

		/// <summary>
		/// 中身に別の値を代入します。
		/// </summary>
		/// <param name="v">
		/// コピーする値
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void reset(const T& v)
		{
			construct(v);
		}

		/// <summary>
		/// 中身がある場合に、その値を引数に関数 f を呼びます。
		/// </summary>
		/// <param name="f">
		/// 中身の値と同じ型を引数にとる関数
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void then(std::function<void(T&)> f)
		{
			if (static_cast<bool>(*this))
			{
				f(value());
			}
		}

		/// <summary>
		/// 中身がある場合に、その値を引数に関数 f を呼びます。
		/// </summary>
		/// <param name="f">
		/// 中身の値と同じ型を引数にとる関数
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void then(std::function<void(const T&)> f) const
		{
			if (static_cast<bool>(*this))
			{
				f(value());
			}
		}

	private:

		void construct(const T& v)
		{
			destroy();

			::new (m_value.address()) value_type(v);

			m_initialized = true;
		}

		void construct(T&& v)
		{
			destroy();

			::new (m_value.address()) value_type(std::forward<value_type>(v));

			m_initialized = true;
		}

		void destroy() SIV_NOEXCEPT
		{
			if (static_cast<bool>(*this))
			{
				get_ptr()->~T();
			}

			m_initialized = false;
		}

		const T* get_ptr() const
		{
			assert(static_cast<bool>(*this));

			return static_cast<const T*>(m_value.address());
		}

		T* get_ptr()
		{
			assert(static_cast<bool>(*this));

			return static_cast<T*>(m_value.address());
		}
	};

	//
	//	Relational operators
	//
	template <class T>
	SIV_CONSTEXPR bool operator==(const Optional<T>& x, const Optional<T>& y)
	{
		if (static_cast<bool>(x) != static_cast<bool>(y))
		{
			return false;
		}

		if (!x)
		{
			return true;
		}

		return *x == *y;
	}

	template <class T>
	SIV_CONSTEXPR bool operator!=(const Optional<T>& x, const Optional<T>& y)
	{
		return !(x == y);
	}

	template <class T>
	SIV_CONSTEXPR bool operator<(const Optional<T>& x, const Optional<T>& y)
	{
		if (!y)
		{
			return false;
		}
		else if (!x)
		{
			return true;
		}

		return *x < *y;
	}

	template <class T>
	SIV_CONSTEXPR bool operator>(const Optional<T>& x, const Optional<T>& y)
	{
		return y<x;
	}

	template <class T>
	SIV_CONSTEXPR bool operator<=(const Optional<T>& x, const Optional<T>& y)
	{
		return !(y < x);
	}

	template <class T>
	SIV_CONSTEXPR bool operator>=(const Optional<T>& x, const Optional<T>& y)
	{
		return !(x < y);
	}

	//
	//	Comparison with nullopt
	//
	template <class T>
	SIV_CONSTEXPR bool operator==(const Optional<T>& x, nullopt_t) SIV_NOEXCEPT
	{
		return !x;
	}

	template <class T>
	SIV_CONSTEXPR bool operator==(nullopt_t, const Optional<T>& x) SIV_NOEXCEPT
	{
		return !x;
	}

	template <class T>
	SIV_CONSTEXPR bool operator!=(const Optional<T>& x, nullopt_t) SIV_NOEXCEPT
	{
		return static_cast<bool>(x);
	}

	template <class T>
	SIV_CONSTEXPR bool operator!=(nullopt_t, const Optional<T>& x) SIV_NOEXCEPT
	{
		return static_cast<bool>(x);
	}

	template <class T>
	SIV_CONSTEXPR bool operator<(const Optional<T>&, nullopt_t) SIV_NOEXCEPT
	{
		return false;
	}

	template <class T>
	SIV_CONSTEXPR bool operator<(nullopt_t, const Optional<T>& x) SIV_NOEXCEPT
	{
		return static_cast<bool>(x);
	}

	template <class T>
	SIV_CONSTEXPR bool operator<=(const Optional<T>& x, nullopt_t) SIV_NOEXCEPT
	{
		return !x;
	}

	template <class T>
	SIV_CONSTEXPR bool operator<=(nullopt_t, const Optional<T>&) SIV_NOEXCEPT
	{
		return true;
	}

	template <class T>
	SIV_CONSTEXPR bool operator>(const Optional<T>& x, nullopt_t) SIV_NOEXCEPT
	{
		return static_cast<bool>(x);
	}

	template <class T>
	SIV_CONSTEXPR bool operator>(nullopt_t, const Optional<T>&) SIV_NOEXCEPT
	{
		return false;
	}

	template <class T>
	SIV_CONSTEXPR bool operator>=(const Optional<T>&, nullopt_t) SIV_NOEXCEPT
	{
		return true;
	}

	template <class T>
	SIV_CONSTEXPR bool operator>=(nullopt_t, const Optional<T>& x) SIV_NOEXCEPT
	{
		return !x;
	}

	//
	//	Comparison with T
	//
	template <class T>
	SIV_CONSTEXPR bool operator==(const Optional<T>& x, const T& v)
	{
		return static_cast<bool>(x) ? (*x == v) : false;
	}

	template <class T>
	SIV_CONSTEXPR bool operator==(const T& v, const Optional<T>& x)
	{
		return static_cast<bool>(x) ? (v == *x) : false;
	}

	template <class T>
	SIV_CONSTEXPR bool operator!=(const Optional<T>& x, const T& v)
	{
		return static_cast<bool>(x) ? !(*x == v) : true;
	}

	template <class T>
	SIV_CONSTEXPR bool operator!=(const T& v, const Optional<T>& x)
	{
		return static_cast<bool>(x) ? !(v == *x) : true;
	}

	template <class T>
	SIV_CONSTEXPR bool operator<(const Optional<T>& x, const T& v)
	{
		return static_cast<bool>(x) ? (*x < v) : true;
	}

	template <class T>
	SIV_CONSTEXPR bool operator<(const T& v, const Optional<T>& x)
	{
		return static_cast<bool>(x) ? (v < *x) : false;
	}

	template <class T>
	SIV_CONSTEXPR bool operator<=(const Optional<T>& x, const T& v)
	{
		return !(x > v);
	}

	template <class T>
	SIV_CONSTEXPR bool operator<=(const T& v, const Optional<T>& x)
	{
		return !(v > x);
	}

	template <class T>
	SIV_CONSTEXPR bool operator>(const Optional<T>& x, const T& v)
	{
		return static_cast<bool>(x) ? (v< *x) : false;
	}

	template <class T>
	SIV_CONSTEXPR bool operator>(const T& v, const Optional<T>& x)
	{
		return static_cast<bool>(x) ? (*x < v) : true;
	}

	template <class T>
	SIV_CONSTEXPR bool operator>=(const Optional<T>& x, const T& v)
	{
		return !(x < v);
	}

	template <class T>
	SIV_CONSTEXPR bool operator>=(const T& v, const Optional<T>& x)
	{
		return !(v < x);
	}

	//
	//	Specialized algorithms
	//
	template <class T>
	SIV_CONSTEXPR Optional<std::decay_t<T>> make_Optional(T&& v)
	{
		return Optional<std::decay_t<T>>(std::forward<T>(v));
	}
}

namespace std
{
	//
	//	Specialized algorithms
	//
	template <class T>
	void swap(s3d::Optional<T>& left, s3d::Optional<T>& y)
#ifdef SIV_CPP11_IMPLEMENTED
		SIV_NOEXCEPT(SIV_NOEXCEPT(left.swap(y)))
#endif
	{
		left.swap(y);
	}

	//
	//	Hash support
	//
	template <class T>
	struct hash<s3d::Optional<T>>
	{
		std::size_t operator() (const s3d::Optional<T>& arg) const
		{
			if (arg)
			{
				return std::hash<T>{}(*arg);
			}

			return 0;
		}
	};
}

# undef SIV_REQUIRES
# undef SIV_ALIGNOF
# undef SIV_ALIGNAS

# ifndef SIV_CPP11_IMPLEMENTED
#	undef SIV_CONSTEXPR
#	undef SIV_NOEXCEPT
# endif
