﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <sstream>
# include "SivString.hpp"
# include "SivOptional.hpp"

namespace s3d
{
	/// <summary>
	/// 文字列をデータ型に変換します。
	/// </summary>
	/// <param name="str">
	/// 変換する文字列
	/// </param>
	/// <returns>
	/// 文字列から変換されたデータ
	/// </returns>
	template <typename Type> inline Type Parse(const String& str)
	{
		Type to;

		if (std::wistringstream{ str.str() } >> to)
		{
			return to;
		}

		return to;
	}

	/// <summary>
	/// 文字列をデータ型に変換します。
	/// </summary>
	/// <param name="str">
	/// 変換する文字列
	/// </param>
	/// <returns>
	/// 文字列から変換されたデータ
	/// </returns>
	template <> inline bool Parse<bool>(const String& str)
	{
		return str.trim().lower() == L"true";
	}

	/// <summary>
	/// 文字列をデータ型に変換します。
	/// </summary>
	/// <param name="str">
	/// 変換する文字列
	/// </param>
	/// <returns>
	/// 文字列から変換されたデータ
	/// </returns>
	template <> inline String Parse<String>(const String& str)
	{
		return str;
	}

	/// <summary>
	/// 文字列をデータ型に変換します。
	/// </summary>
	/// <param name="str">
	/// 変換する文字列
	/// </param>
	/// <param name="defaultValue">
	/// 変換に失敗したときに返す値
	/// </param>
	/// <returns>
	/// 文字列から変換されたデータ。変換に失敗した場合は defaultValue
	/// </returns>
	template <typename Type> inline Type ParseOr(const String& str, const Type& defaultValue)
	{
		Type to;

		if (std::wistringstream{ str.str() } >> to)
		{
			return to;
		}

		return defaultValue;
	}

	/// <summary>
	/// 文字列をデータ型に変換します。
	/// </summary>
	/// <param name="str">
	/// 変換する文字列
	/// </param>
	/// <param name="defaultValue">
	/// 変換に失敗したときに返す値
	/// </param>
	/// <returns>
	/// 文字列から変換されたデータ。変換に失敗した場合は defaultValue
	/// </returns>
	template <> inline bool ParseOr<bool>(const String& str, const bool& defaultValue)
	{
		const String lower = str.trim().lower();

		if (lower == L"true")
		{
			return true;
		}
		else if (lower == L"false")
		{
			return false;
		}

		return defaultValue;
	}

	/// <summary>
	/// 文字列をデータ型に変換します。
	/// </summary>
	/// <param name="str">
	/// 変換する文字列
	/// </param>
	/// <returns>
	/// 文字列から変換されたデータ。
	/// </returns>
	template <> inline String ParseOr<String>(const String& str, const String&)
	{
		return str;
	}

	/// <summary>
	/// 文字列をデータ型に変換します。
	/// </summary>
	/// <param name="str">
	/// 変換する文字列
	/// </param>
	/// <returns>
	/// 文字列から変換されたデータの Optional 型。変換に失敗した場合は none
	/// </returns>
	template <typename Type> inline Optional<Type> ParseOpt(const String& str)
	{
		Type to;

		if (std::wistringstream{ str.str() } >> to)
		{
			return Optional<Type>(to);
		}

		return none;
	}

	/// <summary>
	/// 文字列をデータ型に変換します。
	/// </summary>
	/// <param name="str">
	/// 変換する文字列
	/// </param>
	/// <returns>
	/// 文字列から変換されたデータの Optional 型。変換に失敗した場合は none
	/// </returns>
	template <> inline Optional<bool> ParseOpt<bool>(const String& str)
	{
		const String lower = str.trim().lower();

		if (lower == L"true")
		{
			return Optional<bool>(true);
		}
		else if (lower == L"false")
		{
			return Optional<bool>(false);
		}

		return none;
	}

	/// <summary>
	/// 文字列をデータ型に変換します。
	/// </summary>
	/// <param name="str">
	/// 変換する文字列
	/// </param>
	/// <returns>
	/// 文字列から変換されたデータの Optional 型
	/// </returns>
	template <> inline Optional<String> ParseOpt<String>(const String& str)
	{
		return Optional<String>(str);
	}
}
