﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <memory>
# include <initializer_list>
# include "Siv3DFwd.hpp"
# include "SivPointVector.hpp"
# include "SivTriangle.hpp"
# include "SivPropertyMacro.hpp"

namespace s3d
{
	class PolygonHoles
	{
	private:

		const Array<Vec2>* m_begin;

		const Array<Vec2>* m_end;

	public:

		PolygonHoles(const Array<Vec2>* _begin, const Array<Vec2>* _end)
			: m_begin(_begin)
			, m_end(_end) {}

		bool empty() const
		{
			return !m_begin;
		}

		size_t size() const
		{
			return m_end - m_begin;
		}

		const Array<Vec2>& operator[](unsigned index) const
		{
			return m_begin[index];
		}

		const Array<Vec2>* begin() const
		{
			return m_begin;
		}

		const Array<Vec2>* end() const
		{
			return m_end;
		}
	};

	/// <summary>
	/// 多角形
	/// </summary>
	class Polygon
	{
	private:

		struct PolygonDetail;

		std::shared_ptr<PolygonDetail> m_detail;

		Polygon(PolygonDetail&& detail);

	public:

		Polygon();

		explicit Polygon(const String& pts)
		{
			std::wistringstream(pts.str()) >> *this;
		}

		explicit Polygon(const Array<Vec2>& outer, const Array<Array<Vec2>>& holes = {});

		explicit Polygon(const Vec2* outer, size_t size, const Array<Array<Vec2>>& holes = {});

		Polygon(std::initializer_list<Vec2> outer)
			: Polygon(outer.begin(), outer.size()) {}

		Polygon(const Array<Vec2>& outer, const Array<Array<Vec2>>& holes, const Array<unsigned short>& indices, const RectF& bound);

		explicit operator bool() const { return !isEmpty; }

		Property_Get(bool, isEmpty) const;

		Polygon clone() const;

		Polygon movedBy(double x, double y) const { return movedBy({ x, y }); }

		Polygon movedBy(const Vec2& v) const;

		Polygon& moveBy(double x, double y) { return moveBy({ x, y }); }

		Polygon& moveBy(const Vec2& v);

		Polygon scaled(double s) const { return scaled({ s, s }); }

		Polygon scaled(double sx, double sy) const { return scaled({ sx, sy }); }

		Polygon scaled(const Vec2& s) const;

		Polygon& scale(double s) { return scale({ s, s }); }

		Polygon& scale(double sx, double sy) { return scale({ sx, sy }); }

		Polygon& scale(const Vec2& s);

		Polygon& addHole(const Array<Vec2>& hole);

		Polygon& addHoles(const Array<Array<Vec2>>& holes);

		Polygon operator + (const Vec2& v) const;

		Polygon operator - (const Vec2& v) const { return *this + (-v); }

		Polygon rotated(double angle) const { return rotatedAt({ 0, 0 }, angle); }

		Polygon rotatedAt(const Vec2& pos, double angle) const;

		double area() const;

		double perimeter() const;

		Vec2 centroid() const;

		Property_Get(size_t, num_vertices) const;

		Property_Get(size_t, num_indices) const;

		Property_Get(size_t, num_triangles) const { return num_indices / 3; }

		Property_Get(size_t, num_holes) const;

		Property_Get(bool, hasHoles) const { return !holes().empty(); }

		const Array<Vec2>& outer() const;

		PolygonHoles holes() const;

		PolygonTriangles triangles() const;

		Vec2 vertex(unsigned index) const;

		unsigned short index(unsigned index) const;

		Array<unsigned short> indices() const;

		Triangle triangle(unsigned index) const;

		Property_Get(RectF, boundingRect) const;

		Polygon computeConvexHull() const;

		/// <summary>
		/// ポリゴンの頂点数を削減します。
		/// </summary>
		/// <param name="maxDistance">
		/// 
		/// </param>
		/// <returns>
		/// 頂点数が削減されたポリゴン
		/// </returns>
		Polygon simplified(double maxDistance = 2.0) const;

		/**
		*
		*	@param shape
		*	@return
		*/
		template<typename Shape>
		bool intersects(const Shape& shape) const
		{
			return Geometry2D::Intersect(*this, shape);
		}

		/**
		*
		*	@param shape
		*	@return
		*/
		template<typename Shape>
		bool contains(const Shape& shape) const
		{
			return Geometry2D::Within(shape, *this);
		}

		/**
		*
		*/
		Property_Get(bool, leftClicked) const;

		/**
		*
		*/
		Property_Get(bool, leftPressed) const;

		/**
		*
		*/
		Property_Get(bool, leftReleased) const;

		/**
		*
		*/
		Property_Get(bool, rightClicked) const;

		/**
		*
		*/
		Property_Get(bool, rightPressed) const;

		/**
		*
		*/
		Property_Get(bool, rightReleased) const;

		/**
		*
		*/
		Property_Get(bool, mouseOver) const;

		/**
		*
		*	@param image
		*	@param color
		*/
		void write(Image& image, const Color& color = Palette::Black) const;

		/**
		*
		*	@param image
		*	@param x
		*	@param y
		*	@param color
		*/
		void write(Image& image, int x, int y, const Color& color = Palette::Black) const;

		/**
		*
		*	@param image
		*	@param pos
		*	@param color
		*/
		void write(Image& image, const Point& pos, const Color& color = Palette::Black) const;

		/**
		*
		*	@param image
		*	@param color
		*/
		void overwrite(Image& image, const Color& color = Palette::Black) const;

		/**
		*
		*	@param image
		*	@param x
		*	@param y
		*	@param color
		*/
		void overwrite(Image& image, int x, int y, const Color& color = Palette::Black) const;

		/**
		*
		*	@param image
		*	@param pos
		*	@param color
		*/
		void overwrite(Image& image, const Point& pos, const Color& color = Palette::Black) const;

		/**
		*
		*	@param thickness
		*	@param col
		*/
		void writeFrame(Image& image, const Color& color = Palette::Black) const;

		/**
		*
		*	@param thickness
		*	@param col
		*/
		void writeFrame(Image& image, int thickness, const Color& color = Palette::Black) const;

		/**
		*
		*	@param x
		*	@param y
		*	@param thickness
		*	@param color
		*/
		void writeFrame(Image& image, int x, int y, int thickness = 1, const Color& color = Palette::Black) const;

		/**
		*
		*	@param pos
		*	@param thickness
		*	@param color
		*/
		void writeFrame(Image& image, const Point& pos, int thickness = 1, const Color& color = Palette::Black) const;

		/**
		*
		*	@param thickness
		*	@param col
		*/
		void overwriteFrame(Image& image, const Color& color = Palette::Black) const;

		/**
		*
		*	@param thickness
		*	@param col
		*/
		void overwriteFrame(Image& image, int thickness, const Color& color = Palette::Black) const;

		/**
		*
		*	@param x
		*	@param y
		*	@param thickness
		*	@param color
		*/
		void overwriteFrame(Image& image, int x, int y, int thickness = 1, const Color& color = Palette::Black) const;

		/**
		*
		*	@param pos
		*	@param thickness
		*	@param color
		*/
		void overwriteFrame(Image& image, const Point& pos, int thickness = 1, const Color& color = Palette::Black) const;


		/**
		*
		*	@param color
		*/
		void draw(const Color& color = Palette::White) const;

		/**
		*
		*	@param x
		*	@param y
		*	@param color
		*/
		void draw(double x, double y, const Color& color = Palette::White) const;

		/**
		*
		*	@param pos
		*	@param color
		*/
		void draw(const Vec2& pos, const Color& color = Palette::White) const;

		/**
		*
		*	@param thickness
		*	@param color
		*/
		void drawFrame(double thickness = 1.0, const Color& color = Palette::White) const;

		/**
		*
		*	@param x
		*	@param y
		*	@param thickness
		*	@param color
		*/
		void drawFrame(double x, double y, double thickness = 1.0, const Color& color = Palette::White) const;

		/**
		*
		*	@param pos
		*	@param thickness
		*	@param color
		*/
		void drawFrame(const Vec2& pos, double thickness = 1.0, const Color& color = Palette::White) const;

		/**
		*
		*	@param color
		*/
		void drawWireframe(const Color& color = Palette::White) const;

		/**
		*
		*	@param x
		*	@param y
		*	@param color
		*/
		void drawWireframe(double x, double y, const Color& color = Palette::White) const;

		/**
		*
		*	@param pos
		*	@param color
		*/
		void drawWireframe(const Vec2& pos, const Color& color = Palette::White) const;

		const PolygonDetail* detail() const;
	};

	class Polygon_triangle_iterator
		: public std::iterator<std::input_iterator_tag, Triangle>
	{
	private:

		const Polygon* m_pPolygon = nullptr;

		unsigned m_index;

	public:

		Polygon_triangle_iterator(
			const Polygon* pPolygon,
			unsigned index = 0
			)
			: m_pPolygon(pPolygon)
			, m_index(index) {}

		Polygon_triangle_iterator& operator++()
		{
			++m_index;

			return *this;
		}

		Triangle operator * () const
		{
			return m_pPolygon->triangle(m_index);
		}

		bool operator == (const Polygon_triangle_iterator& iterator)
		{
			return  m_pPolygon == iterator.m_pPolygon && m_index == iterator.m_index;
		}

		bool operator != (const Polygon_triangle_iterator& iterator)
		{
			return !(m_index == iterator.m_index);
		}
	};

	class PolygonTriangles
	{
	private:

		const Polygon* m_pPolygon = nullptr;

	public:

		PolygonTriangles(const Polygon* pPolygon)
			: m_pPolygon(pPolygon) {}

		const Polygon_triangle_iterator begin() const
		{
			return Polygon_triangle_iterator(m_pPolygon, 0);
		}

		const Polygon_triangle_iterator end() const
		{
			return Polygon_triangle_iterator(m_pPolygon, m_pPolygon->num_triangles);
		}
	};

	/**
	*
	*	@param os
	*	@param lineStr
	*	@return
	*/
	template<typename CharType>
	inline
	std::basic_ostream<CharType>&
		operator << (
		std::basic_ostream<CharType>& os,
		const Polygon& polygon
		)
	{
		os << CharType('(');

		os << CharType('(');

		bool b = false;

		for (const auto& point : polygon.outer())
		{
			if (Exchange(b, true))
			{
				os << CharType(',');
			}

			os << point;
		}

		os << CharType(')');

		if (!polygon.holes().empty())
		{
			os << CharType(',');

			os << CharType('(');

			b = false;

			for (const auto& hole : polygon.holes())
			{
				if (Exchange(b, true))
				{
					os << CharType(',');

					os << CharType('(');
				}

				bool b2 = false;

				for (const auto& point : hole)
				{
					if (Exchange(b2, true))
					{
						os << CharType(',');
					}

					os << point;
				}

				os << CharType(')');
			}
		}

		return os << CharType(')');
	}

	/**
	*
	*	@param is
	*	@param lineStr
	*	@return
	*/
	template<typename CharType>
	inline
	std::basic_istream<CharType>&
		operator >> (
		std::basic_istream<CharType>& is,
		Polygon& polygon
		)
	{
		CharType unused;

		is >> unused >> unused;

		Array<Vec2> pts;

		for (;;)
		{
			Vec2 v;

			is >> v;

			if (!is)
			{
				break;
			}

			is >> unused;

			pts.push_back(v);

			if (unused == CharType(')'))
			{
				break;
			}
		}

		is >> unused;

		if (unused != CharType(','))
		{
			polygon = Polygon(pts);

			return is;
		}

		Array<Array<Vec2>> holes;

		is >> unused;

		for (;;)
		{
			Array<Vec2> hole;

			for (;;)
			{
				Vec2 v;

				is >> v;

				if (!is)
				{
					break;
				}

				is >> unused;

				hole.push_back(v);

				if (unused == CharType(')'))
				{
					holes.push_back(std::move(hole));

					break;
				}
			}

			if (!is)
			{
				break;
			}

			is >> unused;

			if (unused != CharType(','))
			{
				polygon = Polygon(pts, holes);

				return is;
			}
			else
			{
				is >> unused;
			}
		}

		return is;
	}
}
