﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <cassert>
# include <type_traits>
# include <random>
# include <initializer_list>
# include "SivPropertyMacro.hpp"
# include "SivArray.hpp"

namespace s3d
{
	/// <summary>
	/// 乱数生成エンジン
	/// </summary>
	class RNG
	{
	private:

		// MT 11213b (周期 2^11213-1, メモリ消費 2816B)
		using mt11213b = std::mersenne_twister_engine<unsigned, 32, 351, 175, 19, 0xccab8ee7,
			11, 0xffffffff, 7, 0x31b6ab00, 15, 0xffe50000, 17, 1812433253>;

		mt11213b m_engine;

	public:

		using result_type = mt11213b::result_type;

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		/// <remarks>
		/// シードはランダムに設定されます。
		/// </remarks>
		RNG();

		/// <summary>
		/// シードを指定して乱数生成器を作成します。
		/// </summary>
		/// <param name="seed">
		/// シード
		/// </param>
		explicit RNG(unsigned seed)
			: m_engine(seed) {}

		/// <summary>
		/// シードを指定して乱数生成器を作成します。
		/// </summary>
		/// <param name="seeds">
		/// シード
		/// </param>
		explicit RNG(const Array<unsigned>& seeds);

		/// <summary>
		/// シードを設定しなおします。
		/// </summary>
		/// <param name="seed">
		/// シード
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void seed(unsigned seed)
		{
			m_engine.seed(seed);
		}

		/// <summary>
		/// シードを設定しなおします。
		/// </summary>
		/// <param name="seeds">
		/// シード
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void seed(const Array<unsigned>& seeds);

		/// <summary>
		/// 生成される乱数の最小値を返します。
		/// </summary>
		/// <returns>
		/// 生成される乱数の最小値
		/// </returns>
		static unsigned min()
		{
			return mt11213b::min();
		}

		/// <summary>
		/// 生成される乱数の最大値を返します。
		/// </summary>
		/// <returns>
		/// 生成される乱数の最大値
		/// </returns>
		static unsigned max()
		{
			return mt11213b::max();
		}

		/// <summary>
		/// 乱数を生成します。
		/// </summary>
		/// <returns>
		/// 乱数
		/// </returns>
		unsigned operator()()
		{
			return m_engine();
		}
	};

	/// <summary>
	/// 一様分布
	/// </summary>
	/// <remarks>
	/// 乱数の型と範囲を定義します。
	/// </remarks>
	template<typename Type>
	class UniformDistribution
	{
	private:

		using DistributionType = std::conditional_t<std::is_floating_point<Type>::value,
			std::uniform_real_distribution<Type>, std::uniform_int_distribution<Type>>;

		DistributionType m_distribution;

	public:

		/// <summary>
		/// 一様分布を設定します。
		/// </summary>
		/// <param name="min">
		/// 生成する乱数の最小値
		/// </param>
		/// <param name="max">
		/// 生成する乱数の最大値
		/// </param>
		UniformDistribution(Type min, Type max)
			: m_distribution(min, max) {}

		/// <summary>
		/// 一様分布の乱数を生成します。
		/// </summary>
		/// <param name="engine">
		/// 乱数生成エンジン
		/// </param>
		/// <returns>
		/// 乱数
		/// </returns>
		Type operator()(RNG& engine) const
		{
			return m_distribution(engine);
		}

		/// <summary>
		/// 一様分布を設定します。
		/// </summary>
		/// <param name="min">
		/// 生成する乱数の最小値
		/// </param>
		/// <param name="max">
		/// 生成する乱数の最大値
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void set(Type min, Type max)
		{
			m_distribution.param({ min, max });
		}

		/// <summary>
		/// 生成する乱数の最小値
		/// </summary>
		Property_Get(Type, min) const
		{
			return distribution.min();
		}

		/// <summary>
		/// 生成する乱数の最大値
		/// </summary>
		Property_Get(Type, max) const
		{
			return distribution.max();
		}
	};

	/// <summary>
	/// 正規分布
	/// </summary>
	/// <remarks>
	/// 正規分布を定義します。
	/// </remarks>
	class NormalDistribution
	{
	private:

		std::normal_distribution<double> m_distribution;

	public:

		/// <summary>
		/// 正規分布を設定します。
		/// </summary>
		/// <param name="mean">
		/// 平均
		/// </param>
		/// <param name="sigma">
		/// 標準偏差
		/// </param>
		NormalDistribution(double mean, double sigma)
			: m_distribution(mean, sigma){}

		/// <summary>
		/// 正規分布の乱数を生成します。
		/// </summary>
		/// <param name="engine">
		/// 乱数生成エンジン
		/// </param>
		/// <returns>
		/// 乱数
		/// </returns>
		double operator()(RNG& engine)
		{
			return m_distribution(engine);
		}

		/// <summary>
		/// 正規分布を設定します。
		/// </summary>
		/// <param name="mean">
		/// 平均
		/// </param>
		/// <param name="sigma">
		/// 標準偏差
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void set(double mean, double sigma)
		{
			m_distribution.param({ mean, sigma });
		}

		/// <summary>
		/// 平均
		/// </summary>
		Property_Get(double, mean) const
		{
			return m_distribution.mean();
		}

		/// <summary>
		/// 標準偏差
		/// </summary>
		Property_Get(double, sigma) const
		{
			return m_distribution.sigma();
		}
	};

	/// <summary>
	/// グローバルな乱数エンジンにアクセスします。
	/// </summary>
	/// <returns>
	/// グローバルな乱数エンジンへの参照
	/// </returns>
	RNG& DefaultRNG();

	/// <summary>
	/// グローバルな乱数エンジンのシードを再設定します。
	/// </summary>
	/// <param name="seed">
	/// シード
	/// </param>
	/// <returns>
	/// なし
	/// </returns>
	inline void Reseed(unsigned seed)
	{
		DefaultRNG().seed(seed);
	}

	/// <summary>
	/// [0.0-1.0] の範囲の乱数を返します。
	/// </summary>
	/// <remarks>
	/// グローバルな乱数エンジンを使用します。
	/// </remarks>
	/// <returns>
	/// [0.0-1.0] の範囲の乱数
	/// </returns>
	double Random();

	/// <summary>
	/// 指定した型と範囲の乱数を返します。
	/// </summary>
	/// <param name="min">
	/// 生成したい乱数の最小値
	/// </param>
	/// <param name="max">
	/// 生成したい乱数の最大値
	/// </param>
	/// <remarks>
	/// グローバルな乱数エンジンを使用します。
	/// </remarks>
	/// <returns>
	/// 指定した型と範囲の乱数
	/// </returns>
	template<typename Type> inline Type Random(Type min, Type max)
	{
		return UniformDistribution<Type>(min, max)(DefaultRNG());
	}

	/// <summary>
	/// 0 から、指定した範囲までの乱数を返します。
	/// </summary>
	/// <param name="max">
	/// 生成したい乱数の最大値
	/// </param>
	/// <remarks>
	/// グローバルな乱数エンジンを使用します。
	/// </remarks>
	/// <returns>
	/// 0 以上 max 以下の乱数
	/// </returns>
	template<typename Type> inline Type Random(Type max)
	{
		return Random<Type>(0, max);
	}

	/// <summary>
	/// 指定した型と範囲の乱数を返します。
	/// </summary>
	/// <param name="min">
	/// 生成したい乱数の最小値
	/// </param>
	/// <param name="max">
	/// 生成したい乱数の最大値
	/// </param>
	/// <remarks>
	/// グローバルな乱数エンジンを使用します。
	/// </remarks>
	/// <returns>
	/// 指定した型と範囲の乱数
	/// </returns>
	template<> inline wchar_t Random<wchar_t>(wchar_t min, wchar_t max)
	{
		return static_cast<wchar_t>(UniformDistribution<unsigned short>(min, max)(DefaultRNG()));
	}

	/// <summary>
	/// 2 値の乱数を返します。
	/// </summary>
	/// <param name="p">
	/// true を返す確率 [0.0, 1.0]
	/// </param>
	/// <remarks>
	/// グローバルな乱数エンジンを使用します。
	/// </remarks>
	/// <returns>
	/// 確率 p でtrue, 確率 (1-p) で false
	/// </returns>
	inline bool RandomBool(double p)
	{
		return std::bernoulli_distribution(p)(DefaultRNG());
	}

	/// <summary>
	/// リストの中からランダムに値を選択します。
	/// </summary>
	/// <param name="ilist">
	/// 値のリスト
	/// </param>
	/// <remarks>
	/// グローバルな乱数エンジンを使用します。
	/// </remarks>
	/// <returns>
	/// リストの中からランダムに選択された値
	/// </returns>
	template<typename Type> inline Type RandomSelect(std::initializer_list<Type> ilist)
	{
		assert(ilist.size() > 0);

		return *(ilist.begin() + (ilist.size() == 1 ? 0 : Random <size_t>(0, ilist.size() - 1)));
	}

	/// <summary>
	/// コンテナの中身をシャッフルします。
	/// </summary>
	/// <param name="c">
	/// コンテナ
	/// </param>
	/// <returns>
	/// なし
	/// </returns>
	template<typename Container> inline void Shuffle(Container& c)
	{
		std::shuffle(c.begin(), c.end(), RNG{});
	}

	/// <summary>
	/// 指定した範囲の値をシャッフルします。
	/// </summary>
	/// <param name="first">
	/// 範囲の開始位置
	/// </param>
	/// <param name="last">
	/// 範囲の終了位置
	/// </param>
	/// <returns>
	/// なし
	/// </returns>
	template<typename RandomIt> inline void Shuffle(RandomIt first, RandomIt last)
	{
		std::shuffle(first, last, RNG{});
	}
}

