﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "SivRectangle.hpp"
# include "SivPolygon.hpp"

namespace s3d
{
	struct RoundRect
	{
#pragma warning(disable:4201)
		union
		{
			struct { RectF rect; double r; };
			struct { double x, y, w, h, r; };
		};
#pragma warning(default:4201)

		RoundRect() {}

		/**
		*
		*	@param pts
		*/
		explicit RoundRect(const String& str)
		{
			std::wistringstream(str.str()) >> *this;
		}

		RoundRect(double _x, double _y, double _w, double _h, double _r)
			: rect(_x, _y, _w, _h)
			, r(_r) {}

		RoundRect(const Vec2& pos, const Vec2& size, double _r)
			: rect(pos, size)
			, r(_r) {}

		RoundRect(const RectF& rect, double _r)
			: rect(rect)
			, r(_r) {}

		/**
		*
		*	@param _x
		*	@param _y
		*	@return
		*/
		RoundRect& setPos(double _x, double _y) { rect.setPos(_x, _y); return *this; }

		/**
		*
		*	@param _pos
		*	@return
		*/
		RoundRect& setPos(const Vec2& _pos) { rect.setPos(_pos); return *this; }

		/**
		*
		*	@param _w
		*	@param _h
		*	@return
		*/
		RoundRect& setSize(double _w, double _h) { rect.setSize(_w, _h); return *this; }

		/**
		*
		*	@param _size
		*	@return
		*/
		RoundRect& setSize(const Vec2& _size) { rect.setSize(_size); return *this; }

		/**
		*
		*	@param _x
		*	@param _y
		*	@return
		*/
		RoundRect& setCenter(double _x, double _y) { rect.setCenter(_x, _y); return *this; }

		/**
		*
		*	@param _pos
		*	@return
		*/
		RoundRect& setCenter(const Vec2& _pos) { rect.setCenter(_pos); return *this; }

		/**
		*
		*	@param x
		*	@param y
		*	@return
		*/
		RoundRect movedBy(double x, double y) const { return movedBy({ x, y }); }

		/**
		*
		*	@param v
		*	@return
		*/
		RoundRect movedBy(const Vec2& v) const
		{
			return{ rect.movedBy(v), r };
		}

		/**
		*
		*	@param x
		*	@param y
		*	@return
		*/
		RoundRect& moveBy(double x, double y) { return moveBy({ x, y }); }

		/**
		*
		*	@param v
		*	@return
		*/
		RoundRect& moveBy(const Vec2& v)
		{
			rect.moveBy(v);

			return *this;
		}

		/**
		*
		*	@param xy
		*	@return
		*/
		RoundRect stretched(double xy) const { return{ rect.stretched(xy), r }; }

		/**
		*
		*	@param _x
		*	@param _y
		*	@return
		*/
		RoundRect stretched(double _x, double _y) const { return{ rect.stretched(_x, _y), r }; }

		/**
		*
		*	@param xy
		*	@return
		*/
		RoundRect stretched(const Vec2& xy) const { return{ rect.stretched(xy), r }; }

		/**
		*
		*	@param v
		*	@return
		*/
		RoundRect operator + (const Vec2& v) const { return movedBy(v); }

		/**
		*
		*	@param v
		*	@return
		*/
		RoundRect operator - (const Vec2& v) const { return movedBy(-v); }

		/**
		*
		*	@return
		*/
		double area() const;

		/**
		*
		*	@return
		*/
		double perimeter() const;

		/**
		*
		*	@param shape
		*	@return
		*/
		template<typename Shape>
		bool intersects(const Shape& shape) const
		{
			return Geometry2D::Intersect(*this, shape);
		}

		/**
		*
		*	@param shape
		*	@return
		*/
		template<typename Shape>
		bool contains(const Shape& shape) const
		{
			return Geometry2D::Within(shape, *this);
		}

		/**
		*
		*/
		Property_Get(bool, leftClicked) const;

		/**
		*
		*/
		Property_Get(bool, leftPressed) const;

		/**
		*
		*/
		Property_Get(bool, leftReleased) const;

		/**
		*
		*/
		Property_Get(bool, rightClicked) const;

		/**
		*
		*/
		Property_Get(bool, rightPressed) const;

		/**
		*
		*/
		Property_Get(bool, rightReleased) const;

		/**
		*
		*/
		Property_Get(bool, mouseOver) const;

		/**
		*
		*	@param image
		*	@param color
		*/
		void write(Image& image, const Color& color = Palette::Black) const;

		void overwrite(Image& image, const Color& color = Palette::Black) const;

		void writeFrame(Image& image, const Color& color = Palette::Black) const;

		void writeFrame(Image& image, int thickness, const Color& color = Palette::Black) const;

		void overwriteFrame(Image& image, const Color& color = Palette::Black) const;

		void overwriteFrame(Image& image, int thickness, const Color& color = Palette::Black) const;

		/**
		*
		*	@param col
		*	@return
		*/
		const RoundRect& draw(const Color& color = Palette::White) const;

		/**
		*
		*	@param col
		*	@return
		*/
		void drawFlatBottom(const Color& color = Palette::White) const;

		/**
		*
		*	@param thickness
		*	@param col
		*	@return
		*/
		const RoundRect& drawFrame(double innerThickness = 1.0, double outerThickness = 0.0, const Color& color = Palette::White) const;

		Polygon asPolygon() const;

		Polygon asPolygonFlatBottom() const;

		Array<Vec2> getOuterVertices(double offset = 0.0) const;

		Array<Vec2> getOuterVerticesFlatBottom(double offset = 0.0) const;
	};

	/**
	*
	*	@param os
	*	@param quad
	*	@return
	*/
	template<typename CharType>
	inline
	std::basic_ostream<CharType>&
		operator << (
		std::basic_ostream<CharType>& os,
		const RoundRect& roundRect
		)
	{
		return	os << CharType('(')
			<< roundRect.x << CharType(',')
			<< roundRect.y << CharType(',')
			<< roundRect.w << CharType(',')
			<< roundRect.h << CharType(',')
			<< roundRect.r << CharType(')');
	}

	/**
	*
	*	@param is
	*	@param quad
	*	@return
	*/
	template<typename CharType>
	inline
	std::basic_istream<CharType>&
		operator >> (
		std::basic_istream<CharType>& is,
		RoundRect& roundRect
		)
	{
		CharType unused;

		return	is >> unused
			>> roundRect.x >> unused
			>> roundRect.y >> unused
			>> roundRect.w >> unused
			>> roundRect.h >> unused
			>> roundRect.r >> unused;
	}
}
