﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <utility>
# include "SivColor.hpp"

namespace s3d
{
	enum class TextureFilter : unsigned char
	{
       MinMagMipPoint				= 0,
       MinMagPointMipLinear			= 0x1,
       MinPointMagLinearMipPoint	= 0x4,
       MinPointMagMipLinear			= 0x5,
       MinLinearMagMipPoint			= 0x10,
       MinLinearMagPointMipLinear	= 0x11,
       MinMagLinearMipPoint			= 0x14,
       MinMagMipLinear				= 0x15,
       Anisotropic					= 0x55,   
	};

	enum class TextureAddressMode : unsigned char
	{
        Wrap		= 1,
        Mirror		= 2,
        Clamp		= 3,
        Border		= 4,
        MirrorOnce	= 5
	};

	struct SamplerState
	{
		SamplerState(
			TextureAddressMode _addressU = TextureAddressMode::Wrap,
			TextureAddressMode _addressV = TextureAddressMode::Wrap,
			TextureAddressMode _addressW = TextureAddressMode::Wrap,
			TextureFilter _filter = TextureFilter::MinMagMipLinear,
			unsigned char _maxAnisotropy = 16,
			const Color& _borderColor = Color(0,0)
			)
			: filter(_filter)
			, addressU(_addressU)
			, addressV(_addressV)
			, addressW(_addressW)
			, maxAnisotropy(_maxAnisotropy)
			, borderColor(_borderColor) {}

		SamplerState(
			TextureAddressMode _address,
			TextureFilter _filter,
			unsigned char _maxAnisotropy = 16,
			const Color& _borderColor = Color(0, 0)
			)
			: filter(_filter)
			, addressU(_address)
			, addressV(_address)
			, addressW(_address)
			, maxAnisotropy(_maxAnisotropy)
			, borderColor(_borderColor) {}

		using DataType = unsigned __int64;

#pragma warning(disable:4201)
		union
		{
			struct
			{
				TextureAddressMode addressU : 4;

				TextureAddressMode addressV : 4;

				TextureAddressMode addressW : 8;

				TextureFilter filter;
				
				unsigned char maxAnisotropy;

				Color borderColor;
			};

			DataType _data;
		};
#pragma warning(default:4201)

		bool operator == (const SamplerState& s) const
		{
			return _data == s._data;
		}

		bool operator != (const SamplerState& s) const
		{
			return _data != s._data;
		}

		enum class Predefined
		{
			WrapPoint,
			WrapLinear,
			WrapAniso,
			MirrorPoint,
			MirrorLinear,
			MirrorAniso,
			ClampPoint,
			ClampLinear,
			ClampAniso,
			Default2D = WrapLinear,
			Default3D = WrapAniso,
		};

		SamplerState(Predefined predefined);

		const static Predefined WrapPoint = Predefined::WrapPoint;

		const static Predefined WrapLinear = Predefined::WrapLinear;

		const static Predefined WrapAniso = Predefined::WrapAniso;

		const static Predefined MirrorPoint = Predefined::MirrorPoint;

		const static Predefined MirrorLinear = Predefined::MirrorLinear;

		const static Predefined MirrorAniso = Predefined::MirrorAniso;

		const static Predefined ClampPoint = Predefined::ClampPoint;

		const static Predefined ClampLinear = Predefined::ClampLinear;

		const static Predefined ClampAniso = Predefined::ClampAniso;

		const static Predefined Default2D = Predefined::Default2D;

		const static Predefined Default3D = Predefined::Default3D;
	};

	static_assert(sizeof(SamplerState) == sizeof(SamplerState::DataType), "");
}

template<>
struct std::hash<s3d::SamplerState>
{
	size_t operator()(const s3d::SamplerState& keyVal) const
	{
		return std::hash<s3d::SamplerState::DataType>()(keyVal._data);
	}
};
