﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "SivPropertyMacro.hpp"

namespace s3d
{
	/// <summary>
	/// タイマー
	/// </summary>
	/// <remarks>
	/// TimerMillisec, TimerSec, TimerMinute が使用します。
	/// </remarks>
	template <int resolutionMilliseconds>
	class TimerBase
	{
	public:

		/// <summary>
		/// タイマーを開始・再開します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void start();

		/// <summary>
		/// タイマーの経過時間を返します。
		/// </summary>
		/// <returns>
		/// タイマーの経過時間
		/// </returns>
		unsigned elapsed() const;

		/// <summary>
		/// タイマーが動作中であるかを示します。
		/// </summary>
		/// <remarks>
		/// タイマーが開始されている、または開始後一時停止中である場合 true, それ以外の場合は false
		/// </remarks>
		Property_Get(bool, isActive) const;

		/// <summary>
		/// タイマーを一時停止します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void pause();

		/// <summary>
		/// タイマーを停止し、経過時間を 0 にリセットします。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void reset();

		/// <summary>
		/// 経過時間を 0 にリセットして、タイマーを開始します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void restart();

		/// <summary>
		/// タイマーの経過時間を変更します。
		/// </summary>
		/// <param name="time">
		/// 新しく設定する経過時間（単位は elapsed() と同じ）
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void set(unsigned time);

	private:

		unsigned m_startTimeMillisec = 0;

		unsigned m_accumulationMillisec = 0;

		bool m_isActive = false;

		bool m_pausing = true;

		unsigned elapsedMillisec() const;
	};

	/// <summary>
	/// 1 ミリ秒単位のタイマー
	/// </summary>
	using TimerMillisec = TimerBase<1>;

	/// <summary>
	/// 1 秒単位のタイマー
	/// </summary>
	using TimerSec = TimerBase<1000>;

	/// <summary>
	/// 1 分単位のタイマー
	/// </summary>
	using TimerMinute = TimerBase<1000 * 60>;

	/// <summary>
	/// マイクロ秒単位のタイマー
	/// </summary>
	class TimerMicrosec
	{
	public:

		/// <summary>
		/// タイマーを開始・再開します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void start();

		/// <summary>
		/// タイマーの経過時間を返します。
		/// </summary>
		/// <returns>
		/// タイマーの経過時間
		/// </returns>
		unsigned long long elapsed() const;

		/// <summary>
		/// タイマーが動作中であるかを示します。
		/// </summary>
		/// <remarks>
		/// タイマーが開始されている、または開始後一時停止中である場合 true, それ以外の場合は false
		/// </remarks>
		Property_Get(bool, isActive) const;

		/// <summary>
		/// タイマーを一時停止します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void pause();

		/// <summary>
		/// タイマーを停止し、経過時間を 0 にリセットします。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void reset();

		/// <summary>
		/// 経過時間を 0 にリセットして、タイマーを開始します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void restart();

		/// <summary>
		/// タイマーの経過時間を変更します。
		/// </summary>
		/// <param name="time">
		/// 新しく設定する経過時間（マイクロ秒）
		/// </param>
		/// <returns>
		/// なし
		/// </returns>
		void set(unsigned long long time);

	private:

		unsigned long long m_startTimeMicrosec = 0;

		unsigned long long m_accumulationMicrosec = 0;

		bool m_isActive = false;

		bool m_pausing = true;
	};
}

# include "SivTimer.inl"
