﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# include "SivTime.hpp"

namespace s3d
{
	/// <summary>
	/// タイマーを開始・再開します。
	/// </summary>
	/// <returns>
	/// なし
	/// </returns>
	template<int resolutionMilliseconds> void TimerBase<resolutionMilliseconds>::start()
	{
		if (!m_pausing)
		{
			return;
		}

		m_startTimeMillisec = Time::GetMillisec();

		m_isActive = true;

		m_pausing = false;
	}

	/// <summary>
	/// タイマーの経過時間を返します。
	/// </summary>
	/// <returns>
	/// タイマーの経過時間
	/// </returns>
	template<int resolutionMilliseconds> unsigned TimerBase<resolutionMilliseconds>::elapsed() const
	{
		const unsigned t = Time::GetMillisec();

		if (!m_isActive)
		{
			return 0;
		}

		if (m_pausing)
		{
			return m_accumulationMillisec / resolutionMilliseconds;
		}

		return (m_accumulationMillisec + (t - m_startTimeMillisec)) / resolutionMilliseconds;
	}

	template<int resolutionMilliseconds> bool TimerBase<resolutionMilliseconds>::_get_isActive() const
	{
		return m_isActive;
	}

	/// <summary>
	/// タイマーを一時停止します。
	/// </summary>
	/// <returns>
	/// なし
	/// </returns>
	template<int resolutionMilliseconds> void TimerBase<resolutionMilliseconds>::pause()
	{
		m_accumulationMillisec = elapsedMillisec();

		m_pausing = true;
	}

	/// <summary>
	/// タイマーを停止し、経過時間を 0 にリセットします。
	/// </summary>
	/// <returns>
	/// なし
	/// </returns>
	template<int resolutionMilliseconds> void TimerBase<resolutionMilliseconds>::reset()
	{
		m_accumulationMillisec = 0;

		m_isActive = false;

		m_pausing = true;
	}

	/// <summary>
	/// 経過時間を 0 にリセットして、タイマーを開始します。
	/// </summary>
	/// <returns>
	/// なし
	/// </returns>
	template<int resolutionMilliseconds> void TimerBase<resolutionMilliseconds>::restart()
	{
		reset();

		start();
	}

	/// <summary>
	/// タイマーの経過時間を変更します。
	/// </summary>
	/// <param name="time">
	/// 新しく設定する経過時間（単位は elapsed() と同じ）
	/// </param>
	/// <returns>
	/// なし
	/// </returns>
	template<int resolutionMilliseconds> void TimerBase<resolutionMilliseconds>::set(const unsigned time)
	{
		m_isActive |= (time != 0);

		m_accumulationMillisec = time * resolutionMilliseconds;

		m_startTimeMillisec = Time::GetMillisec();
	}

	template<int resolutionMilliseconds> unsigned TimerBase<resolutionMilliseconds>::elapsedMillisec() const
	{
		if (!m_isActive)
		{
			return 0;
		}

		if (m_pausing)
		{
			return m_accumulationMillisec;
		}

		return (m_accumulationMillisec + (Time::GetMillisec() - m_startTimeMillisec));
	}

	/// <summary>
	/// タイマーを開始・再開します。
	/// </summary>
	/// <returns>
	/// なし
	/// </returns>
	inline void TimerMicrosec::start()
	{
		if (!m_pausing)
		{
			return;
		}

		m_isActive = true;

		m_pausing = false;

		m_startTimeMicrosec = Time::GetMicrosec();
	}

	/// <summary>
	/// タイマーの経過時間を返します。
	/// </summary>
	/// <returns>
	/// タイマーの経過時間
	/// </returns>
	inline unsigned long long TimerMicrosec::elapsed() const
	{
		const unsigned long long t = Time::GetMicrosec();

		if (!m_isActive)
		{
			return 0;
		}

		if (m_pausing)
		{
			return m_accumulationMicrosec;
		}

		return (m_accumulationMicrosec + (t - m_startTimeMicrosec));
	}

	inline bool TimerMicrosec::_get_isActive() const
	{
		return m_isActive;
	}

	/// <summary>
	/// タイマーを一時停止します。
	/// </summary>
	/// <returns>
	/// なし
	/// </returns>
	inline void TimerMicrosec::pause()
	{
		m_accumulationMicrosec = elapsed();

		m_pausing = true;
	}

	/// <summary>
	/// タイマーを停止し、経過時間を 0 にリセットします。
	/// </summary>
	/// <returns>
	/// なし
	/// </returns>
	inline void TimerMicrosec::reset()
	{
		m_accumulationMicrosec = 0;

		m_isActive = false;

		m_pausing = true;
	}

	/// <summary>
	/// 経過時間を 0 にリセットして、タイマーを開始します。
	/// </summary>
	/// <returns>
	/// なし
	/// </returns>
	inline void TimerMicrosec::restart()
	{
		reset();

		start();
	}

	/// <summary>
	/// タイマーの経過時間を変更します。
	/// </summary>
	/// <param name="time">
	/// 新しく設定する経過時間（マイクロ秒）
	/// </param>
	/// <returns>
	/// なし
	/// </returns>
	inline void TimerMicrosec::set(const unsigned long long time)
	{
		m_isActive |= (time != 0);

		m_accumulationMicrosec = time;

		m_startTimeMicrosec = Time::GetMicrosec();
	}
}
