﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include <memory>
# include "Siv3DFwd.hpp"
# include "SivColor.hpp"

namespace s3d
{
	/// <summary>
	/// AVI の圧縮方式
	/// </summary>
	enum class VideoCompression
	{
		/// <summary>
		/// Microsoft Video 1 | Quality 40
		/// </summary>
		MSV1_Low,

		/// <summary>
		/// Microsoft Video 1 | Quality 60
		/// </summary>
		MSV1_Medium,

		/// <summary>
		/// Microsoft Video 1 | Quality 80
		/// </summary>
		MSV1_High,

		/// <summary>
		/// Microsoft Video 1 | Quality 100
		/// </summary>
		MSV1_Ultra,

		/// <summary>
		/// Intel Indeo
		/// </summary>
		Indeo,

		/// <summary>
		/// Cinepak
		/// </summary>
		Cinepak,

		/// <summary>
		/// 非圧縮
		/// </summary>
		Uncompressed,

		/// <summary>
		/// デフォルト [Microsoft Video 1 | Quality 60]
		/// </summary>
		Default = MSV1_Medium
	};

	/// <summary>
	/// 動画の書き出し
	/// </summary>
	class VideoWriter
	{
	private:

		class CVideoWriter;

		std::shared_ptr<CVideoWriter> pWriter;

	public:

		/// <summary>
		/// デフォルトコンストラクタ
		/// </summary>
		VideoWriter();

		/// <summary>
		/// 動画を書き込むファイルを開きます。
		/// </summary>
		/// <param name="path">
		/// 作成する動画のファイルパス
		/// </param>
		/// <param name="width">
		/// 動画の幅（ピクセル）
		/// </param>
		/// <param name="height">
		/// 動画の高さ（ピクセル）
		/// </param>
		/// <param name="videoCompression">
		/// AVI の圧縮方式
		/// </param>
		/// <param name="frameTime">
		/// フレームレート。 { a, b } で、 b/a[Hz]
		/// </param>
		VideoWriter(const FilePath& path, int width, int height, VideoCompression videoCompression = VideoCompression::Default, const std::pair<int, int>& frameTime = { 1, 30 });

		/// <summary>
		/// 動画の作成を終了し、保存します。
		/// </summary>
		~VideoWriter();

		/// <summary>
		/// 動画を書き込むファイルを開きます。
		/// </summary>
		/// <param name="path">
		/// 作成する動画のファイルパス
		/// </param>
		/// <param name="width">
		/// 動画の幅（ピクセル）
		/// </param>
		/// <param name="height">
		/// 動画の高さ（ピクセル）
		/// </param>
		/// <param name="videoCompression">
		/// AVI の圧縮方式
		/// </param>
		/// <param name="frameTime">
		/// フレームレート。 { a, b } で、 b/a[Hz]
		/// </param>
		/// <returns>
		/// ファイルのオープンに成功した場合 true, それ以外の場合は false
		/// </returns>
		bool open(const FilePath& path, int width, int height, VideoCompression videoCompression = VideoCompression::MSV1_High, const std::pair<int, int>& frameTime = { 1, 30 });

		/// <summary>
		/// 書き出し用ファイルがオープンされているかを返します。
		/// </summary>
		/// <returns>
		/// ファイルがオープンされている場合 true, それ以外の場合は false
		/// </returns>
		bool isOpened() const;

		/// <summary>
		/// 書き出し用ファイルがオープンされているかを返します。
		/// </summary>
		/// <returns>
		/// ファイルがオープンされている場合 true, それ以外の場合は false
		/// </returns>
		explicit operator bool() const { return isOpened(); }

		/// <summary>
		/// フレームを書き出します。
		/// </summary>
		/// <param name="image">
		/// 書き出すフレームの画像
		/// </param>
		/// <remarks>
		/// image は動画と同じサイズでなければいけません。
		/// </remarks>
		/// <returns>
		/// フレームの書き出しに成功したら true, 失敗したら false
		/// </returns>
		bool write(const Image& image);

		/// <summary>
		/// フレームを書き出します。
		/// </summary>
		/// <param name="image">
		/// 書き出すフレームの画像
		/// </param>
		/// <param name="offset">
		/// 画像の位置のオフセット
		/// </param>
		/// <param name="backgroundColor">
		/// 動画の背景色
		/// </param>
		/// <returns>
		/// フレームの書き出しに成功したら true, 失敗したら false
		/// </returns>
		bool write(const Image& image, const Point& offset, const Color& backgroundColor = Palette::Black);

		/// <summary>
		/// 動画の作成を終了し、保存します。
		/// </summary>
		/// <remarks>
		/// 新しい動画を作成するには、再度 open() する必要があります。
		/// </remarks>
		/// <returns>
		/// 動画の保存に成功したら true, 失敗したら false
		/// </returns>
		bool close();
	};
}
