﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivArray.hpp"
# include "SivOptional.hpp"
# include "SivString.hpp"
# include "SivWaveSample.hpp"
# include "SivMath.hpp"
# include "SivMathConstants.hpp"

namespace s3d
{
	/**
	*	@brief オーディオファイルのフォーマット
	*/
	enum class AudioFormat
	{
		/**
		*
		*/
		Unknown,

		/**
		*
		*/
		WAV,

		/**
		*
		*/
		OggVorbis,

		/**
		*
		*/
		MP3,

		/**
		*
		*/
		AAC,

		/**
		*
		*/
		WMA,

		//Aiff,

		//Flac,

		/**
		*
		*/
		Unspecified = Unknown,
	};

	/**
	*	@brief オーディオの情報
	*/
	struct AudioProperty
	{
		/**
		*
		*/
		AudioFormat format;

		/**
		*
		*/
		unsigned durationMillisec;

		/**
		*
		*/
		unsigned encodingBitrate;

		/**
		*
		*/
		String title;

		/**
		*
		*/
		String artist;

		/**
		*
		*/
		String albumTitle;

		/**
		*
		*/
		String albumArtist;

		/**
		*
		*/
		Optional<unsigned> year;

		/**
		*
		*/
		Optional<unsigned> trackNumber;
	};

	/**
	*	@brief MP3 オーディオの圧縮品質
	*/
	enum class MP3Quality
	{
		/**
		*	32kbps
		*/
		BR32kbps,

		/**
		*	40kbps
		*/
		BR40kbps,

		/**
		*	48kbps
		*/
		BR48kbps,

		/**
		*	56kbps
		*/
		BR56kbps,

		/**
		*	64kbps
		*/
		BR64kbps,

		/**
		*	80kbps
		*/
		BR80kbps,

		/**
		*	96kbps
		*/
		BR96kbps,

		/**
		*	112kbps
		*/
		BR112kbps,

		/**
		*	128kbps
		*/
		BR128kbps,

		/**
		*	160kbps
		*/
		BR160kbps,

		/**
		*	192kbps
		*/
		BR192kbps,

		/**
		*	224kbps
		*/
		BR224kbps,

		/**
		*	256kbps
		*/
		BR256kbps,

		/**
		*	320kbps
		*/
		BR320kbps,

		/**
		*
		*/
		Default = BR192kbps,
	};

	/**
	*	@brief AAC オーディオの圧縮品質
	*/
	enum class AACQuality
	{
		/**
		*	96kbps
		*/
		BR96kbps,

		/**
		*	128kbps
		*/
		BR128kbps,

		/**
		*	160kbps
		*/
		BR160kbps,

		/**
		*	192kbps
		*/
		BR192kbps,

		/**
		*
		*/
		Default = BR160kbps,
	};

	/**
	*	brief 波形
	*/
	class Wave
	{
	private:

		Array<WaveSample> m_data;

		unsigned m_samplingRate = DefaultSamplingRate;

	public:

		using iterator			= Array<WaveSample>::iterator;
		using const_iterator	= Array<WaveSample>::const_iterator;
		using reverse_iterator	= Array<WaveSample>::reverse_iterator;	
		using const_reverse_iterator = Array<WaveSample>::const_reverse_iterator;

		enum
		{
			/**
			*
			*/
			MinSamlingRate = 4000,

			/**
			*
			*/
			DefaultSamplingRate = 44100,

			/**
			*
			*/
			MaxSamplingRate = 192000,
		};

		/**
		*
		*/
		Wave();

		/**
		*
		*/
		~Wave();

		/**
		*
		*	@param sample
		*/
		explicit Wave(unsigned sample);

		/**
		*
		*	@param sample
		*	@param samplingRate
		*/
		Wave(unsigned sample, unsigned samplingRate);

		/**
		*
		*	@param sample
		*	@param samplingRate
		*	@param hz
		*/
		Wave(unsigned sample, unsigned samplingRate, double hz);

		/**
		*
		*	@param path
		*/
		explicit Wave(const FilePath& path);

		Wave(IReader& reader, AudioFormat format = AudioFormat::Unspecified);

		/**
		*
		*	@param wave
		*/
		Wave(const Wave& wave);

		/**
		*
		*	@param wave
		*/
		Wave(Wave && wave);

		/**
		*
		*	@param wave
		*	@return
		*/
		Wave& operator =(const Wave& wave);

		/**
		*
		*	@param wave
		*	@return
		*/
		Wave& operator =(Wave && wave);

		/**
		*
		*/
		void clear();

		/**
		*
		*/
		void fillZero();

		/**
		*
		*	@param samplingRate
		*	@return
		*/
		unsigned setSamplingRate(unsigned samplingRate);

		/**
		*
		*	@return
		*/
		explicit operator bool() const
		{
			return !isEmpty;
		}

		/**
		*
		*	@param sample
		*/
		void resize(unsigned sample);

		/**
		*
		*	@param pos
		*	@return
		*/
		WaveSample& at(unsigned pos);

		/**
		*
		*	@param pos
		*	@return
		*/
		const WaveSample& at(unsigned pos) const;

		/**
		*
		*	@param pos
		*	@return
		*/
		WaveSample& operator [](unsigned pos)
		{
			return m_data[pos];
		}

		/**
		*
		*	@param pos
		*	@return
		*/
		const WaveSample& operator [](unsigned pos) const
		{
			return m_data[pos];
		}

		/**
		*
		*	@return
		*/
		void*
			data(
			)
		{
			return &m_data[0];
		}

		/**
		*
		*	@return
		*/
		const void*
			data(
			) const
		{
			return &m_data[0];
		}

		/**
		*	波形の先頭位置のイテレータを取得します。
		*	@return 波形の先頭位置のイテレータ
		*/
		iterator begin() { return m_data.begin(); }

		/**
		*	波形の先頭位置のイテレータを取得します。
		*	@return 波形の先頭位置のイテレータ
		*/
		const_iterator begin() const { return m_data.begin(); }

		/**
		*	波形の先頭位置のイテレータを取得します。
		*	@return 波形の先頭位置のイテレータ
		*/
		const_iterator cbegin() const { return m_data.cbegin(); }

		/**
		*	波形の終了位置のイテレータを取得します。
		*	@return 波形の終了位置のイテレータ
		*/
		iterator end() { return m_data.end(); }

		/**
		*	波形の終了位置のイテレータを取得します。
		*	@return 波形の終了位置のイテレータ
		*/
		const_iterator end() const { return m_data.end(); }

		/**
		*	波形の終了位置のイテレータを取得します。
		*	@return 波形の終了位置のイテレータ
		*/
		const_iterator cend() const { return m_data.cend(); }

		/**
		*	波形の末尾へのリバースイテレータを取得します。
		*	@return 波形の末尾へのリバースイテレータ
		*/
		reverse_iterator rbegin() { return m_data.rbegin(); }

		/**
		*	波形の末尾へのリバースイテレータを取得します。
		*	@return 波形の末尾へのリバースイテレータ
		*/
		const_reverse_iterator rbegin() const { return m_data.rbegin(); }

		/**
		*	波形の末尾へのリバースイテレータを取得します。
		*	@return 波形の末尾へのリバースイテレータ
		*/
		const_reverse_iterator crbegin() const { return m_data.crbegin(); }

		/**
		*	波形の先頭の前へのリバースイテレータを取得します。
		*	@return 波形の先頭の前へのリバースイテレータ
		*/
		reverse_iterator rend() { return m_data.rend(); }

		/**
		*	波形の先頭の前へのリバースイテレータを取得します。
		*	@return 波形の先頭の前へのリバースイテレータ
		*/
		const_reverse_iterator rend() const { return m_data.rend(); }

		/**
		*	波形の先頭の前へのリバースイテレータを取得します。
		*	@return 波形の先頭の前へのリバースイテレータ
		*/
		const_reverse_iterator crend() const { return m_data.crend(); }

		/**
		*
		*	@param wav
		*/
		void swap(Wave& wav)
		{
			m_data.swap(wav.m_data);
			
			std::swap(m_samplingRate, wav.m_samplingRate);
		}

		/**
		*
		*	@return
		*/
		Wave clone() const
		{
			return *this;
		}

		/*
		*
		*/
		Property_Get(double, lengthSec) const { return static_cast<double>(lengthSample) / m_samplingRate; }

		/*
		*
		*/
		Property_Get(unsigned, lengthSample) const { return m_data.size(); }

		/*
		*
		*/
		Property_Get(unsigned, samplingRate) const { return m_samplingRate; }

		/*
		*
		*/
		Property_Get(unsigned, memorySize) const { return lengthSample * sizeof(WaveSample); }

		/*
		*
		*/
		Property_Get(bool, isEmpty) const { return m_data.empty(); }

		/**
		*
		*	@param path
		*	@param format
		*	@return
		*/
		bool save(const FilePath& path, AudioFormat format = AudioFormat::Unspecified) const;

		/**
		*
		*	@param path
		*	@param quality 圧縮品質 [0,100]
		*	@return
		*/
		bool saveOggVorbis(const FilePath& path, int quality = 60) const;

		/**
		*
		*	@param path
		*	@param quality
		*	@return
		*/
		bool saveMP3(const FilePath& path, MP3Quality quality = MP3Quality::Default) const;

		/**
		*
		*	@param path
		*	@param quality
		*	@return
		*/
		bool saveAAC(const FilePath& path, AACQuality quality = AACQuality::Default) const;

		MemoryReader encode(AudioFormat format) const;

		MemoryReader encodeOggVorbis(int quality = 60) const;

		MemoryReader encodeMP3(MP3Quality quality = MP3Quality::Default) const;

		MemoryReader encodeAAC(AACQuality quality = AACQuality::Default) const;

		/**
		*
		*	@return
		*/
		Wave& swapChannel();

		/**
		*
		*	@return
		*/
		Wave& reverse();
	};

	/**
	*	@brief
	*/
	namespace Audio
	{
		/**
		*
		*	@param path
		*	@return
		*/
		AudioFormat GetFormat(const FilePath& path);

		AudioFormat GetFormat(const IReader& reader);

		AudioFormat GetFormatFromFileName(const FilePath& path);

		/**
		*
		*	@param path
		*	@return
		*/
		Optional<AudioProperty> GetProperty(const FilePath& path);

		/**
		*
		*	@return
		*/
		bool MP3EncoderAvailable();

		/**
		*
		*	@return
		*/
		bool AACEncoderAvailable();

		/**
		*
		*	@return
		*/
		bool WMAEncoderAvailable();
	}

	/**
	*	@brief
	*/
	namespace Waving
	{
		/**
		*	テンポを変更した Wave を返します。
		*	@param wave 元の Wave
		*	@param tempo 1.0 を基準としたテンポ。小さいと遅く、大きいと速くなる
		*	@param quickSeek 品質を少し落として処理を高速化するかのフラグ
		*	@return テンポを変更した Wave
		*/
		void ChangeTempo(const Wave& src, Wave& dst, double tempo, bool quickSeek = false);

		/**
		*	ピッチを変更した Wave を返します。
		*	@param wave 元の Wave
		*	@param pitchSemitones ピッチの変更 (半音上げる場合は 1.0, 1 オクターブ上げる場合は 12.0, 半音下げる場合は -1.0)
		*	@param quickSeek 品質を少し落として処理を高速化するかのフラグ
		*	@return ピッチを変更した Wave
		*/
		void ChangePitchSemitones(const Wave& src, Wave& dst, double pitchSemitones, bool quickSeek = false);

		/**
		*	サンプリングレートを変更した Wave を返します。
		*	@param wave 元の Wave
		*	@param samplingRate サンプリングレート
		*	@param quickSeek 品質を少し落として処理を高速化するかのフラグ
		*	@return サンプリングレートを変更した Wave
		*/
		void ChangeSamplingRate(const Wave& src, Wave& dst, unsigned samplingRate, bool quickSeek = false);

		/**
		*
		*	@param mono
		*	@return
		*/
		WaveSample DoubleToSample(double mono);

		/**
		*
		*	@param left
		*	@param right
		*	@return
		*/
		WaveSample DoubleToSample(double left, double right);

		/**
		*
		*	@param wave
		*	@return
		*/
		void SwapChannel(const Wave& src, Wave& dst);

		/**
		*
		*	@param wave
		*	@return
		*/
		void Reverse(const Wave& wave);

		/**
		*
		*	@param t
		*	@return
		*/
		inline double SinWave440Hz(double t)
		{
			return Math::Sin(t*440.0*Math::Pi*2.0);
		}

		/**
		*
		*	@param lengthSec
		*	@param function
		*	@param rate
		*	@return
		*/
		Wave Synthesize(double lengthSec, std::function<double(double)> function = SinWave440Hz, unsigned rate = Wave::DefaultSamplingRate);
	}
}
