﻿//-----------------------------------------------
//
//	This file is part of the Siv3D Engine.
//
//	Copyright (C) 2008-2014 Ryo Suzuki
//
//	Licensed under the MIT License.
//
//-----------------------------------------------

# pragma once
# include "Siv3DFwd.hpp"
# include "SivArray.hpp"
# include "SivString.hpp"

namespace s3d
{
	/// <summary>
	/// Web カメラの情報
	/// </summary>
	struct WebcamInfo
	{
		/// <summary>
		/// Web カメラの名称
		/// </summary>
		String name;

		/// <summary>
		/// デバイス ID
		/// </summary>
		int deviceID = 0;
	};

	/// <summary>
	/// Web カメラ
	/// </summary>
	/// <remarks>
	/// Web カメラに関する機能を提供します。
	/// </remarks>
	namespace Webcam
	{
		/// <summary>
		/// 利用可能な Web カメラの一覧を取得します。
		/// </summary>
		/// <returns>
		/// 利用可能な Web カメラの一覧
		/// </returns>
		Array<WebcamInfo> Enumerate();

		/// <summary>
		/// Web カメラによる撮影を開始します。
		/// </summary>
		/// <param name="deviceID">
		/// デバイス ID
		/// </param>
		/// <remarks>
		/// ユーザが Web カメラの使用を許可していない場合、Web カメラ使用の認証画面を表示します。
		/// </remarks>
		/// <returns>
		/// Web カメラの起動に成功した場合 true、それ以外の場合は false
		/// </returns>
		bool Start(unsigned deviceID = 0);

		/// <summary>
		/// Web カメラ使用の認証画面を表示します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void RequestAccess();

		/// <summary>
		/// Web カメラの使用がユーザによって許可されているかを返します。
		/// </summary>
		/// <returns>
		/// Web カメラの使用がユーザによって許可されている場合 true, それ以外の場合は false
		/// </returns>
		bool Approved();

		/// <summary>
		/// Web カメラの使用がユーザによって拒否されているかを返します。
		/// </summary>
		/// <returns>
		/// Web カメラの使用がユーザによって拒否されている場合 true, それ以外の場合は false
		/// </returns>
		bool Denied();

		/// <summary>
		/// Web カメラによる撮影を終了します。
		/// </summary>
		/// <returns>
		/// なし
		/// </returns>
		void Stop();

		/// <summary>
		/// Web カメラがアクティブかどうかを示します。
		/// </summary>
		/// <returns>
		/// Web カメラがアクティブな場合 true, それ以外の場合は false
		/// </returns>
		bool IsActive();

		/// <summary>
		/// 新しいフレームがあるかを取得します。
		/// </summary>
		/// <remarks>
		/// GetFrame() すると、最新のフレームがコピーされ、
		/// 次に新しい画像が撮影されるまで HasNewFrame() は false を返します。
		/// </remarks>
		/// <remarks>
		/// ユーザが Web カメラの使用を許可していない場合、常に false を返します。
		/// </remarks>
		/// <returns>
		/// 新しいフレームがある場合 true, それ以外の場合は false
		/// </returns>
		bool HasNewFrame();

		/// <summary>
		/// Web カメラの最新のフレームを取得します。
		/// </summary>
		/// <param name="image">
		/// フレームのコピー先
		/// </param>
		/// <remarks>
		/// ユーザが Web カメラの使用を許可していない場合、関数は失敗します。
		/// </remarks>
		/// <returns>
		/// 取得に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool GetFrame(Image& image);

		/// <summary>
		/// Web カメラの最新のフレームを取得します。
		/// </summary>
		/// <param name="texure">
		/// フレームのコピー先
		/// </param>
		/// <remarks>
		/// ユーザが Web カメラの使用を許可していない場合、関数は失敗します。
		/// </remarks>
		/// <returns>
		/// 取得に成功した場合 true, それ以外の場合は false
		/// </returns>
		bool GetFrame(DynamicTexture& texure);
	}
}
